@@ -13,6 +13,9 @@ struct args {
 	uint32_t len;
 };
 
+void shell_escape(struct workspace *wk, struct sbuf *sb, const char *str);
+void shell_escape_no_dollar(struct workspace *wk, struct sbuf *sb, const char *str);
+void shell_escape_custom(struct workspace *wk, struct sbuf *sb, const char *str, const char *escape_inner, const char *need_escaping);
 void ninja_escape(struct workspace *wk, struct sbuf *sb, const char *str);
 void pkgconf_escape(struct workspace *wk, struct sbuf *sb, const char *str);
 
@@ -21,6 +24,7 @@ void push_args_null_terminated(struct workspace *wk, obj arr, char *const *argv)
 
 obj join_args_plain(struct workspace *wk, obj arr);
 obj join_args_shell(struct workspace *wk, obj arr);
+obj join_args_shell_no_dollar(struct workspace *wk, obj arr);
 obj join_args_ninja(struct workspace *wk, obj arr);
 obj join_args_shell_ninja(struct workspace *wk, obj arr);
 obj join_args_pkgconf(struct workspace *wk, obj arr);
