@@ -5,8 +5,8 @@
 
 #ifndef MUON_FUNCTIONS_ARRAY_H
 #define MUON_FUNCTIONS_ARRAY_H
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 
-extern const struct func_impl_name impl_tbl_array[];
-extern const struct func_impl_name impl_tbl_array_internal[];
+extern const struct func_impl impl_tbl_array[];
+extern const struct func_impl impl_tbl_array_internal[];
 #endif
