@@ -5,7 +5,7 @@
 
 #ifndef MUON_FUNCTIONS_BOOLEAN_H
 #define MUON_FUNCTIONS_BOOLEAN_H
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 
-extern const struct func_impl_name impl_tbl_boolean[];
+extern const struct func_impl impl_tbl_boolean[];
 #endif
