@@ -5,9 +5,9 @@
 
 #ifndef MUON_FUNCTIONS_EXTERNAL_PROGRAM_H
 #define MUON_FUNCTIONS_EXTERNAL_PROGRAM_H
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 
-void find_program_guess_version(struct workspace *wk, const char *path, obj *ver);
+void find_program_guess_version(struct workspace *wk, obj cmd_array, obj version_argument, obj *ver);
 
-extern const struct func_impl_name impl_tbl_external_program[];
+extern const struct func_impl impl_tbl_external_program[5];
 #endif
