@@ -5,7 +5,7 @@
 
 #ifndef FUNCTIONS_xxx_H
 #define FUNCTIONS_xxx_H
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 
-extern const struct func_impl_name impl_tbl_xxx[];
+extern const struct func_impl impl_tbl_xxx[];
 #endif
