@@ -9,6 +9,8 @@
 #include "error.h"
 #include "workspace.h"
 
+struct ast;
+
 enum az_diagnostic {
 	az_diagnostic_unused_variable = 1 << 0,
 	az_diagnostic_reassign_to_conflicting_type = 1 << 1,
@@ -19,6 +21,7 @@ enum az_diagnostic {
 struct az_opts {
 	bool subdir_error;
 	bool eval_trace;
+	bool analyze_project_call_only;
 	enum error_diagnostic_store_replay_opts replay_opts;
 	const char *file_override, *internal_file, *get_definition_for;
 	uint64_t enabled_diagnostics;
@@ -53,4 +56,6 @@ extern struct func_impl_group az_func_impl_group;
 bool do_analyze(struct az_opts *opts);
 
 void eval_trace_print(struct workspace *wk, obj trace);
+
+bool analyze_project_call(struct workspace *wk);
 #endif
