@@ -9,13 +9,15 @@
 #include <stdbool.h>
 #include <stdint.h>
 
-#include "lang/workspace.h"
+struct source;
+struct workspace;
 
 enum vm_compile_mode {
 	vm_compile_mode_fmt = 1 << 1,
 	vm_compile_mode_quiet = 1 << 2,
 	vm_compile_mode_language_extended = 1 << 3,
 	vm_compile_mode_expr = 1 << 4,
+	vm_compile_mode_return_after_project = 1 << 5,
 };
 
 void vm_compile_state_reset(struct workspace *wk);
