@@ -9,23 +9,22 @@
 
 #include "compat.h"
 
-#include <stdarg.h>
 #include <stdio.h>
 
 #include "compilers.h"
 #include "datastructures/bucket_arr.h"
-#include "datastructures/iterator.h"
+#include "iterator.h"
 #include "lang/types.h"
 #include "machines.h"
 
 enum obj_type {
 	/* singleton object types */
-	obj_null,
-	obj_meson,
-	obj_disabler,
+	obj_null = 0,
+	obj_disabler = 1,
+	obj_meson = 2,
+	obj_bool, // obj_bool_true, obj_bool_false
 
 	/* simple object types */
-	obj_bool,
 	obj_file,
 	obj_feature_opt,
 	obj_machine,
@@ -74,15 +73,11 @@ struct obj_typeinfo {
 };
 
 struct obj_func {
-	const char *name;
-	struct ast *ast;
+	const char *name, *desc;
 	enum language_mode lang_mode;
-	uint32_t args_id, block_id, nargs, nkwargs;
-	obj kwarg_defaults, src, scope_stack;
+	uint32_t nargs, nkwargs;
 	type_tag return_type;
 
-	/*--------*/
-
 	uint32_t entry;
 	struct args_norm an[32];
 	struct args_kw akw[64];
@@ -90,7 +85,8 @@ struct obj_func {
 
 struct obj_capture {
 	struct obj_func *func;
-	obj scope_stack, defargs;
+	obj scope_stack, defargs, self;
+	uint32_t native_func;
 };
 
 enum tgt_type {
@@ -117,6 +113,8 @@ enum feature_opt_state {
 	_(python3, "public", true)                    \
 	_(sourceset, "public", true)                  \
 	_(toolchain, "private", true)                 \
+	_(subprojects, "private", true)                 \
+	_(getopt, "private", true)                 \
 	_(cmake, "public", false)                     \
 	_(dlang, "public", false)                     \
 	_(gnome, "public", false)                     \
@@ -169,18 +167,25 @@ struct obj_module {
 	obj exports;
 };
 
+struct obj_array_elem {
+	uint32_t next;
+	obj val;
+};
+
+enum obj_array_flags {
+	obj_array_flag_cow = 1 << 3,
+};
+
 struct obj_array {
-	obj val; // any
-	obj next; // obj_array
-	obj tail; // obj_array
-	uint32_t len;
-	bool have_next;
+	uint32_t head, tail, len;
+	enum obj_array_flags flags;
 };
 
 enum obj_dict_flags {
 	obj_dict_flag_big = 1 << 0,
 	obj_dict_flag_int_key = 1 << 1,
 	obj_dict_flag_dont_expand = 1 << 2,
+	obj_dict_flag_cow = 1 << 3,
 };
 
 struct obj_dict_elem {
@@ -207,24 +212,26 @@ enum build_tgt_flags {
 struct build_dep {
 	enum compiler_language link_language;
 
-	obj link_whole; // obj_array
+	obj link_whole;
 
-	obj link_with; // obj_array
-	obj link_with_not_found; // obj_array
+	obj link_with;
+	obj link_with_not_found;
+	obj frameworks;
 
-	obj link_args; // obj_array
-	obj compile_args; // obj_array
+	obj link_args;
+	obj compile_args;
 
-	obj include_directories; // obj_array
+	obj include_directories;
 
-	obj sources; // obj_array
-	obj objects; // obj_array
+	obj sources;
+	obj objects;
 
-	obj order_deps; // obj_array
-	obj rpath; // obj_array
+	obj order_deps;
+	obj rpath;
 
 	struct {
 		obj deps;
+		obj order_deps;
 		obj link_with;
 		obj link_whole;
 	} raw;
@@ -238,13 +245,19 @@ struct obj_build_target {
 	obj cwd; // obj_string
 	obj build_dir; // obj_string
 	obj soname; // obj_string
+	obj implib; // obj_string
 	obj src; // obj_array
 	obj objects; // obj_array
 	obj args; // obj_dict
+	obj processed_args_pch; // obj_dict
+	obj processed_args; // obj_dict
 	obj link_depends; // obj_array
 	obj generated_pc; // obj_string
 	obj override_options; // obj_array
 	obj required_compilers; // obj_dict
+	obj extra_files; // obj_array
+	obj pch; // obj_dict
+	obj callstack; // obj_array
 
 	struct build_dep dep;
 	struct build_dep dep_internal;
@@ -252,6 +265,7 @@ struct obj_build_target {
 	enum compiler_visibility_type visibility;
 	enum build_tgt_flags flags;
 	enum tgt_type type;
+	enum machine_kind machine;
 };
 
 struct obj_both_libs {
@@ -276,6 +290,7 @@ struct obj_custom_target {
 	obj private_path; // obj_string
 	obj env; // str | list[str] | dict[str] | env
 	obj depfile; // str
+	obj callstack; // obj_array
 	enum custom_target_flags flags;
 };
 
@@ -313,6 +328,7 @@ struct obj_dependency {
 	enum dep_flags flags;
 	enum dependency_type type;
 	enum include_type include_type;
+	enum machine_kind machine;
 };
 
 struct obj_external_program {
@@ -375,15 +391,13 @@ struct obj_test {
 };
 
 struct obj_compiler {
-	obj cmd_arr;
-	obj linker_cmd_arr;
-	obj static_linker_cmd_arr;
+	obj cmd_arr[toolchain_component_count];
+	obj overrides[toolchain_component_count];
+	uint32_t type[toolchain_component_count];
 	obj ver;
 	obj libdirs;
-	obj overrides[3];
-	uint32_t type[3];
 	enum compiler_language lang;
-	bool linker_passthrough;
+	enum machine_kind machine;
 };
 
 enum install_target_type {
@@ -448,6 +462,7 @@ struct obj_option {
 	obj min;
 	obj deprecated;
 	obj description;
+	uint32_t ip;
 	enum option_value_source source;
 	enum build_option_type type;
 	enum build_option_kind kind;
@@ -495,7 +510,7 @@ struct range_params {
 struct obj_iterator {
 	enum obj_iterator_type type;
 	union {
-		struct obj_array *array;
+		struct obj_array_elem *array;
 		struct obj_dict_elem *dict_small;
 		struct {
 			struct hash *h;
@@ -520,12 +535,14 @@ struct obj_clear_mark {
 void make_obj(struct workspace *wk, obj *id, enum obj_type type);
 enum obj_type get_obj_type(struct workspace *wk, obj id);
 
+void make_default_objects(struct workspace *wk);
+
 void obj_set_clear_mark(struct workspace *wk, struct obj_clear_mark *mk);
 void obj_clear(struct workspace *wk, const struct obj_clear_mark *mk);
 
-void set_obj_bool(struct workspace *wk, obj o, bool v);
 bool get_obj_bool(struct workspace *wk, obj o);
-void set_obj_bool(struct workspace *wk, obj o, bool v);
+obj make_obj_bool(struct workspace *wk, bool v);
+obj get_obj_bool_with_default(struct workspace *wk, obj o, bool def);
 obj make_number(struct workspace *wk, int64_t n);
 int64_t get_obj_number(struct workspace *wk, obj o);
 void set_obj_number(struct workspace *wk, obj o, int64_t v);
@@ -574,6 +591,7 @@ struct sbuf;
 const char *obj_type_to_s(enum obj_type t);
 bool s_to_type_tag(const char *s, type_tag *t);
 void obj_to_s(struct workspace *wk, obj o, struct sbuf *sb);
+void obj_to_json(struct workspace *wk, obj o, struct sbuf *sb);
 bool obj_equal(struct workspace *wk, obj left, obj right);
 bool obj_clone(struct workspace *wk_src, struct workspace *wk_dest, obj val, obj *ret);
 
@@ -582,19 +600,17 @@ bool obj_clone(struct workspace *wk_src, struct workspace *wk_dest, obj val, obj
 		char buf[4096];                                   \
 		log_print_prefix(log_debug, buf, ARRAY_LEN(buf)); \
 		log_plain("%s", buf);                             \
-		obj_fprintf(wk, log_file(), __VA_ARGS__);         \
+		obj_lprintf(wk, __VA_ARGS__);         \
 	}
 #define LOBJ(object_id) LO("%s: %o\n", #object_id, object_id)
 
-bool obj_vasprintf(struct workspace *wk, struct sbuf *sb, const char *fmt, va_list ap);
-bool obj_asprintf(struct workspace *wk, struct sbuf *sb, const char *fmt, ...) MUON_ATTR_FORMAT(printf, 3, 4);
-bool obj_vfprintf(struct workspace *wk, FILE *f, const char *fmt, va_list ap) MUON_ATTR_FORMAT(printf, 3, 0);
+bool obj_lprintf(struct workspace *wk, const char *fmt, ...) MUON_ATTR_FORMAT(printf, 2, 3);
 bool obj_fprintf(struct workspace *wk, FILE *f, const char *fmt, ...) MUON_ATTR_FORMAT(printf, 3, 4);
 bool obj_printf(struct workspace *wk, const char *fmt, ...) MUON_ATTR_FORMAT(printf, 2, 3);
 uint32_t obj_vsnprintf(struct workspace *wk, char *buf, uint32_t len, const char *fmt, va_list ap);
 uint32_t obj_snprintf(struct workspace *wk, char *buf, uint32_t len, const char *fmt, ...)
 	MUON_ATTR_FORMAT(printf, 4, 5);
-void obj_inspect(struct workspace *wk, FILE *out, obj val);
+void obj_inspect(struct workspace *wk, obj val);
 
 typedef enum iteration_result (*obj_array_iterator)(struct workspace *wk, void *ctx, obj val);
 void obj_array_push(struct workspace *wk, obj arr, obj child);
@@ -603,10 +619,12 @@ bool obj_array_foreach(struct workspace *wk, obj arr, void *ctx, obj_array_itera
 bool obj_array_foreach_flat(struct workspace *wk, obj arr, void *usr_ctx, obj_array_iterator cb);
 bool obj_array_in(struct workspace *wk, obj arr, obj val);
 bool obj_array_index_of(struct workspace *wk, obj arr, obj val, uint32_t *idx);
+obj *obj_array_index_pointer(struct workspace *wk, obj arr, int64_t i);
 void obj_array_index(struct workspace *wk, obj arr, int64_t i, obj *res);
 void obj_array_extend(struct workspace *wk, obj arr, obj arr2);
 void obj_array_extend_nodup(struct workspace *wk, obj arr, obj arr2);
 void obj_array_dup(struct workspace *wk, obj arr, obj *res);
+obj obj_array_dup_light(struct workspace *wk, obj src);
 bool obj_array_join(struct workspace *wk, bool flat, obj arr, obj join, obj *res);
 void obj_array_tail(struct workspace *wk, obj arr, obj *res);
 void obj_array_set(struct workspace *wk, obj arr, int64_t i, obj v);
@@ -619,16 +637,20 @@ int32_t obj_array_sort_by_str(struct workspace *wk, void *_ctx, obj a, obj b);
 void obj_array_sort(struct workspace *wk, void *usr_ctx, obj arr, obj_array_sort_func func, obj *res);
 obj obj_array_slice(struct workspace *wk, obj arr, int64_t i0, int64_t i1);
 obj obj_array_get_tail(struct workspace *wk, obj arr);
+obj obj_array_get_head(struct workspace *wk, obj arr);
 obj obj_array_pop(struct workspace *wk, obj arr);
+void obj_array_clear(struct workspace *wk, obj arr);
 
 typedef enum iteration_result (*obj_dict_iterator)(struct workspace *wk, void *ctx, obj key, obj val);
 bool obj_dict_foreach(struct workspace *wk, obj dict, void *ctx, obj_dict_iterator cb);
 bool obj_dict_in(struct workspace *wk, obj dict, obj key);
 bool obj_dict_index(struct workspace *wk, obj dict, obj key, obj *res);
 bool obj_dict_index_strn(struct workspace *wk, obj dict, const char *str, uint32_t len, obj *res);
+obj *obj_dict_index_strn_pointer(struct workspace *wk, obj dict, const char *str, uint32_t len);
 bool obj_dict_index_str(struct workspace *wk, obj dict, const char *str, obj *res);
 void obj_dict_set(struct workspace *wk, obj dict, obj key, obj val);
 void obj_dict_dup(struct workspace *wk, obj dict, obj *res);
+void obj_dict_dup_light(struct workspace *wk, obj dict, obj *res);
 void obj_dict_merge(struct workspace *wk, obj dict, obj dict2, obj *res);
 void obj_dict_merge_nodup(struct workspace *wk, obj dict, obj dict2);
 void obj_dict_seti(struct workspace *wk, obj dict, uint32_t key, obj val);
