@@ -14,7 +14,6 @@
 #include <stdio.h>
 
 #include "lang/types.h"
-#include "platform/assert.h"
 
 extern const char *log_level_clr[log_level_count];
 extern const char *log_level_name[log_level_count];
@@ -32,6 +31,8 @@ extern const char *log_level_shortname[log_level_count];
 
 void log_init(void);
 void log_set_file(FILE *log_file);
+struct sbuf;
+void log_set_buffer(struct sbuf *buf);
 void log_set_lvl(enum log_level lvl);
 void log_set_prefix(const char *prefix);
 const char *log_get_prefix(void);
@@ -41,6 +42,8 @@ void log_print(bool nl, enum log_level lvl, const char *fmt, ...) MUON_ATTR_FORM
 bool log_clr(void);
 void log_plain(const char *fmt, ...) MUON_ATTR_FORMAT(printf, 1, 2);
 void log_plainv(const char *fmt, va_list ap);
-FILE *log_file(void);
 bool log_should_print(enum log_level lvl);
+
+FILE *_log_file(void);
+struct sbuf *_log_sbuf(void);
 #endif
