@@ -12,6 +12,8 @@
 #include "platform/uname.h"
 
 enum machine_system {
+	machine_system_uninitialized = 0,
+	machine_system_unknown = 1,
 	machine_system_dragonfly,
 	machine_system_freebsd,
 	machine_system_gnu,
@@ -26,13 +28,13 @@ enum machine_system {
 	machine_system_cygwin,
 	machine_system_msys2,
 	machine_system_darwin,
-	machine_system_unknown,
 };
 
 enum machine_kind {
 	machine_kind_build,
 	machine_kind_host,
 };
+#define machine_kind_count 2 // Keep in sync with above
 
 struct machine_definition {
 	enum machine_kind kind;
@@ -45,7 +47,9 @@ struct machine_definition {
 };
 
 extern struct machine_definition build_machine, host_machine;
+extern const struct machine_definition *machine_definitions[machine_kind_count];
 
+const char *machine_kind_to_s(enum machine_kind kind);
 const char *machine_system_to_s(enum machine_system sys);
 const char *machine_system_to_kernel_name(enum machine_system sys);
 
