@@ -36,9 +36,8 @@ struct win_pipe_inst {
 	OVERLAPPED overlapped;
 	HANDLE handle;
 	HANDLE child_handle;
-	HANDLE event;
 	char overlapped_buf[4 << 10];
-	bool is_pending, is_eof;
+	bool is_reading, is_eof;
 };
 #endif
 
@@ -50,9 +49,11 @@ struct run_cmd_ctx {
 	int status;
 	enum run_cmd_ctx_flags flags;
 #ifdef _WIN32
-	HANDLE process;
+	HANDLE process, ioport;
 	bool close_pipes;
 	struct win_pipe_inst pipe_out, pipe_err;
+	struct sbuf env;
+	uint32_t cnt_open;
 #else
 	int pipefd_out[2], pipefd_err[2];
 	int input_fd;
@@ -78,6 +79,8 @@ bool run_cmd_determine_interpreter(struct source *src,
 	const char **new_argv0,
 	const char **new_argv1);
 
+#define ARGV(...) (char *const *)(const char *const []){ __VA_ARGS__, 0 }
+
 bool run_cmd(struct run_cmd_ctx *ctx, const char *argstr, uint32_t argc, const char *envstr, uint32_t envc);
 bool run_cmd_argv(struct run_cmd_ctx *ctx, char *const *argv, const char *envstr, uint32_t envc);
 enum run_cmd_state run_cmd_collect(struct run_cmd_ctx *ctx);
