@@ -4,7 +4,7 @@
  */
 
 #ifdef TRACY_ENABLE
-#include "TracyC.h"
+#include "tracy/TracyC.h"
 
 #define TracyCZoneAutoS TracyCZoneN(tctx_func, __func__, true)
 #define TracyCZoneAutoE TracyCZoneEnd(tctx_func)
