@@ -15,6 +15,7 @@
 #include "lang/serial.h"
 #include "lang/workspace.h"
 #include "log.h"
+#include "platform/assert.h"
 #include "platform/filesystem.h"
 #include "platform/path.h"
 
@@ -51,6 +52,13 @@ write_custom_target_dat(struct workspace *wk, struct obj_custom_target *tgt, obj
 	sbuf_pushf(wk, &name, "%s%d.dat", get_cstr(wk, tgt->name), custom_tgt_dat_sequence);
 	++custom_tgt_dat_sequence;
 
+	uint32_t i;
+	for (i = 0; i < name.len; ++i) {
+		if (!strchr("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789:-_", name.buf[i])) {
+			name.buf[i] = '_';
+		}
+	}
+
 	SBUF(dirpath);
 	SBUF(dat_path);
 	path_join(wk, &dirpath, wk->muon_private, dir);
@@ -81,18 +89,19 @@ ninja_write_custom_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *c
 	obj outputs, inputs = 0, cmdline;
 
 	if (tgt->input) {
-		relativize_paths(wk, tgt->input, false, &inputs);
+		ca_relativize_paths(wk, tgt->input, false, &inputs);
 	}
 
 	make_obj(wk, &outputs, obj_array);
 	if (tgt->output) {
-		relativize_paths(wk, tgt->output, false, &outputs);
+		ca_relativize_paths(wk, tgt->output, false, &outputs);
 	} else {
 		assert(tgt->name && "unnamed targets cannot have no output");
 		obj name;
 
 		if (ctx->proj->subproject_name) {
-			name = make_strf(wk, "%s@@%s", get_cstr(wk, ctx->proj->subproject_name), get_cstr(wk, tgt->name));
+			name = make_strf(
+				wk, "%s@@%s", get_cstr(wk, ctx->proj->subproject_name), get_cstr(wk, tgt->name));
 		} else {
 			name = tgt->name;
 		}
@@ -111,7 +120,7 @@ ninja_write_custom_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *c
 		obj elem;
 		obj_array_index(wk, tgt->output, 0, &elem);
 
-		relativize_path_push(wk, elem, cmdline);
+		ca_relativize_path_push(wk, elem, cmdline);
 	}
 
 	if (tgt->flags & custom_target_feed) {
@@ -120,7 +129,7 @@ ninja_write_custom_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *c
 		obj elem;
 		obj_array_index(wk, tgt->input, 0, &elem);
 
-		relativize_path_push(wk, elem, cmdline);
+		ca_relativize_path_push(wk, elem, cmdline);
 	}
 
 	if (tgt->env) {
@@ -152,7 +161,7 @@ ninja_write_custom_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *c
 	}
 
 	obj depends_rel;
-	relativize_paths(wk, tgt->depends, false, &depends_rel);
+	ca_relativize_paths(wk, tgt->depends, false, &depends_rel);
 
 	if (tgt->flags & custom_target_build_always_stale) {
 		obj_array_push(wk, depends_rel, make_str(wk, "build_always_stale"));
@@ -170,18 +179,18 @@ ninja_write_custom_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *c
 		rule = "CUSTOM_COMMAND";
 	}
 
-	fprintf(ctx->out, "build %s: %s %s | %s\n"
+	fprintf(ctx->out,
+		"build %s: %s %s | %s\n"
 		" COMMAND = %s\n",
 		get_cstr(wk, outputs),
 		rule,
 		get_cstr(wk, inputs),
 		get_cstr(wk, depends),
-		get_cstr(wk, cmdline)
-		);
+		get_cstr(wk, cmdline));
 
 	if (tgt->depfile) {
 		obj depfile_rel;
-		relativize_path(wk, tgt->depfile, false, &depfile_rel);
+		ca_relativize_path(wk, tgt->depfile, false, &depfile_rel);
 		fprintf(ctx->out, " DEPFILE = %s\n", get_cstr(wk, depfile_rel));
 	}
 
