@@ -5,14 +5,15 @@
 
 #include "compat.h"
 
-#include <assert.h>
 #include <inttypes.h>
 #include <string.h>
 
 #include "coerce.h"
 #include "functions/environment.h"
-#include "lang/interpreter.h"
+#include "lang/object_iterators.h"
+#include "lang/typecheck.h"
 #include "log.h"
+#include "platform/assert.h"
 #include "platform/filesystem.h"
 #include "platform/path.h"
 
@@ -49,13 +50,14 @@ coerce_environment_iter(struct workspace *wk, void *_ctx, obj val)
 
 	const struct str *ss = get_str(wk, val);
 	if (str_has_null(ss)) {
-		interp_error(wk, ctx->err_node, "environment string %o must not contain NUL", val);
+		vm_error_at(wk, ctx->err_node, "environment string %o must not contain NUL", val);
 		return ir_err;
 	}
 
 	const char *eql;
 	if (!(eql = strchr(ss->s, '='))) {
-		interp_error(wk, ctx->err_node, "invalid env element %o; env elements must be of the format key=value", val);
+		vm_error_at(
+			wk, ctx->err_node, "invalid env element %o; env elements must be of the format key=value", val);
 		return ir_err;
 	}
 
@@ -71,20 +73,19 @@ static enum iteration_result
 typecheck_environment_dict_iter(struct workspace *wk, void *_ctx, obj key, obj val)
 {
 	uint32_t err_node = *(uint32_t *)_ctx;
-	const struct str *k = get_str(wk, key),
-			 *v = get_str(wk, val);
+	const struct str *k = get_str(wk, key), *v = get_str(wk, val);
 
 	if (!k->len) {
-		interp_error(wk, err_node, "environment key may not be an empty string (value is '%s')", v->s);
+		vm_error_at(wk, err_node, "environment key may not be an empty string (value is '%s')", v->s);
 		return ir_err;
 	} else if (str_has_null(k)) {
-		interp_error(wk, err_node, "environment key may not contain NUL");
+		vm_error_at(wk, err_node, "environment key may not contain NUL");
 		return ir_err;
 	} else if (str_has_null(v)) {
-		interp_error(wk, err_node, "environment value may not contain NUL");
+		vm_error_at(wk, err_node, "environment value may not contain NUL");
 		return ir_err;
 	} else if (strchr(k->s, '=')) {
-		interp_error(wk, err_node, "environment key '%s' contains '='", k->s);
+		vm_error_at(wk, err_node, "environment key '%s' contains '='", k->s);
 		return ir_err;
 	}
 
@@ -103,19 +104,17 @@ coerce_key_value_dict(struct workspace *wk, uint32_t err_node, obj val, obj *res
 
 	enum obj_type t = get_obj_type(wk, val);
 	switch (t) {
-	case obj_string:
-		return coerce_environment_iter(wk, &ctx, val) != ir_err;
-	case obj_array:
-		return obj_array_foreach_flat(wk, val, &ctx, coerce_environment_iter);
+	case obj_string: return coerce_environment_iter(wk, &ctx, val) != ir_err;
+	case obj_array: return obj_array_foreach_flat(wk, val, &ctx, coerce_environment_iter);
 	case obj_dict:
-		if (!typecheck_dict(wk, err_node, val, obj_string)) {
+		if (!typecheck(wk, err_node, val, make_complex_type(wk, complex_type_nested, tc_dict, tc_string))) {
 			return false;
 		}
 
 		*res = val;
 		break;
 	default:
-		interp_error(wk, err_node, "unable to coerce type '%s' into key=value dict", obj_type_to_s(t));
+		vm_error_at(wk, err_node, "unable to coerce type '%s' into key=value dict", obj_type_to_s(t));
 		return false;
 	}
 
@@ -144,7 +143,7 @@ coerce_include_type(struct workspace *wk, const struct str *str, uint32_t err_no
 		}
 	}
 
-	interp_error(wk, err_node, "invalid value for include_type: %s", str->s);
+	vm_error_at(wk, err_node, "invalid value for include_type: %s", str->s);
 	return false;
 }
 
@@ -160,9 +159,7 @@ coerce_num_to_string(struct workspace *wk, uint32_t node, obj val, obj *res)
 		*res = val;
 		break;
 	}
-	default:
-		interp_error(wk, node, "unable to coerce %o to string", val);
-		return false;
+	default: vm_error_at(wk, node, "unable to coerce %o to string", val); return false;
 	}
 
 	return true;
@@ -179,9 +176,7 @@ coerce_string(struct workspace *wk, uint32_t node, obj val, obj *res)
 			*res = make_str(wk, "false");
 		}
 		break;
-	case obj_file:
-		*res = *get_obj_file(wk, val);
-		break;
+	case obj_file: *res = *get_obj_file(wk, val); break;
 	case obj_number: {
 		*res = make_strf(wk, "%" PRId64, get_obj_number(wk, val));
 		break;
@@ -190,9 +185,61 @@ coerce_string(struct workspace *wk, uint32_t node, obj val, obj *res)
 		*res = val;
 		break;
 	}
-	default:
-		interp_error(wk, node, "unable to coerce %o to string", val);
-		return false;
+	case obj_feature_opt: {
+		const char *s = 0;
+		switch (get_obj_feature_opt(wk, val)) {
+		case feature_opt_auto: s = "auto"; break;
+		case feature_opt_enabled: s = "enabled"; break;
+		case feature_opt_disabled: s = "disabled"; break;
+		}
+
+		*res = make_strf(wk, "<option %s>", s);
+		break;
+	}
+	case obj_array: {
+		obj strs, v, s;
+		make_obj(wk, &strs, obj_array);
+		obj_array_for(wk, val, v) {
+			if (!coerce_string(wk, node, v, &s)) {
+				return false;
+			}
+
+			obj_array_push(wk, strs, s);
+		}
+
+		obj joined;
+		obj_array_join(wk, false, strs, make_str(wk, ", "), &joined);
+
+		*res = make_str(wk, "[");
+		str_apps(wk, res, joined);
+		str_app(wk, res, "]");
+		break;
+	}
+	case obj_dict: {
+		obj strs, k, v, s;
+		make_obj(wk, &strs, obj_array);
+		obj_dict_for(wk, val, k, v) {
+			if (!coerce_string(wk, node, v, &s)) {
+				return false;
+			}
+
+			obj kv = make_str(wk, "'");
+			str_apps(wk, &kv, k);
+			str_app(wk, &kv, "': ");
+			str_apps(wk, &kv, s);
+
+			obj_array_push(wk, strs, kv);
+		}
+
+		obj joined;
+		obj_array_join(wk, false, strs, make_str(wk, ", "), &joined);
+
+		*res = make_str(wk, "{");
+		str_apps(wk, res, joined);
+		str_app(wk, res, "}");
+		break;
+	}
+	default: vm_error_at(wk, node, "unable to coerce %o to string", val); return false;
 	}
 
 	return true;
@@ -230,17 +277,15 @@ coerce_string_array(struct workspace *wk, uint32_t node, obj arr, obj *res)
 }
 
 bool
-coerce_executable(struct workspace *wk, uint32_t node, obj val, obj *res)
+coerce_executable(struct workspace *wk, uint32_t node, obj val, obj *res, obj *args)
 {
 	obj str;
+	*args = 0;
 
 	enum obj_type t = get_obj_type(wk, val);
 	switch (t) {
-	case obj_file:
-		str = *get_obj_file(wk, val);
-		break;
-	case obj_both_libs:
-		val = get_obj_both_libs(wk, val)->dynamic_lib;
+	case obj_file: str = *get_obj_file(wk, val); break;
+	case obj_both_libs: val = get_obj_both_libs(wk, val)->dynamic_lib;
 	/* fallthrough */
 	case obj_build_target: {
 		struct obj_build_target *o = get_obj_build_target(wk, val);
@@ -252,19 +297,44 @@ coerce_executable(struct workspace *wk, uint32_t node, obj val, obj *res)
 		str = sbuf_into_str(wk, &dest);
 		break;
 	}
+	case obj_python_installation:
+		val = get_obj_python_installation(wk, val)->prog;
+		/* fallthrough */
 	case obj_external_program: {
 		struct obj_external_program *o = get_obj_external_program(wk, val);
 		if (!o->found) {
-			interp_error(wk, node, "a not found external_program cannot be used here");
+			vm_error_at(wk, node, "a not found external_program cannot be used here");
 			return ir_err;
 		}
 
-		str = o->full_path;
+		obj_array_index(wk, o->cmd_array, 0, &str);
+		uint32_t cmd_array_len = get_obj_array(wk, o->cmd_array)->len;
+		if (cmd_array_len > 1) {
+			*args = obj_array_slice(wk, o->cmd_array, 1, cmd_array_len);
+		}
 		break;
 	}
-	default:
-		interp_error(wk, node, "unable to coerce '%s' into executable", obj_type_to_s(t));
-		return false;
+	case obj_custom_target: {
+		struct obj_custom_target *o = get_obj_custom_target(wk, val);
+
+		uint32_t i = 0;
+		obj v;
+		obj_array_for(wk, o->output, v) {
+			if (i == 0) {
+				str = *get_obj_file(wk, v);
+			} else {
+				if (!args) {
+					make_obj(wk, args, obj_array);
+				}
+
+				obj_array_push(wk, *args, *get_obj_file(wk, v));
+			}
+
+			++i;
+		}
+		break;
+	}
+	default: vm_error_at(wk, node, "unable to coerce '%s' into executable", obj_type_to_s(t)); return false;
 	}
 
 	*res = str;
@@ -285,22 +355,17 @@ coerce_requirement(struct workspace *wk, struct args_kw *kw_required, enum requi
 			}
 		} else if (t == obj_feature_opt) {
 			switch (get_obj_feature_opt(wk, kw_required->val)) {
-			case feature_opt_disabled:
-				*requirement = requirement_skip;
-				break;
-			case feature_opt_enabled:
-				*requirement = requirement_required;
-				break;
-			case feature_opt_auto:
-				*requirement = requirement_auto;
-				break;
+			case feature_opt_disabled: *requirement = requirement_skip; break;
+			case feature_opt_enabled: *requirement = requirement_required; break;
+			case feature_opt_auto: *requirement = requirement_auto; break;
 			}
 		} else {
-			interp_error(wk, kw_required->node, "expected type %s or %s, got %s",
+			vm_error_at(wk,
+				kw_required->node,
+				"expected type %s or %s, got %s",
 				obj_type_to_s(obj_bool),
 				obj_type_to_s(obj_feature_opt),
-				obj_type_to_s(t)
-				);
+				obj_type_to_s(t));
 			return false;
 		}
 	} else {
@@ -365,18 +430,19 @@ coerce_into_file(struct workspace *wk, struct coerce_into_files_ctx *ctx, obj va
 
 		switch (ctx->mode) {
 		case mode_input:
-			if (!coerce_string_to_file(wk, get_cstr(wk, current_project(wk)->cwd), val, file)) {
+			if (!coerce_string_to_file(wk, workspace_cwd(wk), val, file)) {
 				return ir_err;
 			}
 
 			if (!ctx->exists(get_file_path(wk, *file))) {
-				interp_error(wk, ctx->node, "%s %o does not exist", ctx->type, val);
+				vm_error_at(wk, ctx->node, "%s %o does not exist", ctx->type, val);
 				return ir_err;
 			}
 			break;
 		case mode_output:
-			if (!path_is_basename(get_cstr(wk, val))) {
-				interp_error(wk, ctx->node, "output file '%s' contains path separators", get_cstr(wk, val));
+			if (!path_is_basename(get_cstr(wk, val)) && wk->vm.lang_mode != language_extended) {
+				vm_error_at(
+					wk, ctx->node, "output file '%s' contains path separators", get_cstr(wk, val));
 				return ir_err;
 			}
 
@@ -384,14 +450,11 @@ coerce_into_file(struct workspace *wk, struct coerce_into_files_ctx *ctx, obj va
 			make_obj(wk, file, obj_file);
 			*get_obj_file(wk, *file) = sbuf_into_str(wk, &buf);
 			break;
-		default:
-			assert(false);
-			return ir_err;
+		default: assert(false); return ir_err;
 		}
 		break;
 	}
-	case obj_both_libs:
-		val = get_obj_both_libs(wk, val)->dynamic_lib;
+	case obj_both_libs: val = get_obj_both_libs(wk, val)->dynamic_lib;
 	/* fallthrough */
 	case obj_build_target: {
 		if (ctx->mode == mode_output) {
@@ -415,8 +478,7 @@ coerce_into_file(struct workspace *wk, struct coerce_into_files_ctx *ctx, obj va
 		break;
 	default:
 type_error:
-		interp_error(wk, ctx->node, "unable to coerce object with type %s into %s",
-			obj_type_to_s(t), ctx->type);
+		vm_error_at(wk, ctx->node, "unable to coerce object with type %s into %s", obj_type_to_s(t), ctx->type);
 		return false;
 	}
 
@@ -435,8 +497,8 @@ coerce_into_files_iter(struct workspace *wk, void *_ctx, obj val)
 			goto type_error;
 		}
 
-		if (!obj_array_foreach(wk, get_obj_custom_target(wk, val)->output,
-			ctx, coerce_custom_target_output_iter)) {
+		if (!obj_array_foreach(
+			    wk, get_obj_custom_target(wk, val)->output, ctx, coerce_custom_target_output_iter)) {
 			return ir_err;
 		}
 		break;
@@ -455,8 +517,7 @@ coerce_into_files_iter(struct workspace *wk, void *_ctx, obj val)
 	}
 	default:
 type_error:
-		interp_error(wk, ctx->node, "unable to coerce object with type %s into %s",
-			obj_type_to_s(t), ctx->type);
+		vm_error_at(wk, ctx->node, "unable to coerce object with type %s into %s", obj_type_to_s(t), ctx->type);
 		return ir_err;
 	}
 
@@ -464,8 +525,13 @@ type_error:
 }
 
 static bool
-_coerce_files(struct workspace *wk, uint32_t node, obj val, obj *res,
-	const char *type_name, exists_func exists, enum coerce_into_files_mode mode,
+_coerce_files(struct workspace *wk,
+	uint32_t node,
+	obj val,
+	obj *res,
+	const char *type_name,
+	exists_func exists,
+	enum coerce_into_files_mode mode,
 	const char *output_dir)
 {
 	make_obj(wk, res, obj_array);
@@ -480,14 +546,11 @@ _coerce_files(struct workspace *wk, uint32_t node, obj val, obj *res,
 	};
 
 	switch (get_obj_type(wk, val)) {
-	case obj_array:
-		return obj_array_foreach_flat(wk, val, &ctx, coerce_into_files_iter);
+	case obj_array: return obj_array_foreach_flat(wk, val, &ctx, coerce_into_files_iter);
 	default:
 		switch (coerce_into_files_iter(wk, &ctx, val)) {
-		case ir_err:
-			return false;
-		default:
-			return true;
+		case ir_err: return false;
+		default: return true;
 		}
 	}
 }
@@ -540,7 +603,7 @@ include_directories_iter(struct workspace *wk, void *_ctx, obj v)
 		obj_array_push(wk, ctx->res, v);
 		return ir_cont;
 	} else if (t != obj_string) {
-		interp_error(wk, ctx->node, "unable to coerce %o to include_directory", v);
+		vm_error_at(wk, ctx->node, "unable to coerce %o to include_directory", v);
 		return ir_err;
 	}
 
@@ -551,14 +614,14 @@ include_directories_iter(struct workspace *wk, void *_ctx, obj v)
 
 	if (!path_is_absolute(p)) {
 		SBUF(abs);
-		path_join(wk, &abs, get_cstr(wk, current_project(wk)->cwd), p);
+		path_join(wk, &abs, workspace_cwd(wk), p);
 		path = sbuf_into_str(wk, &abs);
 	}
 
 	p = get_cstr(wk, path);
 
 	if (!fs_dir_exists(p)) {
-		interp_error(wk, ctx->node, "directory '%s' does not exist", get_cstr(wk, path));
+		vm_error_at(wk, ctx->node, "directory '%s' does not exist", get_cstr(wk, path));
 		return ir_err;
 	}
 
@@ -601,3 +664,17 @@ coerce_include_dirs(struct workspace *wk, uint32_t node, obj val, bool is_system
 	*res = ctx.res;
 	return true;
 }
+
+enum machine_kind
+coerce_machine_kind(struct workspace *wk, struct args_kw *native_kw)
+{
+	if (native_kw && native_kw->set) {
+		if (get_obj_bool(wk, native_kw->val)) {
+			return machine_kind_build;
+		} else {
+			return machine_kind_host;
+		}
+	}
+
+	return machine_kind_host;
+}
