@@ -6,12 +6,10 @@
 #include "compat.h"
 
 #include <string.h>
-#include <strings.h>
 
 #include "formats/lines.h"
 #include "formats/tap.h"
 #include "lang/string.h"
-#include "log.h"
 
 struct tap_parse_ctx {
 	struct tap_parse_result *res;
@@ -20,16 +18,16 @@ struct tap_parse_ctx {
 };
 
 static enum iteration_result
-tap_parse_line_cb(void *_ctx, char *line, size_t _)
+tap_parse_line_cb(void *_ctx, const char *line, size_t len)
 {
 	struct tap_parse_ctx *ctx = _ctx;
-	struct str l = WKSTR(line), rest;
+	struct str l = { .s = line, .len = len }, rest;
 	bool ok;
 
 	if (str_startswith(&l, &WKSTR("1..")) && l.len > 3) {
 		struct str i = { .s = &l.s[3], l.len - 3 };
 		int64_t plan_count;
-		if (str_to_i(&i, &plan_count) && plan_count > 0) {
+		if (str_to_i(&i, &plan_count, false) && plan_count > 0) {
 			ctx->have_plan = true;
 			ctx->res->total = plan_count;
 		}
@@ -40,22 +38,27 @@ tap_parse_line_cb(void *_ctx, char *line, size_t _)
 		return ir_cont;
 	} else if (str_startswith(&l, &WKSTR("ok"))) {
 		ok = true;
-		rest = (struct str) { .s = &l.s[2], .len = l.len - 2 };
+		rest = (struct str){ .s = &l.s[2], .len = l.len - 2 };
 	} else if (str_startswith(&l, &WKSTR("not ok"))) {
 		ok = false;
-		rest = (struct str) { .s = &l.s[6], .len = l.len - 6 };
+		rest = (struct str){ .s = &l.s[6], .len = l.len - 6 };
 	} else {
 		return ir_cont;
 	}
 
-	enum { none, todo, skip, } directive = none;
+	enum {
+		none,
+		todo,
+		skip,
+	} directive
+		= none;
 	{
 		char *directive_str;
 		if ((directive_str = strstr(rest.s, " # "))) {
 			directive_str += 3;
-			if (strncasecmp(directive_str, "todo", 4) == 0) {
+			if (str_startswithi(&WKSTR(directive_str), &WKSTR("todo"))) {
 				directive = todo;
-			} else if (strncasecmp(directive_str, "skip", 4) == 0) {
+			} else if (str_startswithi(&WKSTR(directive_str), &WKSTR("skip"))) {
 				directive = skip;
 			}
 		}
@@ -80,11 +83,13 @@ tap_parse_line_cb(void *_ctx, char *line, size_t _)
 }
 
 void
-tap_parse(char *buf, uint64_t buf_len, struct tap_parse_result *res)
+tap_parse(const char *buf, uint64_t buf_len, struct tap_parse_result *res)
 {
 	struct tap_parse_ctx ctx = { .res = res };
 
-	each_line(buf, buf_len, &ctx, tap_parse_line_cb);
+	each_line_const(buf, buf_len, &ctx, tap_parse_line_cb);
+
+	res->have_plan = ctx.have_plan;
 
 	if (!ctx.have_plan) {
 		res->total = res->pass + res->skip + res->fail;
