@@ -8,7 +8,6 @@
 #include "lang/func_lookup.h"
 #include "functions/disabler.h"
 #include "lang/typecheck.h"
-#include "log.h"
 
 static bool
 func_disabler_found(struct workspace *wk, obj self, obj *res)
@@ -17,8 +16,7 @@ func_disabler_found(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	make_obj(wk, res, obj_bool);
-	set_obj_bool(wk, *res, false);
+	*res = make_obj_bool(wk, false);
 	return true;
 }
 
