@@ -7,20 +7,20 @@
 
 #include <string.h>
 
-#include "lang/func_lookup.h"
+#include "buf_size.h"
+#include "functions/compiler.h"
 #include "functions/file.h"
+#include "lang/func_lookup.h"
 #include "lang/typecheck.h"
-#include "log.h"
 
-bool
-file_is_linkable(struct workspace *wk, obj file)
+static bool
+file_ends_with_suffix(struct workspace *wk, obj file, const char *suffixes[], uint32_t len)
 {
 	const struct str *s = get_str(wk, *get_obj_file(wk, file));
 
-	const char *suffs[] = { ".a", ".dll", ".lib", ".so", ".dylib", NULL };
 	uint32_t i;
-	for (i = 0; suffs[i]; ++i) {
-		if (str_endswith(s, &WKSTR(suffs[i]))) {
+	for (i = 0; i < len; ++i) {
+		if (str_endswith(s, &WKSTR(suffixes[i]))) {
 			return true;
 		}
 	}
@@ -28,6 +28,26 @@ file_is_linkable(struct workspace *wk, obj file)
 	return false;
 }
 
+bool
+file_is_dynamic_lib(struct workspace *wk, obj file)
+{
+	const char *exts[] = { COMPILER_DYNAMIC_LIB_EXTS };
+	return file_ends_with_suffix(wk, file, exts, ARRAY_LEN(exts));
+}
+
+bool
+file_is_static_lib(struct workspace *wk, obj file)
+{
+	const char *exts[] = { COMPILER_STATIC_LIB_EXTS };
+	return file_ends_with_suffix(wk, file, exts, ARRAY_LEN(exts));
+}
+
+bool
+file_is_linkable(struct workspace *wk, obj file)
+{
+	return  file_is_static_lib(wk, file) || file_is_dynamic_lib(wk, file);
+}
+
 static bool
 func_file_full_path(struct workspace *wk, obj self, obj *res)
 {
