@@ -7,22 +7,21 @@
 
 #include "compat.h"
 
-#include <assert.h>
 #include <string.h>
 
 #include "buf_size.h"
 #include "coerce.h"
-#include "error.h"
 #include "functions/build_target.h"
 #include "functions/file.h"
 #include "functions/generator.h"
 #include "functions/kernel/build_target.h"
 #include "functions/kernel/dependency.h"
-#include "functions/machine.h"
 #include "install.h"
-#include "lang/interpreter.h"
+#include "lang/typecheck.h"
 #include "log.h"
+#include "machines.h"
 #include "options.h"
+#include "platform/assert.h"
 #include "platform/filesystem.h"
 #include "platform/path.h"
 
@@ -39,7 +38,7 @@ enum build_target_kwargs {
 	bt_kw_link_whole,
 	bt_kw_version,
 	bt_kw_build_by_default,
-	bt_kw_extra_files, // TODO
+	bt_kw_extra_files,
 	bt_kw_target_type,
 	bt_kw_name_prefix,
 	bt_kw_name_suffix,
@@ -53,25 +52,21 @@ enum build_target_kwargs {
 	bt_kw_export_dynamic,
 	bt_kw_vs_module_defs, // TODO
 	bt_kw_gnu_symbol_visibility,
-	bt_kw_native, // TODO
+	bt_kw_native,
 	bt_kw_darwin_versions, // TODO
 	bt_kw_implib, // TODO
 	bt_kw_gui_app, // TODO
 	bt_kw_link_language, // TODO
 	bt_kw_win_subsystem, // TODO
 	bt_kw_override_options,
-
-	/* lang args */
-	bt_kw_c_pch, // TODO
-	bt_kw_cpp_pch, // TODO
-	bt_kw_c_args,
-	bt_kw_cpp_args,
-	bt_kw_objc_args,
-	bt_kw_masm_args, // TODO
-	bt_kw_nasm_args, // TODO
 	bt_kw_link_args,
 
-	bt_kwargs_count,
+#define E(lang, s) bt_kw_##lang##s
+#define TOOLCHAIN_ENUM(lang) E(lang, _args), E(lang, _static_args), E(lang, _shared_args), E(lang, _pch),
+	FOREACH_COMPILER_EXPOSED_LANGUAGE(TOOLCHAIN_ENUM)
+#undef TOOLCHAIN_ENUM
+#undef E
+		bt_kwargs_count,
 };
 
 static enum iteration_result
@@ -98,16 +93,26 @@ determine_linker_from_objects_iter(struct workspace *wk, void *_ctx, obj val)
 
 	enum compiler_language fl;
 
+	/*
+	 * Try to see if the file looks like
+	 * path/to/object.language.object_extension
+	 *
+	 * This means we expect two extensions, the first one will be stripped,
+	 * and then the second will be used to determine the language of the
+	 * file.
+	 */
+
 	const struct str *o = get_str(wk, *get_obj_file(wk, val));
-	if (!str_endswith(o, &WKSTR(".o"))) {
+	SBUF(path);
+	path_basename(wk, &path, o->s);
+	const char *first_dot = strrchr(path.buf, '.');
+	path.len = first_dot - path.buf;
+	path.buf[path.len] = 0;
+
+	if (!strrchr(path.buf, '.')) {
 		return ir_cont;
 	}
 
-	SBUF(path);
-	sbuf_pushs(wk, &path, o->s);
-	path.len -= 2;
-	path.buf[path.len] = 0;
-
 	if (!filename_to_compiler_language(path.buf, &fl)) {
 		/* LOG_E("unable to determine language for '%s'", get_cstr(wk, src->dat.file)); */
 		return ir_cont;
@@ -138,7 +143,7 @@ build_tgt_determine_linker(struct workspace *wk, uint32_t err_node, struct obj_b
 		obj comp;
 		uint32_t i;
 		for (i = 0; i < ARRAY_LEN(clink_langs); ++i) {
-			if (obj_dict_geti(wk, current_project(wk)->compilers, clink_langs[i], &comp)) {
+			if (obj_dict_geti(wk, current_project(wk)->toolchains[tgt->machine], clink_langs[i], &comp)) {
 				tgt->dep_internal.link_language = clink_langs[i];
 				break;
 			}
@@ -146,7 +151,7 @@ build_tgt_determine_linker(struct workspace *wk, uint32_t err_node, struct obj_b
 	}
 
 	if (!tgt->dep_internal.link_language) {
-		interp_error(wk, err_node, "unable to determine linker for target");
+		vm_error_at(wk, err_node, "unable to determine linker for target");
 		return false;
 	}
 
@@ -176,6 +181,11 @@ process_source_include(struct workspace *wk, struct process_build_tgt_sources_ct
 	struct obj_build_target *tgt = get_obj_build_target(wk, ctx->tgt_id);
 	obj_array_push(wk, tgt->dep_internal.order_deps, sbuf_into_str(wk, &path));
 
+	if (!tgt->dep_internal.raw.order_deps) {
+		make_obj(wk, &tgt->dep_internal.raw.order_deps, obj_array);
+	}
+	obj_array_push(wk, tgt->dep_internal.raw.order_deps, val);
+
 	if (!ctx->implicit_include_directories) {
 		return true;
 	}
@@ -203,8 +213,9 @@ build_tgt_push_source_files_iter(struct workspace *wk, void *_ctx, obj val)
 	}
 
 	enum compiler_language lang;
-	if (!filename_to_compiler_language(get_file_path(wk, val), &lang)
-	    || languages[lang].is_header) {
+	if (!filename_to_compiler_language(get_file_path(wk, val), &lang) || languages[lang].is_header) {
+		obj_array_push(wk, tgt->extra_files, val);
+
 		// process every file that is either a header, or isn't
 		// recognized, as a header
 		if (!process_source_include(wk, ctx, val)) {
@@ -255,13 +266,34 @@ static bool
 type_from_kw(struct workspace *wk, uint32_t node, obj t, enum tgt_type *res)
 {
 	const char *tgt_type = get_cstr(wk, t);
-	struct { char *name; enum tgt_type type; } tgt_tbl[] = {
-		{ "executable", tgt_executable, },
-		{ "shared_library", tgt_dynamic_library, },
-		{ "shared_module", tgt_shared_module, },
-		{ "static_library", tgt_static_library, },
-		{ "both_libraries", tgt_dynamic_library | tgt_static_library, },
-		{ "library", get_option_default_library(wk), },
+	struct {
+		char *name;
+		enum tgt_type type;
+	} tgt_tbl[] = {
+		{
+			"executable",
+			tgt_executable,
+		},
+		{
+			"shared_library",
+			tgt_dynamic_library,
+		},
+		{
+			"shared_module",
+			tgt_shared_module,
+		},
+		{
+			"static_library",
+			tgt_static_library,
+		},
+		{
+			"both_libraries",
+			tgt_dynamic_library | tgt_static_library,
+		},
+		{
+			"library",
+			get_option_default_library(wk),
+		},
 		{ 0 },
 	};
 	uint32_t i;
@@ -274,7 +306,7 @@ type_from_kw(struct workspace *wk, uint32_t node, obj t, enum tgt_type *res)
 	}
 
 	if (!tgt_tbl[i].name) {
-		interp_error(wk, node, "unsupported target type '%s'", tgt_type);
+		vm_error_at(wk, node, "unsupported target type '%s'", tgt_type);
 		return false;
 	}
 
@@ -307,7 +339,6 @@ setup_dllname(struct workspace *wk, struct obj_build_target *tgt, const char *pl
 {
 	if (dllver) {
 		tgt->soname = make_strf(wk, "%s-%s.dll", plain_name, get_cstr(wk, dllver));
-		return;
 	} else if (ver) {
 		char buf[BUF_SIZE_1k];
 
@@ -316,16 +347,28 @@ setup_dllname(struct workspace *wk, struct obj_build_target *tgt, const char *pl
 		if (tmp) {
 			*tmp = '\0';
 			tgt->soname = make_strf(wk, "%s-%s.dll", plain_name, buf);
-			return;
 		}
 	}
 
-	tgt->soname = make_strf(wk, "%s.dll", plain_name);
+	if (!tgt->soname) {
+		tgt->soname = make_strf(wk, "%s.dll", plain_name);
+	}
+
+	if (tgt->type == tgt_dynamic_library) {
+		SBUF(implib);
+		sbuf_pushf(wk, &implib, "%s-implib.lib", plain_name);
+		SBUF(path);
+		path_join(wk, &path, get_cstr(wk, tgt->build_dir), implib.buf);
+		tgt->implib = sbuf_into_str(wk, &path);
+	}
 }
 
 static bool
-setup_shared_object_symlinks(struct workspace *wk, struct obj_build_target *tgt,
-	const char *plain_name, obj *plain_name_install, obj *soname_install)
+setup_shared_object_symlinks(struct workspace *wk,
+	struct obj_build_target *tgt,
+	const char *plain_name,
+	obj *plain_name_install,
+	obj *soname_install)
 {
 	SBUF(soname_symlink);
 	SBUF(plain_name_symlink);
@@ -344,7 +387,8 @@ setup_shared_object_symlinks(struct workspace *wk, struct obj_build_target *tgt,
 		*soname_install = sbuf_into_str(wk, &soname_symlink);
 	}
 
-	if (!str_eql(get_str(wk, tgt->soname), &WKSTR(plain_name))) {
+	if (!str_eql(&WKSTR(plain_name), get_str(wk, tgt->soname))
+		&& !str_eql(&WKSTR(plain_name), get_str(wk, tgt->build_name))) {
 		path_join(wk, &plain_name_symlink, get_cstr(wk, tgt->build_dir), plain_name);
 
 		if (!fs_make_symlink(get_cstr(wk, tgt->soname), plain_name_symlink.buf, true)) {
@@ -358,26 +402,30 @@ setup_shared_object_symlinks(struct workspace *wk, struct obj_build_target *tgt,
 }
 
 static bool
-determine_target_build_name(struct workspace *wk, struct obj_build_target *tgt, obj sover, obj ver,
-	obj name_pre, obj name_suff, char plain_name[BUF_SIZE_2k])
+determine_target_build_name(struct workspace *wk,
+	struct obj_build_target *tgt,
+	obj sover,
+	obj ver,
+	obj name_pre,
+	obj name_suff,
+	char plain_name[BUF_SIZE_2k])
 {
 	char ver_dll[BUF_SIZE_1k];
 	const char *pref, *suff, *ver_suff = NULL;
-	enum machine_system sys = machine_system();
 
 	*ver_dll = '\0';
 
 	switch (tgt->type) {
 	case tgt_executable:
 		pref = "";
-		if (sys == machine_system_windows || sys == machine_system_cygwin) {
+		if (host_machine.is_windows) {
 			suff = "exe";
 		} else {
 			suff = NULL;
 		}
 		break;
 	case tgt_static_library:
-		if (sys == machine_system_cygwin) {
+		if (host_machine.sys == machine_system_cygwin) {
 			pref = "cyg";
 		} else {
 			pref = "lib";
@@ -386,12 +434,15 @@ determine_target_build_name(struct workspace *wk, struct obj_build_target *tgt,
 		break;
 	case tgt_shared_module:
 	case tgt_dynamic_library:
-		if (sys == machine_system_cygwin) {
+		if (host_machine.sys == machine_system_cygwin) {
 			pref = "cyg";
+		} else if (host_machine.sys == machine_system_windows) {
+			pref = "";
 		} else {
 			pref = "lib";
 		}
-		if (sys == machine_system_windows || sys == machine_system_cygwin) {
+
+		if (host_machine.is_windows) {
 			suff = "dll";
 			if (sover) {
 				strncpy(ver_dll, get_cstr(wk, sover), sizeof(ver_dll) - 1);
@@ -399,12 +450,12 @@ determine_target_build_name(struct workspace *wk, struct obj_build_target *tgt,
 				strncpy(ver_dll, get_cstr(wk, ver), sizeof(ver_dll) - 1);
 				char *ver_tmp = strchr(ver_dll, '.');
 				if (ver_tmp) {
-					*ver_tmp = '\0';
+					*ver_tmp = 0;
 				} else {
-					*ver_dll = '\0';
+					*ver_dll = 0;
 				}
 			}
-		} else if (sys == machine_system_darwin) {
+		} else if (host_machine.sys == machine_system_darwin) {
 			suff = "dylib";
 		} else {
 			suff = "so";
@@ -415,9 +466,7 @@ determine_target_build_name(struct workspace *wk, struct obj_build_target *tgt,
 			}
 		}
 		break;
-	default:
-		assert(false && "unreachable");
-		return false;
+	default: assert(false && "unreachable"); return false;
 	}
 
 	if (name_pre) {
@@ -428,11 +477,23 @@ determine_target_build_name(struct workspace *wk, struct obj_build_target *tgt,
 		suff = get_cstr(wk, name_suff);
 	}
 
-	if (sys == machine_system_windows || sys == machine_system_cygwin) {
+	if (host_machine.is_windows) {
 		snprintf(plain_name, BUF_SIZE_2k, "%s%s", pref, get_cstr(wk, tgt->name));
-		tgt->build_name = make_strf(wk, "%s%s%s%s%s", plain_name, *ver_dll ? "-" : "", *ver_dll ? ver_dll : "", suff ? "." : "", suff ? suff : "");
+		tgt->build_name = make_strf(wk,
+			"%s%s%s%s%s",
+			plain_name,
+			*ver_dll ? "-" : "",
+			*ver_dll ? ver_dll : "",
+			suff ? "." : "",
+			suff ? suff : "");
 	} else {
-		snprintf(plain_name, BUF_SIZE_2k, "%s%s%s%s", pref, get_cstr(wk, tgt->name), suff ? "." : "", suff ? suff : "");
+		snprintf(plain_name,
+			BUF_SIZE_2k,
+			"%s%s%s%s",
+			pref,
+			get_cstr(wk, tgt->name),
+			suff ? "." : "",
+			suff ? suff : "");
 		tgt->build_name = make_strf(wk, "%s%s%s", plain_name, ver_suff ? "." : "", ver_suff ? ver_suff : "");
 	}
 
@@ -440,8 +501,12 @@ determine_target_build_name(struct workspace *wk, struct obj_build_target *tgt,
 }
 
 static bool
-create_target(struct workspace *wk, struct args_norm *an, struct args_kw *akw,
-	enum tgt_type type, bool ignore_sources, obj *res)
+create_target(struct workspace *wk,
+	struct args_norm *an,
+	struct args_kw *akw,
+	enum tgt_type type,
+	bool ignore_sources,
+	obj *res)
 {
 	char plain_name[BUF_SIZE_2k + 1] = { 0 };
 	make_obj(wk, res, obj_build_target);
@@ -450,22 +515,25 @@ create_target(struct workspace *wk, struct args_norm *an, struct args_kw *akw,
 	tgt->name = an[0].val;
 	tgt->cwd = current_project(wk)->cwd;
 	tgt->build_dir = current_project(wk)->build_dir;
+	tgt->machine = coerce_machine_kind(wk, &akw[bt_kw_native]);
+	tgt->callstack = vm_callstack(wk);
 	make_obj(wk, &tgt->args, obj_dict);
 	make_obj(wk, &tgt->src, obj_array);
 	make_obj(wk, &tgt->required_compilers, obj_dict);
+	make_obj(wk, &tgt->extra_files, obj_array);
 	build_dep_init(wk, &tgt->dep_internal);
 
 	{ // linker args (process before dependencies so link_with libs come first on link line
 		if (akw[bt_kw_link_with].set) {
-			if (!dep_process_link_with(wk, akw[bt_kw_link_with].node,
-				akw[bt_kw_link_with].val, &tgt->dep_internal)) {
+			if (!dep_process_link_with(
+				    wk, akw[bt_kw_link_with].node, akw[bt_kw_link_with].val, &tgt->dep_internal)) {
 				return false;
 			}
 		}
 
 		if (akw[bt_kw_link_whole].set) {
-			if (!dep_process_link_whole(wk, akw[bt_kw_link_whole].node,
-				akw[bt_kw_link_whole].val, &tgt->dep_internal)) {
+			if (!dep_process_link_whole(
+				    wk, akw[bt_kw_link_whole].node, akw[bt_kw_link_whole].val, &tgt->dep_internal)) {
 				return false;
 			}
 		}
@@ -475,11 +543,20 @@ create_target(struct workspace *wk, struct args_norm *an, struct args_kw *akw,
 		dep_process_deps(wk, akw[bt_kw_dependencies].val, &tgt->dep_internal);
 	}
 
+	if (!deps_check_machine_matches(wk,
+		    tgt->name,
+		    tgt->machine,
+		    akw[bt_kw_link_with].val,
+		    akw[bt_kw_link_whole].val,
+		    akw[bt_kw_dependencies].val)) {
+		return false;
+	}
+
 	if (akw[bt_kw_override_options].set) { // override options
 		if (!parse_and_set_override_options(wk,
-			akw[bt_kw_override_options].node,
-			akw[bt_kw_override_options].val,
-			&tgt->override_options)) {
+			    akw[bt_kw_override_options].node,
+			    akw[bt_kw_override_options].val,
+			    &tgt->override_options)) {
 			return false;
 		}
 	}
@@ -491,7 +568,8 @@ create_target(struct workspace *wk, struct args_norm *an, struct args_kw *akw,
 				pic = get_obj_bool(wk, akw[bt_kw_pic].val);
 
 				if (!pic && tgt->type & (tgt_dynamic_library | tgt_shared_module)) {
-					interp_error(wk, akw[bt_kw_pic].node, "shared libraries must be compiled as pic");
+					vm_error_at(
+						wk, akw[bt_kw_pic].node, "shared libraries must be compiled as pic");
 					return false;
 				}
 			} else {
@@ -516,7 +594,7 @@ create_target(struct workspace *wk, struct args_norm *an, struct args_kw *akw,
 				pie = get_obj_bool(wk, akw[bt_kw_pie].set);
 
 				if (pie && (tgt->type & tgt_executable) != tgt_executable) {
-					interp_error(wk, akw[bt_kw_pie].node, "pie cannot be set for non-executables");
+					vm_error_at(wk, akw[bt_kw_pie].node, "pie cannot be set for non-executables");
 					return false;
 				}
 			} else if ((tgt->type & tgt_executable) == tgt_executable) {
@@ -550,7 +628,7 @@ create_target(struct workspace *wk, struct args_norm *an, struct args_kw *akw,
 			} else if (str_eql(str, &WKSTR("inlineshidden"))) {
 				tgt->visibility = compiler_visibility_inlineshidden;
 			} else {
-				interp_error(wk, vis->node, "unknown visibility '%s'", get_cstr(wk, vis->val));
+				vm_error_at(wk, vis->node, "unknown visibility '%s'", get_cstr(wk, vis->val));
 				return false;
 			}
 
@@ -565,8 +643,13 @@ create_target(struct workspace *wk, struct args_norm *an, struct args_kw *akw,
 		}
 	}
 
-	if (!determine_target_build_name(wk, tgt, sover, akw[bt_kw_version].val,
-		akw[bt_kw_name_prefix].val, akw[bt_kw_name_suffix].val, plain_name)) {
+	if (!determine_target_build_name(wk,
+		    tgt,
+		    sover,
+		    akw[bt_kw_version].val,
+		    akw[bt_kw_name_prefix].val,
+		    akw[bt_kw_name_suffix].val,
+		    plain_name)) {
 		return false;
 	}
 
@@ -581,10 +664,9 @@ create_target(struct workspace *wk, struct args_norm *an, struct args_kw *akw,
 		tgt->private_path = sbuf_into_str(wk, &path);
 	}
 
-	bool implicit_include_directories =
-		akw[bt_kw_implicit_include_directories].set
-				? get_obj_bool(wk, akw[bt_kw_implicit_include_directories].val)
-				: true;
+	bool implicit_include_directories = akw[bt_kw_implicit_include_directories].set ?
+						    get_obj_bool(wk, akw[bt_kw_implicit_include_directories].val) :
+						    true;
 
 	{ // sources
 		if (akw[bt_kw_objects].set) {
@@ -599,6 +681,9 @@ create_target(struct workspace *wk, struct args_norm *an, struct args_kw *akw,
 			make_obj(wk, &tgt->objects, obj_array);
 		}
 
+		obj_array_extend(wk, tgt->objects, tgt->dep_internal.objects);
+		obj_array_dedup_in_place(wk, &tgt->objects);
+
 		if (!ignore_sources) {
 			obj sources = an[1].val;
 
@@ -624,15 +709,17 @@ create_target(struct workspace *wk, struct args_norm *an, struct args_kw *akw,
 			tgt->src = deduped;
 		}
 
-		if (!get_obj_array(wk, tgt->src)->len &&
-		    !get_obj_array(wk, tgt->objects)->len &&
-		    !akw[bt_kw_link_whole].set
-		    && tgt->type != tgt_static_library) {
-			uint32_t node = akw[bt_kw_sources].set? akw[bt_kw_sources].node : an[1].node;
+		if (!get_obj_array(wk, tgt->src)->len && !get_obj_array(wk, tgt->objects)->len
+			&& !akw[bt_kw_link_whole].set && tgt->type != tgt_static_library) {
+			uint32_t node = akw[bt_kw_sources].set ? akw[bt_kw_sources].node : an[1].node;
 
-			interp_error(wk, node, "target declared with no linkable sources");
+			vm_error_at(wk, node, "target declared with no linkable sources");
 			return false;
 		}
+
+		if (akw[bt_kw_extra_files].set) {
+			obj_array_extend(wk, tgt->extra_files, akw[bt_kw_extra_files].val);
+		}
 	}
 
 	{ // include directories
@@ -663,18 +750,66 @@ create_target(struct workspace *wk, struct args_norm *an, struct args_kw *akw,
 		static struct {
 			enum build_target_kwargs kw;
 			enum compiler_language l;
+			bool static_only, shared_only;
 		} lang_args[] = {
-			{ bt_kw_c_args, compiler_language_c },
-			{ bt_kw_cpp_args, compiler_language_cpp },
-			{ bt_kw_objc_args, compiler_language_objc },
-			{ bt_kw_nasm_args, compiler_language_nasm },
+#define E(lang, s, st, sh) { bt_kw_##lang##s, compiler_language_##lang, st, sh }
+#define TOOLCHAIN_ENUM(lang) \
+	E(lang, _args, false, false), E(lang, _static_args, true, false), E(lang, _shared_args, false, true),
+			FOREACH_COMPILER_EXPOSED_LANGUAGE(TOOLCHAIN_ENUM)
+#undef TOOLCHAIN_ENUM
+#undef E
 		};
 
+		{ // copy c or cpp args to assembly args
+			if (akw[bt_kw_c_args].set) {
+				obj_dict_seti(wk, tgt->args, compiler_language_assembly, akw[bt_kw_c_args].val);
+			} else if (akw[bt_kw_cpp_args].set) {
+				obj_dict_seti(wk, tgt->args, compiler_language_assembly, akw[bt_kw_cpp_args].val);
+			}
+		}
+
 		uint32_t i;
 		for (i = 0; i < ARRAY_LEN(lang_args); ++i) {
-			if (akw[lang_args[i].kw].set) {
-				obj_dict_seti(wk, tgt->args, lang_args[i].l, akw[lang_args[i].kw].val);
+			if (!akw[lang_args[i].kw].set) {
+				continue;
+			} else if (lang_args[i].static_only && type != tgt_static_library) {
+				continue;
+			} else if (lang_args[i].shared_only && type != tgt_dynamic_library) {
+				continue;
 			}
+
+			obj_dict_seti(wk, tgt->args, lang_args[i].l, akw[lang_args[i].kw].val);
+		}
+	}
+
+	{ // pch
+		struct {
+			struct args_kw *kw;
+			enum compiler_language l;
+		} pch_args[] = {
+#define E(lang, s) { &akw[bt_kw_##lang##s], compiler_language_##lang },
+#define TOOLCHAIN_ENUM(lang) \
+	E(lang, _pch)
+			FOREACH_COMPILER_EXPOSED_LANGUAGE(TOOLCHAIN_ENUM)
+#undef TOOLCHAIN_ENUM
+#undef E
+		};
+
+		uint32_t i;
+		for (i = 0; i < ARRAY_LEN(pch_args); ++i) {
+			if (!pch_args[i].kw->set) {
+				continue;
+			}
+
+			obj pch_file;
+			if (!coerce_file(wk, pch_args[i].kw->node, pch_args[i].kw->val, &pch_file)) {
+				return false;
+			}
+
+			if (!tgt->pch) {
+				make_obj(wk, &tgt->pch, obj_dict);
+			}
+			obj_dict_seti(wk, tgt->pch, pch_args[i].l, pch_file);
 		}
 	}
 
@@ -682,16 +817,14 @@ create_target(struct workspace *wk, struct args_norm *an, struct args_kw *akw,
 
 	// soname handling
 	if (type & (tgt_dynamic_library | tgt_shared_module)) {
-		enum machine_system sys = machine_system();
-
-		if (sys == machine_system_windows || sys == machine_system_cygwin) {
+		if (host_machine.is_windows) {
 			setup_dllname(wk, tgt, plain_name, sover, akw[bt_kw_version].val);
 		} else {
 			setup_soname(wk, tgt, plain_name, sover, akw[bt_kw_version].val);
 
 			if (type == tgt_dynamic_library) {
-				if (!setup_shared_object_symlinks(wk, tgt, plain_name,
-					&plain_name_install, &soname_install)) {
+				if (!setup_shared_object_symlinks(
+					    wk, tgt, plain_name, &plain_name_install, &soname_install)) {
 					return false;
 				}
 			}
@@ -716,26 +849,20 @@ create_target(struct workspace *wk, struct args_norm *an, struct args_kw *akw,
 			install_dir = akw[bt_kw_install_dir].val;
 		} else {
 			switch (type) {
-			case tgt_executable:
-				get_option_value(wk, current_project(wk), "bindir", &install_dir);
-				break;
+			case tgt_executable: get_option_value(wk, current_project(wk), "bindir", &install_dir); break;
 			case tgt_static_library:
 				get_option_value(wk, current_project(wk), "libdir", &install_dir);
 				break;
 			case tgt_dynamic_library:
 			case tgt_shared_module: {
-				enum machine_system sys = machine_system();
-
-				if (sys == machine_system_windows || sys == machine_system_cygwin) {
+				if (host_machine.is_windows) {
 					get_option_value(wk, current_project(wk), "bindir", &install_dir);
 				} else {
 					get_option_value(wk, current_project(wk), "libdir", &install_dir);
 				}
 				break;
 			}
-			default:
-				assert(false && "unreachable");
-				break;
+			default: assert(false && "unreachable"); break;
 			}
 		}
 
@@ -746,21 +873,22 @@ create_target(struct workspace *wk, struct args_norm *an, struct args_kw *akw,
 		path_join(wk, &install_dest, get_cstr(wk, install_dir), get_cstr(wk, tgt->build_name));
 
 		struct obj_install_target *install_tgt;
-		if (!(install_tgt = push_install_target(wk, sbuf_into_str(wk, &install_src),
-			sbuf_into_str(wk, &install_dest), akw[bt_kw_install_mode].val))) {
+		if (!(install_tgt = push_install_target(wk,
+			      sbuf_into_str(wk, &install_src),
+			      sbuf_into_str(wk, &install_dest),
+			      akw[bt_kw_install_mode].val))) {
 			return false;
 		}
 
 		install_tgt->build_target = true;
 
 		if (soname_install) {
-			push_install_target_install_dir(wk, soname_install,
-				install_dir, akw[bt_kw_install_mode].val);
+			push_install_target_install_dir(wk, soname_install, install_dir, akw[bt_kw_install_mode].val);
 		}
 
 		if (plain_name_install) {
-			push_install_target_install_dir(wk, plain_name_install,
-				install_dir, akw[bt_kw_install_mode].val);
+			push_install_target_install_dir(
+				wk, plain_name_install, install_dir, akw[bt_kw_install_mode].val);
 		}
 	}
 
@@ -778,7 +906,7 @@ create_target(struct workspace *wk, struct args_norm *an, struct args_kw *akw,
 		}
 	}
 
-	tgt->dep = (struct build_dep) {
+	tgt->dep = (struct build_dep){
 		.link_language = tgt->dep_internal.link_language,
 		.include_directories = tgt->dep_internal.include_directories,
 		.order_deps = tgt->dep_internal.order_deps,
@@ -790,6 +918,7 @@ create_target(struct workspace *wk, struct args_norm *an, struct args_kw *akw,
 		tgt->dep.link_whole = tgt->dep_internal.link_whole;
 		tgt->dep.link_with = tgt->dep_internal.link_with;
 		tgt->dep.link_with_not_found = tgt->dep_internal.link_with_not_found;
+		tgt->dep.frameworks = tgt->dep_internal.frameworks;
 		obj_array_dup(wk, tgt->dep_internal.link_args, &tgt->dep.link_args);
 	}
 
@@ -817,21 +946,22 @@ typecheck_string_or_empty_array(struct workspace *wk, struct args_kw *kw)
 		kw->val = 0;
 		return true;
 	} else {
-		interp_error(wk, kw->node, "expected string or [], got %s", obj_type_to_s(t));
+		vm_error_at(wk, kw->node, "expected string or [], got %s", obj_type_to_s(t));
 		return false;
 	}
 }
 
 static bool
-tgt_common(struct workspace *wk, uint32_t args_node, obj *res, enum tgt_type type, enum tgt_type argtype, bool tgt_type_from_kw)
+tgt_common(struct workspace *wk, obj *res, enum tgt_type type, enum tgt_type argtype, bool tgt_type_from_kw)
 {
-	struct args_norm an[] = { { obj_string }, { ARG_TYPE_GLOB | tc_coercible_files | tc_generated_list }, ARG_TYPE_NULL };
+	struct args_norm an[]
+		= { { obj_string }, { TYPE_TAG_GLOB | tc_coercible_files | tc_generated_list }, ARG_TYPE_NULL };
 
-	struct args_kw akw[] = {
-		[bt_kw_sources] = { "sources", ARG_TYPE_ARRAY_OF | tc_coercible_files | tc_generated_list },
-		[bt_kw_include_directories] = { "include_directories", ARG_TYPE_ARRAY_OF | tc_coercible_inc },
+	struct args_kw akw[bt_kwargs_count + 1] = {
+		[bt_kw_sources] = { "sources", TYPE_TAG_LISTIFY | tc_coercible_files | tc_generated_list },
+		[bt_kw_include_directories] = { "include_directories", TYPE_TAG_LISTIFY | tc_coercible_inc },
 		[bt_kw_implicit_include_directories] = { "implicit_include_directories", obj_bool },
-		[bt_kw_dependencies] = { "dependencies", ARG_TYPE_ARRAY_OF | tc_dependency },
+		[bt_kw_dependencies] = { "dependencies", TYPE_TAG_LISTIFY | tc_dependency },
 		[bt_kw_install] = { "install", obj_bool },
 		[bt_kw_install_dir] = { "install_dir", obj_string },
 		[bt_kw_install_mode] = { "install_mode", tc_install_mode_kw },
@@ -840,13 +970,14 @@ tgt_common(struct workspace *wk, uint32_t args_node, obj *res, enum tgt_type typ
 		[bt_kw_link_whole] = { "link_whole", tc_link_with_kw },
 		[bt_kw_version] = { "version", obj_string },
 		[bt_kw_build_by_default] = { "build_by_default", obj_bool },
-		[bt_kw_extra_files] = { "extra_files", ARG_TYPE_ARRAY_OF | tc_coercible_files }, // ignored
+		[bt_kw_extra_files] = { "extra_files", TYPE_TAG_LISTIFY | tc_coercible_files },
 		[bt_kw_target_type] = { "target_type", obj_string },
 		[bt_kw_name_prefix] = { "name_prefix", tc_string | tc_array },
 		[bt_kw_name_suffix] = { "name_suffix", tc_string | tc_array },
 		[bt_kw_soversion] = { "soversion", tc_number | tc_string },
-		[bt_kw_link_depends] = { "link_depends", ARG_TYPE_ARRAY_OF | tc_string | tc_file | tc_custom_target },
-		[bt_kw_objects] = { "objects", ARG_TYPE_ARRAY_OF | tc_file | tc_string },
+		[bt_kw_link_depends] = { "link_depends",
+			TYPE_TAG_LISTIFY | tc_string | tc_file | tc_custom_target | tc_build_target | tc_build_target },
+		[bt_kw_objects] = { "objects", TYPE_TAG_LISTIFY | tc_file | tc_string },
 		[bt_kw_pic] = { "pic", obj_bool },
 		[bt_kw_pie] = { "pie", obj_bool },
 		[bt_kw_build_rpath] = { "build_rpath", obj_string },
@@ -855,31 +986,29 @@ tgt_common(struct workspace *wk, uint32_t args_node, obj *res, enum tgt_type typ
 		[bt_kw_vs_module_defs] = { "vs_module_defs", tc_string | tc_file | tc_custom_target },
 		[bt_kw_gnu_symbol_visibility] = { "gnu_symbol_visibility", obj_string },
 		[bt_kw_native] = { "native", obj_bool },
-		[bt_kw_darwin_versions] = { "darwin_versions", ARG_TYPE_ARRAY_OF | tc_string | tc_number },
-		[bt_kw_implib] = { "implib", obj_bool },
+		[bt_kw_darwin_versions] = { "darwin_versions", TYPE_TAG_LISTIFY | tc_string | tc_number },
+		[bt_kw_implib] = { "implib", tc_bool | tc_string },
 		[bt_kw_gui_app] = { "gui_app", obj_bool },
 		[bt_kw_link_language] = { "link_language", obj_string },
 		[bt_kw_win_subsystem] = { "win_subsystem", obj_string },
-		[bt_kw_override_options] = { "override_options", ARG_TYPE_ARRAY_OF | obj_string },
-		/* lang args */
-		[bt_kw_c_pch] = { "c_pch", tc_string | tc_file, },
-		[bt_kw_cpp_pch] = { "cpp_pch", tc_string | tc_file, },
-		[bt_kw_c_args] = { "c_args", ARG_TYPE_ARRAY_OF | obj_string },
-		[bt_kw_cpp_args] = { "cpp_args", ARG_TYPE_ARRAY_OF | obj_string },
-		[bt_kw_objc_args] = { "objc_args", ARG_TYPE_ARRAY_OF | obj_string },
-		[bt_kw_nasm_args] = { "nasm_args", ARG_TYPE_ARRAY_OF | obj_string },
-		[bt_kw_masm_args] = { "masm_args", ARG_TYPE_ARRAY_OF | obj_string },
-		[bt_kw_link_args] = { "link_args", ARG_TYPE_ARRAY_OF | obj_string },
-		0
+		[bt_kw_override_options] = { "override_options", wk->complex_types.options_dict_or_list },
+		[bt_kw_link_args] = { "link_args", TYPE_TAG_LISTIFY | obj_string },
+#define E(lang, s, t) [bt_kw_##lang##s] = { #lang #s, t }
+#define TOOLCHAIN_ENUM(lang)                                                                                 \
+	E(lang, _args, TYPE_TAG_LISTIFY | obj_string), E(lang, _static_args, TYPE_TAG_LISTIFY | obj_string), \
+		E(lang, _shared_args, TYPE_TAG_LISTIFY | obj_string), E(lang, _pch, tc_string | tc_file),
+		FOREACH_COMPILER_EXPOSED_LANGUAGE(TOOLCHAIN_ENUM)
+#undef TOOLCHAIN_ENUM
+#undef E
 	};
 
-	if (!interp_args(wk, args_node, an, NULL, akw)) {
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
 	if (tgt_type_from_kw) {
 		if (!akw[bt_kw_target_type].set) {
-			interp_error(wk, args_node, "missing required kwarg: %s", akw[bt_kw_target_type].key);
+			vm_error(wk, "missing required kwarg: %s", akw[bt_kw_target_type].key);
 			return false;
 		}
 
@@ -888,7 +1017,7 @@ tgt_common(struct workspace *wk, uint32_t args_node, obj *res, enum tgt_type typ
 		}
 	} else {
 		if (akw[bt_kw_target_type].set) {
-			interp_error(wk, akw[bt_kw_target_type].node, "invalid kwarg");
+			vm_error_at(wk, akw[bt_kw_target_type].node, "invalid kwarg");
 			return false;
 		}
 	}
@@ -896,15 +1025,12 @@ tgt_common(struct workspace *wk, uint32_t args_node, obj *res, enum tgt_type typ
 	static const enum tgt_type keyword_validity[bt_kwargs_count] = {
 		[bt_kw_version] = tgt_dynamic_library,
 		[bt_kw_soversion] = tgt_dynamic_library,
-		[bt_kw_vs_module_defs] = tgt_dynamic_library | tgt_static_library | tgt_shared_module,
 	};
 
 	uint32_t i;
 	for (i = 0; i < bt_kwargs_count; ++i) {
-		if (keyword_validity[i]
-		    && akw[i].set
-		    && !(keyword_validity[i] & argtype)) {
-			interp_error(wk, akw[i].node, "invalid kwarg");
+		if (keyword_validity[i] && akw[i].set && !(keyword_validity[i] & argtype)) {
+			vm_error_at(wk, akw[i].node, "invalid kwarg");
 			return false;
 		}
 	}
@@ -915,10 +1041,8 @@ tgt_common(struct workspace *wk, uint32_t args_node, obj *res, enum tgt_type typ
 		return false;
 	}
 
-	if (type == (tgt_static_library | tgt_dynamic_library)
-	    && !akw[bt_kw_pic].set) {
-		make_obj(wk, &akw[bt_kw_pic].val, obj_bool);
-		set_obj_bool(wk, akw[bt_kw_pic].val, true);
+	if (type == (tgt_static_library | tgt_dynamic_library) && !akw[bt_kw_pic].set) {
+		akw[bt_kw_pic].val = make_obj_bool(wk, true);
 		akw[bt_kw_pic].set = true;
 	}
 
@@ -982,47 +1106,44 @@ tgt_common(struct workspace *wk, uint32_t args_node, obj *res, enum tgt_type typ
 }
 
 bool
-func_executable(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_executable(struct workspace *wk, obj _, obj *res)
 {
-	return tgt_common(wk, args_node, res, tgt_executable, tgt_executable, false);
+	return tgt_common(wk, res, tgt_executable, tgt_executable, false);
 }
 
 bool
-func_static_library(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_static_library(struct workspace *wk, obj _, obj *res)
 {
-	return tgt_common(wk, args_node, res, tgt_static_library, tgt_static_library, false);
+	return tgt_common(wk, res, tgt_static_library, tgt_static_library, false);
 }
 
 bool
-func_shared_library(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_shared_library(struct workspace *wk, obj _, obj *res)
 {
-	return tgt_common(wk, args_node, res, tgt_dynamic_library, tgt_dynamic_library, false);
+	return tgt_common(wk, res, tgt_dynamic_library, tgt_dynamic_library, false);
 }
 
 bool
-func_both_libraries(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_both_libraries(struct workspace *wk, obj _, obj *res)
 {
-	return tgt_common(wk, args_node, res, tgt_static_library | tgt_dynamic_library,
-		tgt_static_library | tgt_dynamic_library, false);
+	return tgt_common(
+		wk, res, tgt_static_library | tgt_dynamic_library, tgt_static_library | tgt_dynamic_library, false);
 }
 
 bool
-func_library(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_library(struct workspace *wk, obj _, obj *res)
 {
-	return tgt_common(wk, args_node, res, get_option_default_library(wk),
-		tgt_static_library | tgt_dynamic_library, false);
+	return tgt_common(wk, res, get_option_default_library(wk), tgt_static_library | tgt_dynamic_library, false);
 }
 
 bool
-func_shared_module(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_shared_module(struct workspace *wk, obj _, obj *res)
 {
-	return tgt_common(wk, args_node, res, tgt_shared_module,
-		tgt_shared_module, false);
+	return tgt_common(wk, res, tgt_shared_module, tgt_shared_module, false);
 }
 
 bool
-func_build_target(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_build_target(struct workspace *wk, obj _, obj *res)
 {
-	return tgt_common(wk, args_node, res, 0,
-		tgt_executable | tgt_static_library | tgt_dynamic_library, true);
+	return tgt_common(wk, res, 0, tgt_executable | tgt_static_library | tgt_dynamic_library, true);
 }
