@@ -17,12 +17,11 @@
 #include "functions/kernel/configure_file.h"
 #include "functions/kernel/custom_target.h"
 #include "install.h"
-#include "lang/func_lookup.h"
 #include "lang/object_iterators.h"
 #include "lang/typecheck.h"
 #include "log.h"
+#include "platform/assert.h"
 #include "platform/filesystem.h"
-#include "platform/mem.h"
 #include "platform/path.h"
 #include "platform/run_cmd.h"
 
@@ -52,7 +51,7 @@ file_exists_with_content(struct workspace *wk, const char *dest, const char *out
 static void
 configure_file_skip_whitespace(const struct source *src, uint32_t *i)
 {
-	while (src->src[*i] && strchr(" \t", src->src[*i])) {
+	while (src->src[*i] && is_whitespace_except_newline(src->src[*i])) {
 		++(*i);
 	}
 }
@@ -121,6 +120,7 @@ substitute_config(struct workspace *wk,
 	char tmp_buf[BUF_SIZE_1k] = { 0 };
 
 	for (i = 0; i < src.len; ++i) {
+		location.off = i;
 		if (src.src[i] == '\n') {
 			col = i + 1;
 		}
@@ -271,7 +271,6 @@ write_mesondefine:
 			i += varstart_len;
 			id_start = i;
 			id_location = location;
-			/* id_location.col = id_start - location.col + 1; */
 			i += configure_var_len(&src.src[id_start]);
 
 			if (src.src[i] != varend) {
