@@ -20,6 +20,7 @@
 #include "install.h"
 #include "lang/typecheck.h"
 #include "log.h"
+#include "platform/assert.h"
 #include "platform/filesystem.h"
 #include "platform/path.h"
 
@@ -321,7 +322,7 @@ custom_target_cmd_fmt_iter(struct workspace *wk, void *_ctx, obj val)
 		goto cont;
 	}
 	case obj_compiler: {
-		obj cmd_array = get_obj_compiler(wk, val)->cmd_arr;
+		obj cmd_array = get_obj_compiler(wk, val)->cmd_arr[toolchain_component_compiler];
 
 		if (!obj_array_foreach(wk, cmd_array, ctx, custom_target_cmd_fmt_iter)) {
 			return ir_err;
@@ -474,6 +475,7 @@ make_custom_target(struct workspace *wk, struct make_custom_target_opts *opts, o
 	make_obj(wk, res, obj_custom_target);
 	struct obj_custom_target *tgt = get_obj_custom_target(wk, *res);
 	tgt->name = opts->name;
+	tgt->callstack = vm_callstack(wk);
 
 	// A custom_target won't have a name if it is from a generator
 	if (opts->name) { /* private path */
@@ -591,6 +593,39 @@ make_custom_target(struct workspace *wk, struct make_custom_target_opts *opts, o
 	return true;
 }
 
+bool
+install_custom_target(struct workspace *wk,
+	struct obj_custom_target *tgt,
+	const struct args_kw *kw_install,
+	const struct args_kw *kw_build_by_default,
+	obj install_dir,
+	obj install_mode)
+{
+	bool should_install = get_obj_bool_with_default(wk, kw_install->val, false)
+			      || (!kw_install->set && install_dir);
+
+	if (!should_install) {
+		return true;
+	}
+
+	if (!kw_build_by_default || !kw_build_by_default->set) {
+		tgt->flags |= custom_target_build_by_default;
+	}
+
+	if (!install_dir || !get_obj_array(wk, install_dir)->len) {
+		vm_error(wk, "custom target installation requires install_dir");
+		return false;
+	}
+
+	if (get_obj_array(wk, install_dir)->len == 1) {
+		obj i0;
+		obj_array_index(wk, install_dir, 0, &i0);
+		install_dir = i0;
+	}
+
+	return push_install_targets(wk, 0, tgt->output, install_dir, install_mode, false);
+}
+
 bool
 func_custom_target(struct workspace *wk, obj _, obj *res)
 {
@@ -712,33 +747,13 @@ func_custom_target(struct workspace *wk, obj _, obj *res)
 		tgt->flags |= custom_target_console;
 	}
 
-	if ((akw[kw_install].set && get_obj_bool(wk, akw[kw_install].val))
-		|| (!akw[kw_install].set && akw[kw_install_dir].set)) {
-		if (!akw[kw_install_dir].set || !get_obj_array(wk, akw[kw_install_dir].val)->len) {
-			vm_error_at(wk, akw[kw_install].node, "custom target installation requires install_dir");
-			return false;
-		}
-
-		if (!akw[kw_build_by_default].set) {
-			tgt->flags |= custom_target_build_by_default;
-		}
-
-		obj install_mode_id = 0;
-		if (akw[kw_install_mode].set) {
-			install_mode_id = akw[kw_install_mode].val;
-		}
-
-		obj install_dir = akw[kw_install_dir].val;
-		if (get_obj_array(wk, akw[kw_install_dir].val)->len == 1) {
-			obj i0;
-			obj_array_index(wk, akw[kw_install_dir].val, 0, &i0);
-			install_dir = i0;
-		}
-
-		if (!push_install_targets(
-			    wk, akw[kw_install_dir].node, tgt->output, install_dir, install_mode_id, false)) {
-			return false;
-		}
+	if (!install_custom_target(wk,
+		    tgt,
+		    &akw[kw_install],
+		    &akw[kw_build_by_default],
+		    akw[kw_install_dir].val,
+		    akw[kw_install_mode].val)) {
+		return false;
 	}
 
 	if (!coerce_environment_from_kwarg(wk, &akw[kw_env], false, &tgt->env)) {
@@ -791,7 +806,7 @@ func_vcs_tag(struct workspace *wk, obj _, obj *res)
 			"internal",
 			"eval",
 			"-e",
-			"vcs_tagger.meson",
+			"commands/vcs_tagger.meson",
 			NULL,
 		});
 
@@ -808,13 +823,19 @@ func_vcs_tag(struct workspace *wk, obj _, obj *res)
 		}
 	}
 
+	obj_array_push(wk, command, make_str(wk, "-s"));
 	obj_array_push(wk, command, input);
+	obj_array_push(wk, command, make_str(wk, "-d"));
 	obj_array_push(wk, command, make_str(wk, "@OUTPUT@"));
+	obj_array_push(wk, command, make_str(wk, "-p"));
 	obj_array_push(wk, command, replace_string);
+	obj_array_push(wk, command, make_str(wk, "-f"));
 	obj_array_push(wk, command, fallback);
+	obj_array_push(wk, command, make_str(wk, "-r"));
 	obj_array_push(wk, command, make_str(wk, wk->source_root));
 
 	if (akw[kw_command].set) {
+		obj_array_push(wk, command, make_str(wk, "--"));
 		obj_array_extend(wk, command, akw[kw_command].val);
 	}
 
