@@ -9,7 +9,6 @@
 #include <string.h>
 
 #include "args.h"
-#include "coerce.h"
 #include "error.h"
 #include "formats/editorconfig.h"
 #include "functions/kernel/custom_target.h"
@@ -17,7 +16,9 @@
 #include "lang/func_lookup.h"
 #include "lang/typecheck.h"
 #include "log.h"
+#include "platform/assert.h"
 #include "platform/filesystem.h"
+#include "platform/os.h"
 #include "platform/path.h"
 #include "sha_256.h"
 
@@ -124,8 +125,7 @@ func_module_fs_lookup_common(struct workspace *wk,
 		return false;
 	}
 
-	make_obj(wk, res, obj_bool);
-	set_obj_bool(wk, *res, lookup(path.buf));
+	*res = make_obj_bool(wk, lookup(path.buf));
 	return true;
 }
 
@@ -217,9 +217,8 @@ func_module_fs_is_absolute(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	make_obj(wk, res, obj_bool);
 	// TODO: Handle this
-	set_obj_bool(wk, *res, path_is_absolute(get_cstr(wk, an[0].val)));
+	*res = make_obj_bool(wk, path_is_absolute(get_cstr(wk, an[0].val)));
 	return true;
 }
 
@@ -424,8 +423,7 @@ func_module_fs_is_samepath(struct workspace *wk, obj self, obj *res)
 
 	// TODO: handle symlinks
 
-	make_obj(wk, res, obj_bool);
-	set_obj_bool(wk, *res, strcmp(path1.buf, path2.buf) == 0);
+	*res = make_obj_bool(wk, strcmp(path1.buf, path2.buf) == 0);
 	return true;
 }
 
@@ -474,7 +472,7 @@ func_module_fs_copyfile(struct workspace *wk, obj self, obj *res)
 			"internal",
 			"eval",
 			"-e",
-			"copyfile.meson",
+			"commands/copyfile.meson",
 			"@INPUT@",
 			"@OUTPUT@",
 			NULL,
@@ -496,6 +494,12 @@ func_module_fs_copyfile(struct workspace *wk, obj self, obj *res)
 
 	obj_array_push(wk, current_project(wk)->targets, *res);
 
+	struct obj_custom_target *tgt = get_obj_custom_target(wk, *res);
+
+	if (!install_custom_target(wk, tgt, &akw[kw_install], 0, akw[kw_install_dir].val, 0)) {
+		return false;
+	}
+
 	return true;
 }
 
@@ -537,23 +541,23 @@ func_module_fs_relative_to(struct workspace *wk, obj self, obj *res)
 }
 
 const struct func_impl impl_tbl_module_fs[] = {
-	{ "as_posix", func_module_fs_as_posix, tc_string, true },
-	{ "copyfile", func_module_fs_copyfile, tc_custom_target },
-	{ "exists", func_module_fs_exists, tc_bool },
-	{ "expanduser", func_module_fs_expanduser, tc_string },
-	{ "hash", func_module_fs_hash, tc_string },
-	{ "is_absolute", func_module_fs_is_absolute, tc_bool, true },
-	{ "is_dir", func_module_fs_is_dir, tc_bool },
-	{ "is_file", func_module_fs_is_file, tc_bool },
-	{ "is_samepath", func_module_fs_is_samepath, tc_bool },
-	{ "is_symlink", func_module_fs_is_symlink, tc_bool },
-	{ "name", func_module_fs_name, tc_string, true },
-	{ "parent", func_module_fs_parent, tc_string, true },
-	{ "read", func_module_fs_read, tc_string },
-	{ "relative_to", func_module_fs_relative_to, tc_string, true },
-	{ "replace_suffix", func_module_fs_replace_suffix, tc_string, true },
-	{ "size", func_module_fs_size, tc_number },
-	{ "stem", func_module_fs_stem, tc_string, true },
+	{ "as_posix", func_module_fs_as_posix, tc_string, true, .desc = "Convert backslashes to `/`" },
+	{ "copyfile", func_module_fs_copyfile, tc_custom_target, .desc = "Creates a custom target that copies a file at build time" },
+	{ "exists", func_module_fs_exists, tc_bool, .desc = "Check if a file or directory exists" },
+	{ "expanduser", func_module_fs_expanduser, tc_string, .desc = "Expand a path that starts with a ~ into an absolute path"  },
+	{ "hash", func_module_fs_hash, tc_string, .desc = "Calculate the content hash of a file" },
+	{ "is_absolute", func_module_fs_is_absolute, tc_bool, true, .desc = "Check if a path is absolute" },
+	{ "is_dir", func_module_fs_is_dir, tc_bool, .desc = "Check if a directory exists" },
+	{ "is_file", func_module_fs_is_file, tc_bool, .desc = "Check if a file exists" },
+	{ "is_samepath", func_module_fs_is_samepath, tc_bool, .desc = "Check if two paths point to the same object" },
+	{ "is_symlink", func_module_fs_is_symlink, tc_bool, .desc = "Check if a symlink exists" },
+	{ "name", func_module_fs_name, tc_string, true, .desc = "Get the basename of a path" },
+	{ "parent", func_module_fs_parent, tc_string, true, .desc = "Get the dirname of a path" },
+	{ "read", func_module_fs_read, tc_string, .desc = "Read a file from disk" },
+	{ "relative_to", func_module_fs_relative_to, tc_string, true, .desc = "Get a path relative to another path" },
+	{ "replace_suffix", func_module_fs_replace_suffix, tc_string, true, .desc = "Replace a path's suffix" },
+	{ "size", func_module_fs_size, tc_number, .desc = "Get a file's size in bytes" },
+	{ "stem", func_module_fs_stem, tc_string, true, .desc = "Get the basename of a path with its extension removed" },
 	{ NULL, NULL },
 };
 
@@ -581,16 +585,26 @@ static bool
 func_module_fs_copy(struct workspace *wk, obj self, obj *res)
 {
 	struct args_norm an[] = { { tc_string | tc_file }, { obj_string }, ARG_TYPE_NULL };
-	if (!pop_args(wk, an, NULL)) {
+	enum kwargs {
+		kw_force,
+	};
+	struct args_kw akw[] = {
+		[kw_force] = { "force", obj_bool },
+		0,
+	};
+
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
+	bool force = akw[kw_force].set ? get_obj_bool(wk, akw[kw_force].val) : false;
+
 	SBUF(path);
 	if (!fix_file_path(wk, an[0].node, an[0].val, 0, &path)) {
 		return false;
 	}
 
-	if (!fs_copy_file(path.buf, get_cstr(wk, an[1].val))) {
+	if (!fs_copy_file(path.buf, get_cstr(wk, an[1].val), force)) {
 		return false;
 	}
 	return true;
@@ -679,8 +693,7 @@ func_module_fs_is_basename(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	make_obj(wk, res, obj_bool);
-	set_obj_bool(wk, *res, path_is_basename(get_cstr(wk, an[0].val)));
+	*res = make_obj_bool(wk, path_is_basename(get_cstr(wk, an[0].val)));
 	return true;
 }
 
@@ -706,8 +719,7 @@ func_module_fs_is_subpath(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	make_obj(wk, res, obj_bool);
-	set_obj_bool(wk, *res, path_is_subpath(get_cstr(wk, an[0].val), get_cstr(wk, an[1].val)));
+	*res = make_obj_bool(wk, path_is_subpath(get_cstr(wk, an[0].val), get_cstr(wk, an[1].val)));
 	return true;
 }
 
@@ -836,8 +848,7 @@ func_module_fs_glob(struct workspace *wk, obj self, obj *res)
 			return false;
 		}
 
-		/* path_join(wk, &pat, workspace_cwd(wk), _pat->s); */
-		path_copy(wk, &pat, _pat->s);
+		path_join(wk, &pat, workspace_cwd(wk), _pat->s);
 	}
 
 	SBUF(prefix);
@@ -871,7 +882,8 @@ func_module_fs_glob(struct workspace *wk, obj self, obj *res)
 	make_obj(wk, res, obj_array);
 
 	if (!fs_dir_exists(prefix.buf)) {
-		return true;
+		vm_error(wk, "Path \"%s\" does not exist", prefix.buf);
+		return false;
 	}
 
 	struct func_module_fs_glob_ctx ctx = {
@@ -884,6 +896,94 @@ func_module_fs_glob(struct workspace *wk, obj self, obj *res)
 	return fs_dir_foreach(prefix.buf, &ctx, func_module_fs_glob_cb);
 }
 
+struct delete_suffixes_ctx {
+	const char *base_dir;
+	const char *suffix;
+};
+
+static enum iteration_result
+delete_suffix_recursive(void *_ctx, const char *path)
+{
+	enum iteration_result ir_res = ir_err;
+	struct delete_suffixes_ctx *ctx = _ctx;
+	struct stat sb;
+
+	SBUF(name);
+
+	path_join(NULL, &name, ctx->base_dir, path);
+
+	// Check for existence first, before doing a stat. It's possible that
+	// another process has deleted this file in the time since the iteration
+	// machinery found this file. In this case, the file no longer exists, so
+	// just continue.
+	if (!fs_exists(name.buf)) {
+		sbuf_destroy(&name);
+		return ir_cont;
+	}
+
+	if (!fs_stat(name.buf, &sb)) {
+		goto ret;
+	}
+
+	if (S_ISDIR(sb.st_mode)) {
+		struct delete_suffixes_ctx new_ctx = *ctx;
+		new_ctx.base_dir = name.buf;
+
+		if (!fs_dir_foreach(new_ctx.base_dir, &new_ctx, delete_suffix_recursive)) {
+			goto ret;
+		}
+
+	} else if (S_ISREG(sb.st_mode)) {
+		if (fs_has_extension(name.buf, ctx->suffix) && !fs_remove(name.buf)) {
+			goto ret;
+		}
+
+	} else {
+		LOG_E("unhandled file type: %s", name.buf);
+		goto ret;
+	}
+
+	ir_res = ir_cont;
+
+ret:
+	sbuf_destroy(&name);
+	return ir_res;
+}
+
+static bool
+func_module_fs_delete_with_suffix(struct workspace *wk, obj rcrv, obj *res)
+{
+	struct args_norm an[] = { { tc_string }, { tc_string }, ARG_TYPE_NULL };
+	if (!pop_args(wk, an, NULL)) {
+		return false;
+	}
+
+	const char *base_dir = get_cstr(wk, an[0].val);
+	struct delete_suffixes_ctx ctx = {
+		.base_dir = base_dir,
+		.suffix = get_cstr(wk, an[1].val),
+	};
+
+	return fs_dir_foreach(base_dir, &ctx, delete_suffix_recursive);
+}
+
+static bool
+func_module_fs_canonicalize(struct workspace *wk, obj self, obj *res)
+{
+	struct args_norm an[] = { { tc_string }, ARG_TYPE_NULL };
+	if (!pop_args(wk, an, NULL)) {
+		return false;
+	}
+
+	SBUF(path);
+	if (!fix_file_path(wk, an[0].node, an[0].val, fix_file_path_noexpanduser, &path)) {
+		return false;
+	}
+
+	*res = sbuf_into_str(wk, &path);
+	return true;
+}
+
 const struct func_impl impl_tbl_module_fs_internal[] = {
 	{ "as_posix", func_module_fs_as_posix, tc_string, true },
 	{ "copyfile", func_module_fs_copyfile },
@@ -898,22 +998,24 @@ const struct func_impl impl_tbl_module_fs_internal[] = {
 	{ "name", func_module_fs_name, tc_string, true },
 	{ "parent", func_module_fs_parent, tc_string, true },
 	{ "read", func_module_fs_read, tc_string },
+	{ "relative_to", func_module_fs_relative_to, tc_string, true },
 	{ "replace_suffix", func_module_fs_replace_suffix, tc_string, true },
 	{ "size", func_module_fs_size, tc_number },
 	{ "stem", func_module_fs_stem, tc_string, true },
 	// non-standard muon extensions
-	{ "add_suffix", func_module_fs_add_suffix, tc_string, true },
-	{ "copy", func_module_fs_copy, .fuzz_unsafe = true },
-	{ "cwd", func_module_fs_cwd, tc_string },
-	{ "executable", func_module_fs_executable, tc_string, true },
-	{ "glob", func_module_fs_glob, tc_array },
-	{ "is_basename", func_module_fs_is_basename, tc_bool, true },
-	{ "is_subpath", func_module_fs_is_subpath, tc_bool, true },
-	{ "make_absolute", func_module_fs_make_absolute, tc_string },
-	{ "mkdir", func_module_fs_mkdir, .fuzz_unsafe = true },
-	{ "relative_to", func_module_fs_relative_to, tc_string, true },
-	{ "rmdir", func_module_fs_rmdir, .fuzz_unsafe = true },
-	{ "without_ext", func_module_fs_without_ext, tc_string, true },
-	{ "write", func_module_fs_write, .fuzz_unsafe = true },
+	{ "add_suffix", func_module_fs_add_suffix, tc_string, true, .desc = "Append a suffix to a path" },
+	{ "copy", func_module_fs_copy, .fuzz_unsafe = true, .desc = "Copy a file to a destination at configure time" },
+	{ "cwd", func_module_fs_cwd, tc_string, .desc = "Return the muon's current working directory" },
+	{ "executable", func_module_fs_executable, tc_string, true, .desc = "Prepend ./ to a path if it is a basename" },
+	{ "glob", func_module_fs_glob, tc_array, .desc = "Gather a list of files matching a glob expression" },
+	{ "is_basename", func_module_fs_is_basename, tc_bool, true, .desc = "Check if a path is a basename" },
+	{ "is_subpath", func_module_fs_is_subpath, tc_bool, true, .desc = "Check if a path is a subpath of another path" },
+	{ "make_absolute", func_module_fs_make_absolute, tc_string, .desc = "Make a path absolute by prepending the current working directory" },
+	{ "mkdir", func_module_fs_mkdir, .fuzz_unsafe = true, .desc = "Create a directory" },
+	{ "rmdir", func_module_fs_rmdir, .fuzz_unsafe = true, .desc = "Remove a directory" },
+	{ "without_ext", func_module_fs_without_ext, tc_string, true, .desc = "Get a path with its extension removed" },
+	{ "write", func_module_fs_write, .fuzz_unsafe = true, .desc = "Write a file" },
+	{ "delete_with_suffix", func_module_fs_delete_with_suffix, .desc = "Recursively delete all files with a matching suffix" },
+	{ "canonicalize", func_module_fs_canonicalize, tc_string, true, .desc = "Make a path absolute and replace .."  },
 	{ NULL, NULL },
 };
