@@ -6,11 +6,12 @@
 #include "compat.h"
 
 #include "functions/modules/sourceset.h"
+#include "lang/typecheck.h"
 
 static bool
-func_module_sourceset_source_set(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_module_sourceset_source_set(struct workspace *wk, obj self, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
@@ -20,8 +21,11 @@ func_module_sourceset_source_set(struct workspace *wk, obj rcvr, uint32_t args_n
 	return true;
 }
 
-const struct func_impl_name impl_tbl_module_sourceset[] = {
-	{ "source_set", func_module_sourceset_source_set, tc_source_set, },
+const struct func_impl impl_tbl_module_sourceset[] = {
+	{
+		"source_set",
+		func_module_sourceset_source_set,
+		tc_source_set,
+	},
 	{ NULL, NULL },
 };
-
