@@ -10,7 +10,6 @@
 #include "functions/number.h"
 #include "lang/func_lookup.h"
 #include "lang/typecheck.h"
-#include "log.h"
 
 static bool
 func_number_is_odd(struct workspace *wk, obj self, obj *res)
@@ -19,8 +18,7 @@ func_number_is_odd(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	make_obj(wk, res, obj_bool);
-	set_obj_bool(wk, *res, (get_obj_number(wk, self) & 1) != 0);
+	*res = make_obj_bool(wk, (get_obj_number(wk, self) & 1) != 0);
 	return true;
 }
 
@@ -31,8 +29,7 @@ func_number_is_even(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	make_obj(wk, res, obj_bool);
-	set_obj_bool(wk, *res, (get_obj_number(wk, self) & 1) == 0);
+	*res = make_obj_bool(wk, (get_obj_number(wk, self) & 1) == 0);
 	return true;
 }
 
