@@ -12,36 +12,33 @@
 
 #include "buf_size.h"
 #include "coerce.h"
-#include "functions/common.h"
 #include "functions/string.h"
-#include "guess.h"
-#include "lang/interpreter.h"
-#include "log.h"
+#include "lang/func_lookup.h"
+#include "lang/lexer.h"
+#include "lang/typecheck.h"
 #include "rpmvercmp.h"
+#include "util.h"
 
 static bool
-func_strip(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_strip(struct workspace *wk, obj self, obj *res)
 {
-	struct args_norm ao[] = { { obj_string }, ARG_TYPE_NULL };
-	if (!interp_args(wk, args_node, NULL, ao, NULL)) {
+	struct args_norm an[] = { { obj_string, .optional = true }, ARG_TYPE_NULL };
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
-	*res = str_strip(wk,
-		get_str(wk, rcvr),
-		ao[0].set ? get_str(wk, ao[0].val) : NULL);
+	*res = str_strip(wk, get_str(wk, self), an[0].set ? get_str(wk, an[0].val) : NULL, 0);
 	return true;
 }
 
 static bool
-func_to_upper(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_to_upper(struct workspace *wk, obj self, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
-	*res = str_clone(wk, wk, rcvr);
-
+	*res = str_clone_mutable(wk, self);
 	const struct str *ss = get_str(wk, *res);
 
 	uint32_t i;
@@ -55,13 +52,13 @@ func_to_upper(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 }
 
 static bool
-func_to_lower(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_to_lower(struct workspace *wk, obj self, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
-	*res = str_clone(wk, wk, rcvr);
+	*res = str_clone_mutable(wk, self);
 
 	const struct str *ss = get_str(wk, *res);
 
@@ -78,77 +75,70 @@ func_to_lower(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 bool
 string_format(struct workspace *wk, uint32_t err_node, obj str, obj *res, void *ctx, string_format_cb cb)
 {
-	struct str key;
 	const struct str *ss_in = get_str(wk, str);
+	struct str key, text = { .s = ss_in->s };
+	obj elem;
 
-	uint32_t i, id_start = 0, id_end = 0;
+	uint32_t i;
 	bool reading_id = false;
 
 	*res = make_str(wk, "");
 
 	for (i = 0; i < ss_in->len; ++i) {
-		if (ss_in->s[i] == '@') {
-			if (reading_id) {
-				obj elem;
-				id_end = i + 1;
-
-				if (i == id_start) {
-					str_app(wk, *res, "@");
-					id_start = i + 1;
-					reading_id = true;
-					key.len = 0;
-					continue;
-				}
-
-				key = (struct str){ .s = &ss_in->s[id_start], .len = i - id_start };
+		if (reading_id) {
+			key.len = &ss_in->s[i] - key.s;
 
+			if (ss_in->s[i] == '@') {
 				switch (cb(wk, err_node, ctx, &key, &elem)) {
 				case format_cb_not_found: {
-					interp_error(wk, err_node, "key '%.*s' not found", key.len, key.s);
+					vm_error(wk, "key '%.*s' not found", key.len, key.s);
 					return false;
 				}
-				case format_cb_error:
-					return false;
+				case format_cb_error: return false;
 				case format_cb_found: {
 					obj coerced;
 					if (!coerce_string(wk, err_node, elem, &coerced)) {
 						return false;
 					}
 
-					const struct str *ss = get_str(wk, coerced);
-					str_appn(wk, *res, ss->s, ss->len);
+					str_apps(wk, res, coerced);
+					text.s = &ss_in->s[i + 1];
 					break;
 				}
 				case format_cb_skip: {
-					str_app(wk, *res, "@");
-					i = id_start - 1;
-					id_end = id_start;
-					id_start = 0;
-					reading_id = false;
-					continue;
+					str_appn(wk, res, key.s - 1, key.len + 1);
+					text.s = &ss_in->s[i];
+					--i;
+					break;
 				}
 				}
 
 				reading_id = false;
-			} else {
-				if (i) {
-					str_appn(wk, *res, &ss_in->s[id_end], i - id_end);
-				}
-
-				id_start = i + 1;
-				reading_id = true;
-				key.len = 0;
+			} else if (!is_valid_inside_of_identifier(ss_in->s[i])) {
+				str_appn(wk, res, key.s - 1, key.len + 1);
+				text.s = &ss_in->s[i];
+				reading_id = false;
 			}
+		} else if (ss_in->s[i] == '@' && is_valid_inside_of_identifier(ss_in->s[i + 1])) {
+			text.len = &ss_in->s[i] - text.s;
+			str_appn(wk, res, text.s, text.len);
+			text.s = &ss_in->s[i];
+
+			reading_id = true;
+			key.s = &ss_in->s[i + 1];
+		} else if (ss_in->s[i] == '\\' && ss_in->s[i + 1] == '@') {
+			text.len = &ss_in->s[i] - text.s;
+			str_appn(wk, res, text.s, text.len);
+			text.s = &ss_in->s[i + 1];
+			++i;
 		}
 	}
 
+	text.len = &ss_in->s[i] - text.s;
+	str_appn(wk, res, text.s, text.len);
+
 	if (reading_id) {
-		str_app(wk, *res, "@");
-		str_appn(wk, *res, &ss_in->s[id_start], i - id_start);
-	} else {
-		if (i > id_end) {
-			str_appn(wk, *res, &ss_in->s[id_end], i - id_end);
-		}
+		vm_warning(wk, "unclosed @");
 	}
 
 	return true;
@@ -164,7 +154,7 @@ func_format_cb(struct workspace *wk, uint32_t node, void *_ctx, const struct str
 	struct func_format_ctx *ctx = _ctx;
 	int64_t i;
 
-	if (!str_to_i(key, &i)) {
+	if (!str_to_i(key, &i, false)) {
 		return format_cb_skip;
 	}
 
@@ -178,11 +168,11 @@ func_format_cb(struct workspace *wk, uint32_t node, void *_ctx, const struct str
 }
 
 static bool
-func_format(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_format(struct workspace *wk, obj self, obj *res)
 {
-	struct args_norm an[] = { { ARG_TYPE_GLOB | tc_number | tc_bool | tc_string | tc_file }, ARG_TYPE_NULL };
+	struct args_norm an[] = { { tc_message }, ARG_TYPE_NULL };
 
-	if (!interp_args(wk, args_node, an, NULL, NULL)) {
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
@@ -191,7 +181,7 @@ func_format(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 	};
 
 	obj str;
-	if (!string_format(wk, an[0].node, rcvr, &str, &ctx, func_format_cb)) {
+	if (!string_format(wk, an[0].node, self, &str, &ctx, func_format_cb)) {
 		return false;
 	}
 
@@ -200,21 +190,20 @@ func_format(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 }
 
 static bool
-func_underscorify(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_underscorify(struct workspace *wk, obj self, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
-	*res = str_clone(wk, wk, rcvr);
+	*res = str_clone_mutable(wk, self);
 
 	const struct str *ss = get_str(wk, *res);
 
 	uint32_t i;
 	for (i = 0; i < ss->len; ++i) {
-		if (!(('a' <= ss->s[i] && ss->s[i] <= 'z')
-		      || ('A' <= ss->s[i] && ss->s[i] <= 'Z')
-		      || ('0' <= ss->s[i] && ss->s[i] <= '9'))) {
+		if (!(('a' <= ss->s[i] && ss->s[i] <= 'z') || ('A' <= ss->s[i] && ss->s[i] <= 'Z')
+			    || ('0' <= ss->s[i] && ss->s[i] <= '9'))) {
 			((char *)ss->s)[i] = '_';
 		}
 	}
@@ -223,31 +212,41 @@ func_underscorify(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 }
 
 static bool
-func_split(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_split(struct workspace *wk, obj self, obj *res)
 {
-	struct args_norm ao[] = { { obj_string }, ARG_TYPE_NULL };
+	struct args_norm an[] = { { obj_string, .optional = true }, ARG_TYPE_NULL };
 
-	if (!interp_args(wk, args_node, NULL, ao, NULL)) {
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
-	const struct str *split = ao[0].set ? get_str(wk, ao[0].val) : NULL,
-			 *ss = get_str(wk, rcvr);
+	const struct str *split = an[0].set ? get_str(wk, an[0].val) : NULL, *ss = get_str(wk, self);
 
 	*res = str_split(wk, ss, split);
 	return true;
 }
 
 static bool
-func_join(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_splitlines(struct workspace *wk, obj self, obj *res)
 {
-	struct args_norm an[] = { { ARG_TYPE_GLOB | obj_string }, ARG_TYPE_NULL };
-
-	if (!interp_args(wk, args_node, an, NULL, NULL)) {
+	if (!pop_args(wk, 0, 0)) {
 		return false;
 	}
 
-	return obj_array_join(wk, true, an[0].val, rcvr, res);
+	*res = str_splitlines(wk, get_str(wk, self));
+	return true;
+}
+
+static bool
+func_join(struct workspace *wk, obj self, obj *res)
+{
+	struct args_norm an[] = { { TYPE_TAG_GLOB | obj_string }, ARG_TYPE_NULL };
+
+	if (!pop_args(wk, an, NULL)) {
+		return false;
+	}
+
+	return obj_array_join(wk, true, an[0].val, self, res);
 }
 
 struct version_compare_ctx {
@@ -271,14 +270,38 @@ version_compare_iter(struct workspace *wk, void *_ctx, obj s2)
 	};
 	enum op_type op = op_eq;
 
-	struct { const struct str name; enum op_type op; } ops[] = {
-		{ WKSTR(">="), op_ge, },
-		{ WKSTR(">"),  op_gt, },
-		{ WKSTR("=="), op_eq, },
-		{ WKSTR("!="), op_ne, },
-		{ WKSTR("<="), op_le, },
-		{ WKSTR("<"),  op_lt, },
-		{ WKSTR("="), op_eq, },
+	struct {
+		const struct str name;
+		enum op_type op;
+	} ops[] = {
+		{
+			WKSTR(">="),
+			op_ge,
+		},
+		{
+			WKSTR(">"),
+			op_gt,
+		},
+		{
+			WKSTR("=="),
+			op_eq,
+		},
+		{
+			WKSTR("!="),
+			op_ne,
+		},
+		{
+			WKSTR("<="),
+			op_le,
+		},
+		{
+			WKSTR("<"),
+			op_lt,
+		},
+		{
+			WKSTR("="),
+			op_eq,
+		},
 	};
 
 	uint32_t i;
@@ -294,24 +317,12 @@ version_compare_iter(struct workspace *wk, void *_ctx, obj s2)
 	int8_t cmp = rpmvercmp(ctx->ver1, &ver2);
 
 	switch (op) {
-	case op_eq:
-		ctx->res = cmp == 0;
-		break;
-	case op_ne:
-		ctx->res = cmp != 0;
-		break;
-	case op_gt:
-		ctx->res = cmp == 1;
-		break;
-	case op_ge:
-		ctx->res = cmp >= 0;
-		break;
-	case op_lt:
-		ctx->res = cmp == -1;
-		break;
-	case op_le:
-		ctx->res = cmp <= 0;
-		break;
+	case op_eq: ctx->res = cmp == 0; break;
+	case op_ne: ctx->res = cmp != 0; break;
+	case op_gt: ctx->res = cmp == 1; break;
+	case op_ge: ctx->res = cmp >= 0; break;
+	case op_lt: ctx->res = cmp == -1; break;
+	case op_le: ctx->res = cmp <= 0; break;
 	}
 
 	if (!ctx->res) {
@@ -338,40 +349,39 @@ version_compare(struct workspace *wk, uint32_t err_node, const struct str *ver,
 }
 
 static bool
-func_version_compare(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_version_compare(struct workspace *wk, obj self, obj *res)
 {
 	struct args_norm an[] = { { obj_string }, ARG_TYPE_NULL };
 
-	if (!interp_args(wk, args_node, an, NULL, NULL)) {
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
 	struct version_compare_ctx ctx = {
 		.err_node = an[0].node,
-		.ver1 = get_str(wk, rcvr),
+		.ver1 = get_str(wk, self),
 	};
 
 	if (version_compare_iter(wk, &ctx, an[0].val) == ir_err) {
 		return false;
 	}
 
-	make_obj(wk, res, obj_bool);
-	set_obj_bool(wk, *res, ctx.res);
+	*res = make_obj_bool(wk, ctx.res);
 	return true;
 }
 
 static bool
-func_string_to_int(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_string_to_int(struct workspace *wk, obj self, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
-	const struct str *ss = get_str(wk, rcvr);
+	const struct str *ss = get_str(wk, self);
 
 	int64_t n;
-	if (!str_to_i(ss, &n)) {
-		interp_error(wk, args_node, "unable to parse %o", rcvr);
+	if (!str_to_i(ss, &n, true)) {
+		vm_error(wk, "unable to parse %o", self);
 		return false;
 	}
 
@@ -381,48 +391,46 @@ func_string_to_int(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 }
 
 static bool
-func_string_startswith(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_string_startswith(struct workspace *wk, obj self, obj *res)
 {
 	struct args_norm an[] = { { obj_string }, ARG_TYPE_NULL };
-	if (!interp_args(wk, args_node, an, NULL, NULL)) {
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
-	make_obj(wk, res, obj_bool);
-	set_obj_bool(wk, *res, str_startswith(get_str(wk, rcvr), get_str(wk, an[0].val)));
+	*res = make_obj_bool(wk, str_startswith(get_str(wk, self), get_str(wk, an[0].val)));
 	return true;
 }
 
 static bool
-func_string_endswith(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_string_endswith(struct workspace *wk, obj self, obj *res)
 {
 	struct args_norm an[] = { { obj_string }, ARG_TYPE_NULL };
-	if (!interp_args(wk, args_node, an, NULL, NULL)) {
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
-	make_obj(wk, res, obj_bool);
-	set_obj_bool(wk, *res, str_endswith(get_str(wk, rcvr), get_str(wk, an[0].val)));
+	*res = make_obj_bool(wk, str_endswith(get_str(wk, self), get_str(wk, an[0].val)));
 	return true;
 }
 
 static bool
-func_string_substring(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_string_substring(struct workspace *wk, obj self, obj *res)
 {
-	struct args_norm ao[] = { { obj_number }, { obj_number }, ARG_TYPE_NULL };
-	if (!interp_args(wk, args_node, NULL, ao, NULL)) {
+	struct args_norm an[] = { { obj_number, .optional = true }, { obj_number, .optional = true }, ARG_TYPE_NULL };
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
-	const struct str *s = get_str(wk, rcvr);
+	const struct str *s = get_str(wk, self);
 	int64_t start = 0, end = s->len;
 
-	if (ao[0].set) {
-		start = get_obj_number(wk, ao[0].val);
+	if (an[0].set) {
+		start = get_obj_number(wk, an[0].val);
 	}
 
-	if (ao[1].set) {
-		end = get_obj_number(wk, ao[1].val);
+	if (an[1].set) {
+		end = get_obj_number(wk, an[1].val);
 	}
 
 	if (start < 0) {
@@ -433,33 +441,30 @@ func_string_substring(struct workspace *wk, obj rcvr, uint32_t args_node, obj *r
 		end = s->len + end;
 	}
 
-	if (start > s->len || start < 0) {
-		assert(ao[0].set);
-		interp_error(wk, ao[0].node, "substring start out of bounds");
-		return false;
-	} else if (end > s->len || end < 0) {
-		assert(ao[1].set);
-		interp_error(wk, ao[1].node, "substring end out of bounds");
-		return false;
-	} else if (end < start) {
-		assert(ao[1].set);
-		interp_error(wk, ao[1].node, "substring end before start");
-		return false;
+	if (end < start) {
+		end = start;
 	}
 
-	*res = make_strn(wk, &s->s[start], end - start);
+	start = MAX(0, start);
+
+	if (start > end || start > s->len) {
+		*res = make_str(wk, "");
+		return true;
+	}
+
+	*res = make_strn(wk, &s->s[start], MIN(end - start, s->len - start));
 	return true;
 }
 
 static bool
-func_string_replace(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_string_replace(struct workspace *wk, obj self, obj *res)
 {
 	struct args_norm an[] = { { obj_string }, { obj_string }, ARG_TYPE_NULL };
-	if (!interp_args(wk, args_node, an, NULL, NULL)) {
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
-	const struct str *s = get_str(wk, rcvr);
+	const struct str *s = get_str(wk, self);
 	const struct str *find = get_str(wk, an[0].val);
 	const struct str *replace = get_str(wk, an[1].val);
 	struct str tmp, pre = {
@@ -471,14 +476,14 @@ func_string_replace(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res
 
 	uint32_t i;
 	for (i = 0; i < s->len; ++i) {
-		tmp = (struct str) {
+		tmp = (struct str){
 			.s = &s->s[i],
 			.len = s->len - i,
 		};
 
 		if (str_startswith(&tmp, find)) {
-			str_appn(wk, *res, pre.s, pre.len);
-			str_appn(wk, *res, replace->s, replace->len);
+			str_appn(wk, res, pre.s, pre.len);
+			str_appn(wk, res, replace->s, replace->len);
 			i += find->len;
 			pre.s = &s->s[i];
 			pre.len = 0;
@@ -489,27 +494,27 @@ func_string_replace(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res
 		}
 	}
 
-	str_appn(wk, *res, pre.s, pre.len);
+	str_appn(wk, res, pre.s, pre.len);
 
 	return true;
 }
 
 static bool
-func_string_contains(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_string_contains(struct workspace *wk, obj self, obj *res)
 {
 	struct args_norm an[] = { { obj_string }, ARG_TYPE_NULL };
-	if (!interp_args(wk, args_node, an, NULL, NULL)) {
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
-	const struct str *s = get_str(wk, rcvr);
+	const struct str *s = get_str(wk, self);
 	const struct str *find = get_str(wk, an[0].val);
 	struct str tmp;
 
 	bool found = false;
 	uint32_t i;
 	for (i = 0; i < s->len; ++i) {
-		tmp = (struct str) {
+		tmp = (struct str){
 			.s = &s->s[i],
 			.len = s->len - i,
 		};
@@ -520,18 +525,18 @@ func_string_contains(struct workspace *wk, obj rcvr, uint32_t args_node, obj *re
 		}
 	}
 
-	make_obj(wk, res, obj_bool);
-	set_obj_bool(wk, *res, found);
+	*res = make_obj_bool(wk, found);
 	return true;
 }
 
-const struct func_impl_name impl_tbl_string[] = {
+const struct func_impl impl_tbl_string[] = {
 	{ "contains", func_string_contains, tc_bool, true },
 	{ "endswith", func_string_endswith, tc_bool, true },
 	{ "format", func_format, tc_string, true },
 	{ "join", func_join, tc_string, true },
 	{ "replace", func_string_replace, tc_string, true },
 	{ "split", func_split, tc_array, true },
+	{ "splitlines", func_splitlines, tc_array, true },
 	{ "startswith", func_string_startswith, tc_bool, true },
 	{ "strip", func_strip, tc_string, true },
 	{ "substring", func_string_substring, tc_string, true },
@@ -542,3 +547,34 @@ const struct func_impl_name impl_tbl_string[] = {
 	{ "version_compare", func_version_compare, tc_bool, true },
 	{ NULL, NULL },
 };
+
+static bool
+func_string_length(struct workspace *wk, obj self, obj *res)
+{
+	if (!pop_args(wk, 0, 0)) {
+		return false;
+	}
+
+	*res = make_number(wk, get_str(wk, self)->len);
+	return true;
+}
+
+const struct func_impl impl_tbl_string_internal[] = {
+	{ "contains", func_string_contains, tc_bool, true },
+	{ "endswith", func_string_endswith, tc_bool, true },
+	{ "format", func_format, tc_string, true },
+	{ "join", func_join, tc_string, true },
+	{ "replace", func_string_replace, tc_string, true },
+	{ "split", func_split, tc_array, true },
+	{ "splitlines", func_splitlines, tc_array, true },
+	{ "startswith", func_string_startswith, tc_bool, true },
+	{ "strip", func_strip, tc_string, true },
+	{ "substring", func_string_substring, tc_string, true },
+	{ "to_int", func_string_to_int, tc_number, true },
+	{ "to_lower", func_to_lower, tc_string, true },
+	{ "to_upper", func_to_upper, tc_string, true },
+	{ "underscorify", func_underscorify, tc_string, true },
+	{ "version_compare", func_version_compare, tc_bool, true },
+	{ "length", func_string_length, tc_number, true },
+	{ NULL, NULL },
+};
