@@ -5,17 +5,15 @@
 
 #include "compat.h"
 
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 #include "functions/xxx.h"
-#include "lang/interpreter.h"
-#include "log.h"
 
 static bool
-func_(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_(struct workspace *wk, obj self, obj *res)
 {
 }
 
-const struct func_impl_name impl_tbl_xxx[] = {
+const struct func_impl impl_tbl_xxx[] = {
 	{ "", func_ },
 	{ NULL, NULL },
 };
