@@ -6,8 +6,6 @@
 #include "compat.h"
 
 #include "guess.h"
-#include "log.h"
-#include "platform/assert.h"
 
 bool
 guess_version(struct workspace *wk, const char *src, obj *res)
