@@ -9,13 +9,18 @@
 #include <stdlib.h>
 #include <string.h>
 
+#include "buf_size.h"
+#include "error.h"
 #include "formats/editorconfig.h"
 #include "formats/ini.h"
 #include "lang/fmt.h"
 #include "lang/object_iterators.h"
+#include "lang/parser.h"
 #include "lang/string.h"
 #include "lang/typecheck.h"
-#include "memmem.h"
+#include "lang/workspace.h"
+#include "log.h"
+#include "platform/assert.h"
 #include "platform/mem.h"
 
 enum fmt_frag_flag {
@@ -231,7 +236,7 @@ fmt_write_frag_set_dbg_ws(struct fmt_ctx *f, const struct fmt_frag *pws, struct
 		if (ws->type == fmt_frag_type_ws_newline) {
 			log_plain("newline");
 		} else {
-			obj_fprintf(f->wk, log_file(), "# %o", ws->str);
+			obj_lprintf(f->wk, "# %o", ws->str);
 			if (ws->type == fmt_frag_type_ws_comment) {
 				log_plain(" comment");
 			} else if (ws->type == fmt_frag_type_ws_comment_trailing) {
@@ -264,15 +269,15 @@ fmt_write_frag_set_dbg(struct fmt_ctx *f, struct fmt_frag *p, const struct tree_
 	tree_indent_print(ti);
 
 	if (p->str) {
-		obj_fprintf(f->wk, log_file(), "%o", p->str);
+		obj_lprintf(f->wk, "%o", p->str);
 	} else if (p->type == fmt_frag_type_block) {
-		obj_fprintf(f->wk, log_file(), "block");
+		obj_lprintf(f->wk, "block");
 	} else if (p->type == fmt_frag_type_line) {
-		obj_fprintf(f->wk, log_file(), "line");
+		obj_lprintf(f->wk, "line");
 	} else if (p->enclosing) {
-		obj_fprintf(f->wk, log_file(), "%s", p->enclosing);
+		obj_lprintf(f->wk, "%s", p->enclosing);
 	} else {
-		obj_fprintf(f->wk, log_file(), "?");
+		obj_lprintf(f->wk, "?");
 	}
 
 	if (p->flags) {
@@ -755,7 +760,6 @@ fmt_node_to_token(enum node_type type)
 	case node_type_break: return "break";
 	case node_type_return: return "return";
 	case node_type_assign: return "=";
-	case node_type_plusassign: return "+=";
 	case node_type_or: return "or";
 	case node_type_and: return "and";
 	case node_type_add: return "+";
@@ -872,7 +876,8 @@ fmt_list(struct fmt_ctx *f, struct node *n, struct fmt_frag *fr, enum fmt_list_f
 					bool prev_is_arg_like = str_startswith(s, &WKSTR("-")) && !strchr(s->s, '='),
 					     cur_is_arg_like = cur && str_startswith(get_str(f->wk, cur), &WKSTR("-"));
 
-					if (prev_is_arg_like && !cur_is_arg_like) {
+					if (prev_is_arg_like && !cur_is_arg_like
+						&& !(child->pre_ws && child->pre_ws->type == fmt_frag_type_ws_comment_trailing)) {
 						child->flags |= fmt_frag_flag_stick_line_left;
 					}
 				}
@@ -974,6 +979,7 @@ fmt_node(struct fmt_ctx *f, struct node *n)
 	/* L("formatting %p:%s", (void *)n, node_to_s(f->wk, n)); */
 
 	switch (n->type) {
+	case node_type_maybe_id:
 	case node_type_stringify:
 	case node_type_stmt: UNREACHABLE;
 
@@ -1043,7 +1049,6 @@ fmt_node(struct fmt_ctx *f, struct node *n)
 		break;
 	}
 	case node_type_assign:
-	case node_type_plusassign:
 	case node_type_or:
 	case node_type_and:
 	case node_type_add:
@@ -1059,9 +1064,36 @@ fmt_node(struct fmt_ctx *f, struct node *n)
 	case node_type_gt:
 	case node_type_leq:
 	case node_type_geq: {
-		res = fr = fmt_node(f, n->l);
+		const char *token;
+		if (n->type == node_type_assign && (n->data.type & op_store_flag_add_store)) {
+			token = "+=";
+		} else {
+			token = fmt_node_to_token(n->type);
+		}
+		bool is_member_assign = n->type == node_type_assign && (n->data.type & op_store_flag_member);
 
-		next = fmt_frag_sibling(fr, fmt_frag_s(f, fmt_node_to_token(n->type)));
+		struct node *rhs;
+
+		if (is_member_assign) {
+			res = fr = fmt_node(f, n->r->l);
+			if (n->l->type == node_type_id_lit) {
+				next = fmt_frag_sibling(fr, fmt_frag_s(f, "."));
+				next->flags |= fmt_frag_flag_stick_left;
+				next = fmt_frag_sibling(fr, fmt_node(f, n->l));
+				next->flags |= fmt_frag_flag_stick_left;
+			} else {
+				next = fmt_frag_sibling(fr, fmt_frag(f, fmt_frag_type_expr));
+				next->enclosing = "[]";
+				next->flags |= fmt_frag_flag_stick_left;
+				fmt_frag_child(&next->child, fmt_node(f, n->l));
+			}
+			rhs = n->r->r;
+		} else {
+			res = fr = fmt_node(f, n->l);
+			rhs = n->r;
+		}
+
+		next = fmt_frag_sibling(fr, fmt_frag_s(f, token));
 		switch (n->type) {
 		case node_type_add:
 		case node_type_or:
@@ -1075,7 +1107,7 @@ fmt_node(struct fmt_ctx *f, struct node *n)
 		}
 		}
 
-		next = fmt_frag_sibling(fr, fmt_node(f, n->r));
+		next = fmt_frag_sibling(fr, fmt_node(f, rhs));
 		next->flags |= fmt_frag_flag_stick_line_left;
 		break;
 	}
@@ -1110,16 +1142,16 @@ fmt_node(struct fmt_ctx *f, struct node *n)
 		fmt_list(f, n, fr, 0);
 		break;
 	}
-	case node_type_method: {
-		res = fr = fmt_node(f, n->l->r);
+	case node_type_member: {
+		res = fr = fmt_node(f, n->l);
 		next = fmt_frag_sibling(fr, fmt_frag_s(f, "."));
 		next->flags |= fmt_frag_flag_stick_left;
 		next = fmt_frag_sibling(fr, fmt_node(f, n->r));
 		next->flags |= fmt_frag_flag_stick_left;
-		next = fmt_frag_sibling(fr, fmt_frag(f, fmt_frag_type_expr));
-		fmt_list(f, n->l->l, next, fmt_list_flag_func_args);
-		next->enclosing = "()";
-		next->flags |= fmt_frag_flag_stick_left;
+
+		/* fmt_list(f, n->l->l, next, fmt_list_flag_func_args); */
+		/* next->enclosing = "()"; */
+		/* next->flags |= fmt_frag_flag_stick_left; */
 		break;
 	}
 	case node_type_call: {
@@ -1264,8 +1296,8 @@ fmt_special_function_done:
 		res->type = fmt_frag_type_lines;
 		fr = fmt_frag_child(&res->child, fmt_frag(f, fmt_frag_type_line));
 		fr = fmt_frag_child(&fr->child, fmt_frag_s(f, "func"));
-		if (n->l->l) {
-			next = fmt_frag_sibling(fr, fmt_node(f, n->l->l));
+		if (n->l->l->l) {
+			next = fmt_frag_sibling(fr, fmt_node(f, n->l->l->l));
 			next->flags |= fmt_frag_flag_stick_line_left;
 		}
 		next = fmt_frag_sibling(fr, fmt_frag(f, fmt_frag_type_expr));
