@@ -7,11 +7,14 @@
 
 #include <errno.h>
 #include <inttypes.h>
+#include <stdarg.h>
 #include <stdlib.h>
 #include <string.h>
 
 #include "buf_size.h"
+#include "error.h"
 #include "lang/lexer.h"
+#include "platform/assert.h"
 
 /******************************************************************************
 * token printing
@@ -71,6 +74,7 @@ token_type_to_s(enum token_type type)
 	case token_type_return: return "return";
 	case token_type_bitor: return "|";
 	case token_type_returntype: return "->";
+	case token_type_doc_comment: return "doc comment";
 	}
 
 	UNREACHABLE_RETURN;
@@ -163,6 +167,7 @@ lex_advance_n(struct lexer *lexer, uint32_t n)
 struct lex_str_token_table {
 	struct str str;
 	int32_t token_type;
+	int32_t token_subtype;
 };
 
 static bool
@@ -178,6 +183,7 @@ lex_str_token_lookup(struct lexer *lexer,
 		if (str_eql(&table[i].str, str)) {
 			token->type = table[i].token_type;
 			token->location.len = table[i].str.len;
+			token->data.type = table[i].token_subtype;
 			return true;
 		}
 	}
@@ -451,6 +457,38 @@ lex_string_escape(struct lexer *lexer, struct token *token, struct sbuf *buf)
 	UNREACHABLE_RETURN;
 }
 
+typedef bool(lex_string_escape_fun)(struct lexer *lexer, struct token *token, struct sbuf *buf);
+
+static void
+lex_basic_string(struct lexer *lexer, struct token *token, struct sbuf *buf, char end, lex_string_escape_fun escape)
+{
+	lex_advance(lexer);
+
+	for (; lexer->i < lexer->source->len && lexer->src[lexer->i] != end; lex_advance(lexer)) {
+		switch (lexer->src[lexer->i]) {
+		case 0:
+		case '\n': goto unterminated_string;
+		case '\\': {
+			if (!lex_string_escape(lexer, token, buf)) {
+				return;
+			}
+			break;
+		}
+		default: sbuf_push(lexer->wk, buf, lexer->src[lexer->i]); break;
+		}
+	}
+
+	if (lexer->src[lexer->i] != end) {
+unterminated_string:
+		lex_error_token(lexer, token, "unterminated string");
+		return;
+	}
+
+	lex_advance(lexer);
+
+	token->data.str = sbuf_into_str(lexer->wk, buf);
+}
+
 static void
 lex_string(struct lexer *lexer, struct token *token)
 {
@@ -459,15 +497,16 @@ lex_string(struct lexer *lexer, struct token *token)
 
 	if (str_eql(&lexer_str(multiline_terminator.len), &multiline_terminator)) {
 		lex_advance_n(lexer, multiline_terminator.len);
-		uint32_t start = lexer->i;
 
 		while (lexer->source->len - lexer->i >= multiline_terminator.len
 			&& !str_eql(&lexer_str(multiline_terminator.len), &multiline_terminator)) {
+			if (lexer->src[lexer->i] != '\r') {
+				sbuf_push(lexer->wk, &buf, lexer->src[lexer->i]);
+			}
 			lex_advance(lexer);
 		}
 
 		if (str_eql(&lexer_str(multiline_terminator.len), &multiline_terminator)) {
-			sbuf_pushn(lexer->wk, &buf, &lexer->src[start], lexer->i - start);
 			lex_advance_n(lexer, 3);
 			token->data.str = sbuf_into_str(lexer->wk, &buf);
 		} else {
@@ -477,31 +516,7 @@ lex_string(struct lexer *lexer, struct token *token)
 		return;
 	}
 
-	lex_advance(lexer);
-
-	for (; lexer->i < lexer->source->len && lexer->src[lexer->i] != '\''; lex_advance(lexer)) {
-		switch (lexer->src[lexer->i]) {
-		case 0:
-		case '\n': goto unterminated_string;
-		case '\\': {
-			if (!lex_string_escape(lexer, token, &buf)) {
-				return;
-			}
-			break;
-		}
-		default: sbuf_push(lexer->wk, &buf, lexer->src[lexer->i]); break;
-		}
-	}
-
-	if (lexer->src[lexer->i] != '\'') {
-unterminated_string:
-		lex_error_token(lexer, token, "unterminated string");
-		return;
-	}
-
-	lex_advance(lexer);
-
-	token->data.str = sbuf_into_str(lexer->wk, &buf);
+	lex_basic_string(lexer, token, &buf, '\'', lex_string_escape);
 }
 
 enum lexer_enclosed_state {
@@ -563,13 +578,61 @@ restart:
 		return;
 	}
 
+	if (lexer->mode & lexer_mode_bom_error) {
+		lex_error_token(lexer, token, "Unexpected utf-8 BOM.  File must be in utf-8 with no BOM.");
+		lexer->mode &= ~lexer_mode_bom_error;
+		lexer->i += 3;
+		return;
+	}
+
 	while (is_skipchar(lexer->src[lexer->i])) {
 		if (lexer->src[lexer->i] == '#') {
 			lex_advance(lexer);
 
+			obj doc_comment = 0;
+			if (!(lexer->mode & lexer_mode_fmt) && (lexer->mode & lexer_mode_functions)
+				&& lexer->src[lexer->i] == '#') {
+				lex_advance(lexer);
+				if (strchr(" \t", lexer->src[lexer->i])) {
+					lex_advance(lexer);
+				}
+				doc_comment = make_str(lexer->wk, "");
+			}
+
 			start = lexer->i;
 
-			while (lexer->src[lexer->i] && lexer->src[lexer->i] != '\n') {
+			while (lexer->src[lexer->i]) {
+				if (doc_comment) {
+					if (lexer->src[lexer->i] == '\n') {
+						uint32_t skip = 1;
+						while (strchr(" \t", lexer->src[lexer->i + skip])) {
+							++skip;
+						}
+
+						if (str_startswith(&WKSTR(&lexer->src[lexer->i + skip]), &WKSTR("##"))) {
+							skip += 2;
+
+							uint32_t i;
+							for (i = 0; i < skip; ++i) {
+								lex_advance(lexer);
+							}
+
+							str_app(lexer->wk, &doc_comment, "\n");
+
+							if (strchr(" \t", lexer->src[lexer->i])) {
+								lex_advance(lexer);
+							}
+							continue;
+						} else {
+							break;
+						}
+					} else {
+						str_appn(lexer->wk, &doc_comment, &lexer->src[lexer->i], 1);
+					}
+				} else if (lexer->src[lexer->i] == '\n') {
+					break;
+				}
+
 				lex_advance(lexer);
 			}
 
@@ -596,6 +659,11 @@ restart:
 						}
 					}
 				}
+			} else if (doc_comment) {
+				token->location.off = start;
+				token->type = token_type_doc_comment;
+				token->data.str = doc_comment;
+				return;
 			}
 		} else {
 			lex_advance(lexer);
@@ -740,6 +808,10 @@ lexer_init(struct lexer *lexer, struct workspace *wk, struct source *src, enum l
 		.mode = mode,
 	};
 
+	if (src->len >= 3 && memcmp(src->src, (uint8_t []){ 0xef, 0xbb, 0xbf }, 3) == 0) {
+		lexer->mode |= lexer_mode_bom_error;
+	}
+
 	stack_init(&lexer->stack, 2048);
 
 	if (lexer->mode & lexer_mode_fmt) {
@@ -776,3 +848,200 @@ lexer_is_fmt_comment(const struct str *comment, bool *fmt_on)
 
 	return false;
 }
+
+/******************************************************************************
+* cmake
+******************************************************************************/
+
+static const struct lex_str_token_table cm_lex_keyword_tokens_command[] = {
+	{ WKSTR_STATIC("else"), token_type_else },
+	{ WKSTR_STATIC("elseif"), token_type_elif },
+	{ WKSTR_STATIC("endif"), token_type_endif },
+	{ WKSTR_STATIC("if"), token_type_if },
+};
+
+static const struct lex_str_token_table cm_lex_keyword_tokens_conditional[] = {
+	{ WKSTR_STATIC("NOT"), token_type_not },
+	{ WKSTR_STATIC("AND"), token_type_and },
+	{ WKSTR_STATIC("OR"), token_type_or },
+	{ WKSTR_STATIC("EQUAL"), token_type_eq },
+	{ WKSTR_STATIC("LESS"), '<' },
+	{ WKSTR_STATIC("LESS_EQUAL"), token_type_leq },
+	{ WKSTR_STATIC("GREATER"), '>' },
+	{ WKSTR_STATIC("GREATER_EQUAL"), token_type_geq },
+	{ WKSTR_STATIC("STREQUAL"), token_type_eq, cm_token_subtype_comp_str },
+	{ WKSTR_STATIC("STRLESS"), '<', cm_token_subtype_comp_str },
+	{ WKSTR_STATIC("STRLESS_EQUAL"), token_type_leq, cm_token_subtype_comp_str },
+	{ WKSTR_STATIC("STRGREATER"), '>', cm_token_subtype_comp_str },
+	{ WKSTR_STATIC("STRGREATER_EQUAL"), token_type_geq, cm_token_subtype_comp_str },
+	{ WKSTR_STATIC("VERSION_EQUAL"), token_type_eq, cm_token_subtype_comp_ver },
+	{ WKSTR_STATIC("VERSION_LESS"), '<', cm_token_subtype_comp_ver },
+	{ WKSTR_STATIC("VERSION_LESS_EQUAL"), token_type_leq, cm_token_subtype_comp_ver },
+	{ WKSTR_STATIC("VERSION_GREATER"), '>', cm_token_subtype_comp_ver },
+	{ WKSTR_STATIC("VERSION_GREATER_EQUAL"), token_type_geq, cm_token_subtype_comp_ver },
+	{ WKSTR_STATIC("PATH_EQUAL"), token_type_eq, cm_token_subtype_comp_path },
+	{ WKSTR_STATIC("MATCHES"), token_type_eq, cm_token_subtype_comp_regex },
+};
+
+void
+cm_lexer_next(struct lexer *lexer, struct token *token)
+{
+	uint32_t start;
+
+restart:
+	*token = (struct token){
+		.location = (struct source_location){ .off = lexer->i, .len = 1 },
+	};
+
+	if (lexer->i >= lexer->source->len) {
+		token->type = token_type_eof;
+		return;
+	}
+
+	while (is_skipchar(lexer->src[lexer->i])) {
+		if (lexer->src[lexer->i] == '#') {
+			lex_advance(lexer);
+
+			start = lexer->i;
+
+			while (lexer->src[lexer->i] && lexer->src[lexer->i] != '\n') {
+				lex_advance(lexer);
+			}
+
+			if (lexer->mode & lexer_mode_fmt) {
+				bool fmt_on;
+				obj s;
+
+				s = make_strn(lexer->wk, &lexer->src[start], lexer->i - start);
+				s = str_strip(lexer->wk, get_str(lexer->wk, s), 0, 0);
+				if (lexer_is_fmt_comment(get_str(lexer->wk, s), &fmt_on)) {
+					if (fmt_on) {
+						if (lexer->fmt.in_raw_block) {
+							s = make_strn(lexer->wk,
+								&lexer->src[lexer->fmt.raw_block_start],
+								(start - 1) - lexer->fmt.raw_block_start);
+
+							obj_array_push(lexer->wk, lexer->fmt.raw_blocks, s);
+							lexer->fmt.in_raw_block = false;
+						}
+					} else {
+						if (!lexer->fmt.in_raw_block) {
+							lexer->fmt.raw_block_start = lexer->i;
+							lexer->fmt.in_raw_block = true;
+						}
+					}
+				}
+			}
+		} else {
+			lex_advance(lexer);
+		}
+	}
+
+	token->location.off = lexer->i;
+
+	/* if (is_valid_start_of_identifier(lexer->src[lexer->i])) { */
+	/* 	start = lexer->i; */
+	/* 	struct str str = { &lexer->src[lexer->i] }; */
+
+	/* 	while (is_valid_inside_of_identifier(lexer->src[lexer->i])) { */
+	/* 		lex_advance(lexer); */
+	/* 		++str.len; */
+	/* 	} */
+
+	/* 	token->type = token_type_identifier; */
+	/* 	lex_copy_str(lexer, token, start, lexer->i); */
+	/* 	return; */
+	/* } */
+
+	if (!strchr("()#\" \r\n\t", lexer->src[lexer->i])) {
+		start = lexer->i;
+		struct str str = { &lexer->src[lexer->i] };
+
+		while (!strchr("()#\" \r\n\t", lexer->src[lexer->i])) {
+			lex_advance(lexer);
+			++str.len;
+		}
+
+		token->type = token_type_string;
+
+		if (is_valid_start_of_identifier(lexer->src[start])) {
+			uint32_t i;
+			for (i = 1; i < str.len; ++i) {
+				if (!is_valid_inside_of_identifier(lexer->src[start + i])) {
+					break;
+				}
+			}
+
+			if (i == str.len) {
+				token->type = token_type_identifier;
+			}
+		}
+
+		if (token->type == token_type_identifier) {
+			const struct lex_str_token_table *token_table = 0;
+			uint32_t token_table_len = 0;
+
+			switch (lexer->cm_mode) {
+			case cm_lexer_mode_default: break;
+			case cm_lexer_mode_command:
+				token_table = cm_lex_keyword_tokens_command;
+				token_table_len = ARRAY_LEN(cm_lex_keyword_tokens_command);
+				break;
+			case cm_lexer_mode_conditional:
+				token_table = cm_lex_keyword_tokens_conditional;
+				token_table_len = ARRAY_LEN(cm_lex_keyword_tokens_conditional);
+				break;
+			}
+
+			if (token_table && lex_str_token_lookup(lexer, token, token_table, token_table_len, &str)) {
+				return;
+			}
+		}
+
+		lex_copy_str(lexer, token, start, lexer->i);
+		return;
+	}
+
+	switch (lexer->src[lexer->i]) {
+	case '\n':
+		lex_advance(lexer);
+
+		if (lexer->enclosed_state) {
+			goto restart;
+		} else {
+			token->type = token_type_eol;
+		}
+		return;
+	case '"': {
+		start = lexer->i;
+		token->type = token_type_string;
+
+		SBUF(buf);
+		lex_basic_string(lexer, token, &buf, '"', lex_string_escape);
+		token->location.len = lexer->i - token->location.off;
+		return;
+	}
+	case '(':
+		token->type = lexer->src[lexer->i];
+		lexer_push_pop_enclosed_state(lexer, token->type);
+		break;
+	case ')':
+		token->type = lexer->src[lexer->i];
+		lexer_push_pop_enclosed_state(lexer, token->type);
+		break;
+	case '\0':
+		if (lexer->i != lexer->source->len) {
+			goto unexpected_character;
+		}
+
+		token->type = token_type_eof;
+		break;
+	default:
+unexpected_character:
+		lex_error_token(lexer, token, "unexpected character: '%c'", lexer->src[lexer->i]);
+		break;
+	}
+
+	lex_advance(lexer);
+	return;
+}
