@@ -16,13 +16,15 @@
 #include "error.h"
 #include "lang/object.h"
 #include "lang/object_iterators.h"
-#include "lang/parser.h"
 #include "lang/typecheck.h"
 #include "log.h"
 #include "options.h"
+#include "platform/assert.h"
 #include "platform/mem.h"
 #include "tracy.h"
 
+static bool making_default_objects = false;
+
 static void *
 get_obj_internal(struct workspace *wk, obj id, enum obj_type type)
 {
@@ -34,7 +36,14 @@ get_obj_internal(struct workspace *wk, obj id, enum obj_type type)
 	}
 
 	switch (type) {
+	case obj_null:
+	case obj_disabler:
+	case obj_meson:
 	case obj_bool:
+		if (!making_default_objects) {
+			error_unrecoverable("tried to get singleton object of type %s", obj_type_to_s(type));
+		}
+		// fallthrough
 	case obj_file:
 	case obj_feature_opt:
 	case obj_machine: {
@@ -72,14 +81,36 @@ get_obj_internal(struct workspace *wk, obj id, enum obj_type type)
 	case obj_func:
 	case obj_capture: return bucket_arr_get(&wk->vm.objects.obj_aos[o->t - _obj_aos_start], o->val);
 
-	case obj_null:
-	case obj_meson:
-	case obj_disabler: LOG_E("tried to get singleton object of type %s", obj_type_to_s(type)); abort();
-
 	default: assert(false && "tried to get invalid object type"); return NULL;
 	}
 }
 
+void
+make_default_objects(struct workspace *wk)
+{
+	making_default_objects = true;
+
+	obj id;
+	make_obj(wk, &id, obj_null);
+	assert(id == 0);
+
+	make_obj(wk, &id, obj_disabler);
+	assert(id == obj_disabler);
+
+	make_obj(wk, &id, obj_meson);
+	assert(id == obj_meson);
+
+	make_obj(wk, &id, obj_bool);
+	assert(id == obj_bool_true);
+	*(bool *)get_obj_internal(wk, id, obj_bool) = true;
+
+	make_obj(wk, &id, obj_bool);
+	assert(id == obj_bool_false);
+	*(bool *)get_obj_internal(wk, id, obj_bool) = false;
+
+	making_default_objects = false;
+}
+
 enum obj_type
 get_obj_type(struct workspace *wk, obj id)
 {
@@ -90,7 +121,27 @@ get_obj_type(struct workspace *wk, obj id)
 bool
 get_obj_bool(struct workspace *wk, obj o)
 {
-	return *(bool *)get_obj_internal(wk, o, obj_bool);
+	if (o == obj_bool_true) {
+		return true;
+	} else if (o == obj_bool_false) {
+		return false;
+	} else {
+		UNREACHABLE;
+	}
+
+	/* return *(bool *)get_obj_internal(wk, o, obj_bool); */
+}
+
+obj
+make_obj_bool(struct workspace *wk, bool v)
+{
+	return v ? obj_bool_true : obj_bool_false;
+}
+
+obj
+get_obj_bool_with_default(struct workspace *wk, obj o, bool def)
+{
+	return o ? get_obj_bool(wk, o) : def;
 }
 
 obj
@@ -108,12 +159,6 @@ get_obj_number(struct workspace *wk, obj o)
 	return *(int64_t *)get_obj_internal(wk, o, obj_number);
 }
 
-void
-set_obj_bool(struct workspace *wk, obj o, bool v)
-{
-	*(bool *)get_obj_internal(wk, o, obj_bool) = v;
-}
-
 void
 set_obj_number(struct workspace *wk, obj o, int64_t v)
 {
@@ -220,11 +265,15 @@ make_obj(struct workspace *wk, obj *id, enum obj_type type)
 	*id = wk->vm.objects.objs.len;
 
 	switch (type) {
-	case obj_bool:
-	case obj_file:
 	case obj_null:
-	case obj_meson:
 	case obj_disabler:
+	case obj_meson:
+	case obj_bool:
+		if (!making_default_objects) {
+			UNREACHABLE;
+		}
+		// fallthrough
+	case obj_file:
 	case obj_machine:
 	case obj_feature_opt: val = 0; break;
 
@@ -326,7 +375,7 @@ static struct {
 	enum obj_type t;
 	const char *name;
 } obj_names[obj_type_count] = {
-	{ .t = obj_null, .name = "void" },
+	{ .t = obj_null, .name = "null" },
 	{ .t = obj_compiler, .name = "compiler" },
 	{ .t = obj_dependency, .name = "dep" },
 	{ .t = obj_meson, .name = "meson" },
@@ -550,100 +599,87 @@ obj_equal(struct workspace *wk, obj left, obj right)
 	}
 }
 
-/*
+/*******************************************************************************
  * arrays
- */
+ ******************************************************************************/
+
+static void
+obj_array_copy_on_write(struct workspace *wk, struct obj_array *a, obj arr)
+{
+	struct obj_array cur;
+
+	if (!(a->flags & obj_array_flag_cow)) {
+		return;
+	}
+
+	cur = *a;
+	*a = (struct obj_array){ 0 };
+
+	obj v;
+	obj_array_for_array(wk, &cur, v) {
+		obj_array_push(wk, arr, v);
+	}
+}
 
 bool
 obj_array_foreach(struct workspace *wk, obj arr, void *ctx, obj_array_iterator cb)
 {
-	struct obj_array *a = get_obj_array(wk, arr);
-
-	if (!a->len) {
-		return true;
-	}
-
-	while (true) {
-		switch (cb(wk, ctx, a->val)) {
+	obj v;
+	obj_array_for(wk, arr, v) {
+		switch (cb(wk, ctx, v)) {
 		case ir_cont: break;
 		case ir_done: return true;
 		case ir_err: return false;
 		}
-
-		if (!a->have_next) {
-			break;
-		}
-
-		a = get_obj_array(wk, a->next);
 	}
 
 	return true;
 }
 
-struct obj_array_foreach_flat_ctx {
-	void *usr_ctx;
-	obj_array_iterator cb;
-};
-
-static enum iteration_result
-obj_array_foreach_flat_iter(struct workspace *wk, void *_ctx, obj val)
-{
-	struct obj_array_foreach_flat_ctx *ctx = _ctx;
-
-	if (get_obj_type(wk, val) == obj_array) {
-		if (!obj_array_foreach(wk, val, ctx, obj_array_foreach_flat_iter)) {
-			return ir_err;
-		} else {
-			return ir_cont;
-		}
-	} else if (get_obj_type(wk, val) == obj_typeinfo && get_obj_typeinfo(wk, val)->type == tc_array) {
-		// skip typeinfo arrays as they wouldn't be yielded if they
-		// were real arrays
-		return ir_cont;
-	} else {
-		return ctx->cb(wk, ctx->usr_ctx, val);
-	}
-
-	return ir_cont;
-}
-
 bool
 obj_array_foreach_flat(struct workspace *wk, obj arr, void *usr_ctx, obj_array_iterator cb)
 {
-	struct obj_array_foreach_flat_ctx ctx = {
-		.usr_ctx = usr_ctx,
-		.cb = cb,
-	};
+	obj v;
+	obj_array_flat_for_(wk, arr, v, iter) {
+		switch (cb(wk, usr_ctx, v)) {
+		case ir_cont: break;
+		case ir_done: {
+			obj_array_flat_iter_end(wk, &iter);
+			return true;
+		}
+		case ir_err: {
+			obj_array_flat_iter_end(wk, &iter);
+			return false;
+		}
+		}
+	}
 
-	return obj_array_foreach(wk, arr, &ctx, obj_array_foreach_flat_iter);
+	return true;
 }
 
 void
 obj_array_push(struct workspace *wk, obj arr, obj child)
 {
-	obj child_arr;
-	struct obj_array *a, *tail, *c;
-
-	if (!(a = get_obj_array(wk, arr))->len) {
-		a->tail = arr;
-		a->len = 1;
-		a->val = child;
-		a->have_next = false;
-		return;
-	}
-
-	make_obj(wk, &child_arr, obj_array);
-	c = get_obj_array(wk, child_arr);
-	c->val = child;
+	struct obj_array_elem *e;
+	struct obj_array *a;
 
 	a = get_obj_array(wk, arr);
-	tail = get_obj_array(wk, a->tail);
-	assert(!tail->have_next);
+	obj_array_copy_on_write(wk, a, arr);
 
-	tail->have_next = true;
-	tail->next = child_arr;
+	uint32_t next = wk->vm.objects.array_elems.len;
 
-	a->tail = child_arr;
+	if (!a->len) {
+		a->head = next;
+	}
+
+	bucket_arr_push(&wk->vm.objects.array_elems, &(struct obj_array_elem){ .val = child });
+
+	if (a->len) {
+		e = bucket_arr_get(&wk->vm.objects.array_elems, a->tail);
+		e->next = next;
+	}
+
+	a->tail = next;
 	++a->len;
 }
 
@@ -657,34 +693,20 @@ obj_array_prepend(struct workspace *wk, obj *arr, obj val)
 	*arr = prepend;
 }
 
-struct obj_array_index_of_iter_ctx {
-	obj l;
-	bool res;
-	uint32_t i;
-};
-
-static enum iteration_result
-obj_array_index_of_iter(struct workspace *wk, void *_ctx, obj v)
-{
-	struct obj_array_index_of_iter_ctx *ctx = _ctx;
-
-	if (obj_equal(wk, ctx->l, v)) {
-		ctx->res = true;
-		return ir_done;
-	}
-
-	++ctx->i;
-	return ir_cont;
-}
-
 bool
 obj_array_index_of(struct workspace *wk, obj arr, obj val, uint32_t *idx)
 {
-	struct obj_array_index_of_iter_ctx ctx = { .l = val };
-	obj_array_foreach(wk, arr, &ctx, obj_array_index_of_iter);
+	obj v;
+	uint32_t i = 0;
+	obj_array_for(wk, arr, v) {
+		if (obj_equal(wk, val, v)) {
+			*idx = i;
+			return true;
+		}
+		++i;
+	}
 
-	*idx = ctx.i;
-	return ctx.res;
+	return false;
 }
 
 bool
@@ -694,83 +716,109 @@ obj_array_in(struct workspace *wk, obj arr, obj val)
 	return obj_array_index_of(wk, arr, val, &_);
 }
 
-struct obj_array_index_iter_ctx {
-	obj res, i, tgt;
-};
-
-static enum iteration_result
-obj_array_index_iter(struct workspace *wk, void *_ctx, obj v)
+static obj *
+obj_array_index_pointer_raw(struct workspace *wk, obj arr, int64_t i)
 {
-	struct obj_array_index_iter_ctx *ctx = _ctx;
-
-	if (ctx->i == ctx->tgt) {
-		ctx->res = v;
-		return ir_done;
+	obj v;
+	int64_t j = 0;
+	obj_array_for_(wk, arr, v, iter) {
+		(void)v;
+		if (j == i) {
+			return &iter.e->val;
+		}
+		++j;
 	}
 
-	++ctx->i;
-	return ir_cont;
+	return 0;
+}
+
+obj *
+obj_array_index_pointer(struct workspace *wk, obj arr, int64_t i)
+{
+	obj_array_copy_on_write(wk, get_obj_array(wk, arr), arr);
+	return obj_array_index_pointer_raw(wk, arr, i);
 }
 
 void
 obj_array_index(struct workspace *wk, obj arr, int64_t i, obj *res)
 {
-	struct obj_array_index_iter_ctx ctx = { .tgt = i };
-	assert(i >= 0 && i < get_obj_array(wk, arr)->len);
-	obj_array_foreach(wk, arr, &ctx, obj_array_index_iter);
-	*res = ctx.res;
+	obj *a = obj_array_index_pointer_raw(wk, arr, i);
+	assert(a);
+	*res = *a;
 }
 
 obj
 obj_array_get_tail(struct workspace *wk, obj arr)
 {
-	return get_obj_array(wk, get_obj_array(wk, arr)->tail)->val;
+	uint32_t tail = get_obj_array(wk, arr)->tail;
+	struct obj_array_elem *e = bucket_arr_get(&wk->vm.objects.array_elems, tail);
+	return e->val;
 }
 
-struct obj_array_dup_ctx {
-	obj *arr;
-};
-
-static enum iteration_result
-obj_array_dup_iter(struct workspace *wk, void *_ctx, obj v)
+obj
+obj_array_get_head(struct workspace *wk, obj arr)
 {
-	struct obj_array_dup_ctx *ctx = _ctx;
-	obj_array_push(wk, *ctx->arr, v);
-	return ir_cont;
+	uint32_t head = get_obj_array(wk, arr)->head;
+	struct obj_array_elem *e = bucket_arr_get(&wk->vm.objects.array_elems, head);
+	return e->val;
 }
 
 void
 obj_array_dup(struct workspace *wk, obj arr, obj *res)
 {
-	struct obj_array_dup_ctx ctx = { .arr = res };
 	make_obj(wk, res, obj_array);
-	obj_array_foreach(wk, arr, &ctx, obj_array_dup_iter);
+	obj v;
+	obj_array_for(wk, arr, v) {
+		obj_array_push(wk, *res, v);
+	}
 }
 
+static void
+obj_array_dup_into_light(struct workspace *wk, obj src, obj dst)
+{
+	struct obj_array *a_dst = get_obj_array(wk, dst), *a_src = get_obj_array(wk, src);
+	*a_dst = *a_src;
+	a_dst->flags |= obj_array_flag_cow;
+	a_src->flags |= obj_array_flag_cow;
+}
+
+obj
+obj_array_dup_light(struct workspace *wk, obj src)
+{
+	obj res;
+	make_obj(wk, &res, obj_array);
+	obj_array_dup_into_light(wk, src, res);
+	return res;
+}
+
+// mutates arr and consumes arr2
 void
 obj_array_extend_nodup(struct workspace *wk, obj arr, obj arr2)
 {
-	struct obj_array *a, *b, *tail;
+	struct obj_array *a, *b;
+	struct obj_array_elem *tail;
 
 	if (!(b = get_obj_array(wk, arr2))->len) {
 		return;
 	}
 
-	if (!(a = get_obj_array(wk, arr))->len) {
-		struct obj_array_dup_ctx ctx = { .arr = &arr };
-		obj_array_foreach(wk, arr2, &ctx, obj_array_dup_iter);
+	a = get_obj_array(wk, arr);
+	obj_array_copy_on_write(wk, a, arr);
+
+	if (!a->len) {
+		obj_array_dup_into_light(wk, arr2, arr);
 		return;
 	}
 
-	tail = get_obj_array(wk, a->tail);
-	assert(!tail->have_next);
-	tail->have_next = true;
-	tail->next = arr2;
+	tail = bucket_arr_get(&wk->vm.objects.array_elems, a->tail);
+	assert(!tail->next);
+	tail->next = b->head;
 
 	a->tail = b->tail;
 	a->len += b->len;
 }
 
+// mutates arr without modifying arr2
 void
 obj_array_extend(struct workspace *wk, obj arr, obj arr2)
 {
@@ -851,82 +899,74 @@ obj_array_tail(struct workspace *wk, obj arr, obj *res)
 {
 	const struct obj_array *a = get_obj_array(wk, arr);
 
-	if (a->have_next) {
-		struct obj_array *n = get_obj_array(wk, a->next);
+	make_obj(wk, res, obj_array);
+
+	if (a->len > 1) {
+		struct obj_array *n = get_obj_array(wk, *res);
+		struct obj_array_elem *head = bucket_arr_get(&wk->vm.objects.array_elems, a->head);
+
+		n->head = head->next;
 		n->tail = a->tail;
 		n->len = a->len;
-		*res = a->next;
-	} else {
-		// the tail of a zero or single element array is an empty array
-		make_obj(wk, res, obj_array);
 	}
 }
 
 void
 obj_array_set(struct workspace *wk, obj arr, int64_t i, obj v)
 {
-	assert(i >= 0 && i < get_obj_array(wk, arr)->len);
-
-	uint32_t j = 0;
-
-	while (true) {
-		if (j == i) {
-			get_obj_array(wk, arr)->val = v;
-			return;
-		}
-
-		assert(get_obj_array(wk, arr)->have_next);
-		arr = get_obj_array(wk, arr)->next;
-		++j;
-	}
-
-	assert(false && "unreachable");
+	obj *p = obj_array_index_pointer(wk, arr, i);
+	assert(p);
+	*p = v;
 }
 
 void
 obj_array_del(struct workspace *wk, obj arr, int64_t i)
 {
-	uint32_t j = 0;
-	obj p = arr;
-	struct obj_array *head = get_obj_array(wk, arr), *prev, *next, *del;
+	struct obj_array *a = get_obj_array(wk, arr);
+	obj_array_copy_on_write(wk, a, arr);
 
-	assert(i >= 0 && i < head->len);
+	struct obj_array_elem *head = bucket_arr_get(&wk->vm.objects.array_elems, a->head), *prev = 0;
+	uint32_t head_idx = a->head, prev_idx = 0;
 
+	assert(i >= 0 && i < a->len);
+
+#if 0
 	if (i == 0) {
-		if (head->have_next) {
-			next = get_obj_array(wk, head->next);
-			next->len = head->len - 1;
-			next->tail = head->tail;
-			*head = *next;
+		if (a->len >= 1) {
+			head = bucket_arr_get(&wk->vm.objects.array_elems, a->head);
+			a->head = head->next;
+			--a->len;
 		} else {
-			*head = (struct obj_array){ 0 };
+			*a = (struct obj_array){ 0 };
 		}
 
 		return;
 	}
+#endif
 
+	int64_t j = 0;
 	while (true) {
 		if (j == i) {
-			del = get_obj_array(wk, arr);
 			break;
 		}
 
-		p = arr;
-		assert(get_obj_array(wk, arr)->have_next);
-		arr = get_obj_array(wk, arr)->next;
+		prev_idx = head_idx;
+		prev = head;
+		head_idx = head->next;
+		head = bucket_arr_get(&wk->vm.objects.array_elems, head_idx);
 		++j;
 	}
 
-	prev = get_obj_array(wk, p);
-
-	if (del->have_next) {
-		prev->next = del->next;
+	if (i == 0) {
+		a->head = head->next;
+	} else if (i == a->len - 1) {
+		a->tail = prev_idx;
+		prev->next = 0;
 	} else {
-		head->tail = p;
-		prev->have_next = false;
+		prev->next = head->next;
 	}
 
-	--head->len;
+	--a->len;
 }
 
 obj
@@ -937,6 +977,13 @@ obj_array_pop(struct workspace *wk, obj arr)
 	return t;
 }
 
+void
+obj_array_clear(struct workspace *wk, obj arr)
+{
+	struct obj_array *a = get_obj_array(wk, arr);
+	*a = (struct obj_array){ 0 };
+}
+
 static enum iteration_result
 obj_array_dedup_iter(struct workspace *wk, void *_ctx, obj val)
 {
@@ -1058,130 +1105,107 @@ obj_array_sort(struct workspace *wk, void *usr_ctx, obj arr, obj_array_sort_func
 	arr_destroy(&da);
 }
 
-struct obj_array_slice_ctx {
-	int64_t i, i0, i1;
-	obj res;
-};
-
-static enum iteration_result
-obj_array_slice_iter(struct workspace *wk, void *_ctx, obj v)
-{
-	struct obj_array_slice_ctx *ctx = _ctx;
-
-	if (ctx->i >= ctx->i0) {
-		obj_array_push(wk, ctx->res, v);
-	}
-
-	++ctx->i;
-
-	if (ctx->i > ctx->i1) {
-		return ir_done;
-	}
-
-	return ir_cont;
-}
-
 obj
-obj_array_slice(struct workspace *wk, obj a, int64_t i0, int64_t i1)
+obj_array_slice(struct workspace *wk, obj a, int64_t start, int64_t end)
 {
-	struct obj_array *arr = get_obj_array(wk, a);
-	if (!(bounds_adjust(arr->len, &i0) && bounds_adjust(arr->len, &i1))) {
-		assert(false && "index out of bounds");
+	struct obj_array *src = get_obj_array(wk, a);
+
+	obj res;
+	make_obj(wk, &res, obj_array);
+	struct obj_array *dst = get_obj_array(wk, res);
+
+	if (start == end) {
+		// empty slice
+		return res;
 	}
 
-	struct obj_array_slice_ctx ctx = {
-		.i0 = i0,
-		.i1 = i1,
-	};
+	const bool tail_slice = end == src->len;
+	uint32_t prev_elem = src->head;
 
-	make_obj(wk, &ctx.res, obj_array);
+	obj v;
+	obj_array_for_array_(wk, src, v, iter) {
+		if (iter.i >= end) {
+			break;
+		}
 
-	obj_array_foreach(wk, a, &ctx, obj_array_slice_iter);
+		if (iter.i >= start) {
+			if (tail_slice) {
+				src->flags |= obj_array_flag_cow;
 
-	return ctx.res;
+				dst->len = src->len - start;
+				dst->head = prev_elem;
+				dst->tail = src->tail;
+				dst->flags |= obj_array_flag_cow;
+				return res;
+			}
+
+			obj_array_push(wk, res, v);
+		}
+
+		prev_elem = iter.e->next;
+	}
+
+	return res;
 }
 
-/*
+/*******************************************************************************
  * dictionaries
- */
+ ******************************************************************************/
 
 bool
 obj_dict_foreach(struct workspace *wk, obj dict, void *ctx, obj_dict_iterator cb)
 {
-	struct obj_dict *d = get_obj_dict(wk, dict);
-	if (!d->len) {
-		return true;
-	}
-
-	if (d->flags & obj_dict_flag_big) {
-		uint32_t i;
-		struct hash *h = bucket_arr_get(&wk->vm.objects.dict_hashes, d->data);
-		for (i = 0; i < h->keys.len; ++i) {
-			void *_key = arr_get(&h->keys, i);
-			uint64_t *_val = hash_get(h, _key);
-			obj key = *_val >> 32;
-			obj val = *_val & 0xffffffff;
-
-			switch (cb(wk, ctx, key, val)) {
-			case ir_cont: break;
-			case ir_done: return true;
-			case ir_err: return false;
-			}
-		}
-	} else {
-		struct obj_dict_elem *e = bucket_arr_get(&wk->vm.objects.dict_elems, d->data);
-
-		while (true) {
-			switch (cb(wk, ctx, e->key, e->val)) {
-			case ir_cont: break;
-			case ir_done: return true;
-			case ir_err: return false;
-			}
-
-			if (!e->next) {
-				break;
-			}
-			e = bucket_arr_get(&wk->vm.objects.dict_elems, e->next);
+	obj k, v;
+	obj_dict_for(wk, dict, k, v) {
+		switch (cb(wk, ctx, k, v)) {
+		case ir_cont: break;
+		case ir_done: return true;
+		case ir_err: return false;
 		}
 	}
 
 	return true;
 }
 
-struct obj_dict_dup_ctx {
-	obj *dict;
-};
-
-static enum iteration_result
-obj_dict_dup_iter(struct workspace *wk, void *_ctx, obj key, obj val)
-{
-	struct obj_dict_dup_ctx *ctx = _ctx;
-	obj_dict_set(wk, *ctx->dict, key, val);
-	return ir_cont;
-}
-
 void
 obj_dict_dup(struct workspace *wk, obj dict, obj *res)
 {
-	struct obj_dict_dup_ctx ctx = { .dict = res };
 	make_obj(wk, res, obj_dict);
-	obj_dict_foreach(wk, dict, &ctx, obj_dict_dup_iter);
+
+	obj k, v;
+	obj_dict_for(wk, dict, k, v) {
+		obj_dict_set(wk, *res, k, v);
+	}
 }
 
-static enum iteration_result
-obj_dict_merge_iter(struct workspace *wk, void *_ctx, obj key, obj val)
+void
+obj_dict_dup_light(struct workspace *wk, obj dict, obj *res)
 {
-	obj *res = _ctx;
-
-	obj_dict_set(wk, *res, key, val);
-
-	return ir_cont;
+	make_obj(wk, res, obj_dict);
+	struct obj_dict *new = get_obj_dict(wk, *res), *cur = get_obj_dict(wk, dict);
+	*new = *cur;
+	// TODO: We have to set cow on both dicts because we don't know which one
+	// will be modified.
+	// Ideally, we would unset the one dict's cow flag when the other one gets
+	// written to, but that would require refcounting.
+	//
+	// a = {}
+	// b = a
+	// c = a
+	//
+	// Now a, b, and c all point at the same underlying dict.  We'd need to
+	// track the references and only remove cow when there is a single reference.
+	cur->flags |= obj_dict_flag_cow;
+	new->flags |= obj_dict_flag_cow;
 }
 
 void
 obj_dict_merge_nodup(struct workspace *wk, obj dict, obj dict2)
 {
-	obj_dict_foreach(wk, dict2, &dict, obj_dict_merge_iter);
+	obj k, v;
+	obj_dict_for(wk, dict2, k, v) {
+		obj_dict_set(wk, dict, k, v);
+	}
 }
 
 void
@@ -1219,9 +1243,7 @@ _obj_dict_index(struct workspace *wk,
 	obj dict,
 	union obj_dict_key_comparison_key *key,
 	obj_dict_key_comparison_func comp,
-	obj **res,
-	uint64_t **ures)
-
+	obj **res)
 {
 	struct obj_dict *d = get_obj_dict(wk, dict);
 	if (!d->len) {
@@ -1230,13 +1252,17 @@ _obj_dict_index(struct workspace *wk,
 
 	if (d->flags & obj_dict_flag_big) {
 		struct hash *h = bucket_arr_get(&wk->vm.objects.dict_hashes, d->data);
+		uint64_t *uv;
+
 		if (d->flags & obj_dict_flag_int_key) {
-			*ures = hash_get(h, &key->num);
+			uv = hash_get(h, &key->num);
 		} else {
-			*ures = hash_get_strn(h, key->string.s, key->string.len);
+			uv = hash_get_strn(h, key->string.s, key->string.len);
 		}
 
-		if (*ures) {
+		if (uv) {
+			union obj_dict_big_dict_value *val = (union obj_dict_big_dict_value *)uv;
+			*res = &val->val.val;
 			return true;
 		}
 	} else {
@@ -1258,23 +1284,54 @@ _obj_dict_index(struct workspace *wk,
 	return false;
 }
 
+static obj *
+obj_dict_index_strn_pointer_raw(struct workspace *wk, obj dict, const char *str, uint32_t len)
+{
+	obj *r = 0;
+	union obj_dict_key_comparison_key key = { .string = { .s = str, .len = len } };
+
+	if (!_obj_dict_index(wk, dict, &key, obj_dict_key_comparison_func_string, &r)) {
+		return 0;
+	}
+
+	return r;
+}
+
+static void
+obj_dict_copy_on_write(struct workspace *wk, obj dict)
+{
+	struct obj_dict *d = get_obj_dict(wk, dict), cur;
+
+	if (!(d->flags & obj_dict_flag_cow)) {
+		return;
+	}
+
+	cur = *d;
+	*d = (struct obj_dict){ 0 };
+
+	obj k, v;
+	obj_dict_for_dict(wk, &cur, k, v)
+	{
+		obj_dict_set(wk, dict, k, v);
+	}
+}
+
+obj *
+obj_dict_index_strn_pointer(struct workspace *wk, obj dict, const char *str, uint32_t len)
+{
+	obj_dict_copy_on_write(wk, dict);
+	return obj_dict_index_strn_pointer_raw(wk, dict, str, len);
+}
+
 bool
 obj_dict_index_strn(struct workspace *wk, obj dict, const char *str, uint32_t len, obj *res)
 {
-	uint64_t *ur = 0;
-	obj *r = 0;
-	union obj_dict_key_comparison_key key = { .string = {
-							  .s = str,
-							  .len = len,
-						  } };
-
-	if (!_obj_dict_index(wk, dict, &key, obj_dict_key_comparison_func_string, &r, &ur)) {
-		return false;
+	obj *r = obj_dict_index_strn_pointer_raw(wk, dict, str, len);
+	if (r) {
+		*res = *r;
+		return true;
 	}
-
-	*res = r ? *r : (*ur & 0xffffffff);
-
-	return true;
+	return false;
 }
 
 bool
@@ -1298,13 +1355,15 @@ obj_dict_in(struct workspace *wk, obj dict, obj key)
 }
 
 static void
-_obj_dict_set(struct workspace *wk,
+obj_dict_set_impl(struct workspace *wk,
 	obj dict,
 	union obj_dict_key_comparison_key *k,
 	obj_dict_key_comparison_func comp,
 	obj key,
 	obj val)
 {
+	obj_dict_copy_on_write(wk, dict);
+
 	struct obj_dict *d = get_obj_dict(wk, dict);
 
 	assert(key);
@@ -1332,15 +1391,13 @@ _obj_dict_set(struct workspace *wk,
 		d->tail = 0; // unnecessary but nice
 
 		while (true) {
-			uint64_t uv = ((uint64_t)(e->key) << 32) | e->val;
+			union obj_dict_big_dict_value val = { .val = { .key = e->key, .val = e->val } };
 
 			if (d->flags & obj_dict_flag_int_key) {
-				hash_set(h, &key, uv);
+				hash_set(h, &e->key, val.u64);
 			} else {
 				const struct str *ss = get_str(wk, e->key);
-				/* LO("setting %s, %d to %ld, (%o=%o)\n", ss->s, ss->len, uv, (obj)(uv
-         * >> 32), (obj)(uv & 0xffffffff)); */
-				hash_set_strn(h, ss->s, ss->len, uv);
+				hash_set_strn(h, ss->s, ss->len, val.u64);
 			}
 
 			if (!e->next) {
@@ -1351,14 +1408,9 @@ _obj_dict_set(struct workspace *wk,
 		d->flags |= obj_dict_flag_big;
 	}
 
-	uint64_t *ur;
 	obj *r = 0;
-	if (_obj_dict_index(wk, dict, k, comp, &r, &ur)) {
-		if (r) {
-			*r = val;
-		} else {
-			*ur = ((uint64_t)key << 32) | val;
-		}
+	if (_obj_dict_index(wk, dict, k, comp, &r)) {
+		*r = val;
 		return;
 	}
 
@@ -1369,7 +1421,8 @@ _obj_dict_set(struct workspace *wk,
 			hash_set(h, &key, val);
 		} else {
 			const struct str *ss = get_str(wk, key);
-			hash_set_strn(h, ss->s, ss->len, ((uint64_t)key << 32) | val);
+			union obj_dict_big_dict_value big_val = { .val = { .key = key, .val = val } };
+			hash_set_strn(h, ss->s, ss->len, big_val.u64);
 		}
 		d->len = h->len;
 	} else {
@@ -1394,12 +1447,14 @@ obj_dict_set(struct workspace *wk, obj dict, obj key, obj val)
 	union obj_dict_key_comparison_key k = {
 		.string = *get_str(wk, key),
 	};
-	_obj_dict_set(wk, dict, &k, obj_dict_key_comparison_func_string, key, val);
+	obj_dict_set_impl(wk, dict, &k, obj_dict_key_comparison_func_string, key, val);
 }
 
 static void
 _obj_dict_del(struct workspace *wk, obj dict, union obj_dict_key_comparison_key *key, obj_dict_key_comparison_func comp)
 {
+	obj_dict_copy_on_write(wk, dict);
+
 	struct obj_dict *d = get_obj_dict(wk, dict);
 	if (!d->len) {
 		return;
@@ -1482,28 +1537,30 @@ void
 obj_dict_seti(struct workspace *wk, obj dict, uint32_t key, obj val)
 {
 	union obj_dict_key_comparison_key k = { .num = key };
-	_obj_dict_set(wk, dict, &k, obj_dict_key_comparison_func_int, key, val);
+	struct obj_dict *d = get_obj_dict(wk, dict);
+	d->flags |= obj_dict_flag_int_key;
+	obj_dict_set_impl(wk, dict, &k, obj_dict_key_comparison_func_int, key, val);
 }
 
 bool
 obj_dict_geti(struct workspace *wk, obj dict, uint32_t key, obj *val)
 {
-	uint64_t *ur = 0;
 	obj *r = 0;
 	if (_obj_dict_index(wk,
 		    dict,
 		    &(union obj_dict_key_comparison_key){ .num = key },
 		    obj_dict_key_comparison_func_int,
-		    &r,
-		    &ur)) {
-		*val = r ? *r : (*ur & 0xffffffff);
+		    &r)) {
+		*val = *r;
 		return true;
 	}
 
 	return false;
 }
 
-/* */
+/*******************************************************************************
+ * obj_iterable_foreach
+ ******************************************************************************/
 
 struct obj_iterable_foreach_ctx {
 	void *ctx;
@@ -1593,10 +1650,8 @@ obj_clone(struct workspace *wk_src, struct workspace *wk_dest, obj val, obj *ret
 		make_obj(wk_dest, ret, t);
 		set_obj_number(wk_dest, *ret, get_obj_number(wk_src, val));
 		return true;
-	case obj_bool:
-		make_obj(wk_dest, ret, t);
-		set_obj_bool(wk_dest, *ret, get_obj_bool(wk_src, val));
-		return true;
+	case obj_disabler: *ret = val; return true;
+	case obj_bool: *ret = val; return true;
 	case obj_string: {
 		*ret = str_clone(wk_src, wk_dest, val);
 		return true;
@@ -1846,7 +1901,7 @@ static void
 obj_to_s_str(struct workspace *wk, struct obj_to_s_ctx *ctx, obj s)
 {
 	sbuf_push(wk, ctx->sb, '\'');
-	str_unescape(wk, ctx->sb, get_str(wk, s), true);
+	str_escape(wk, ctx->sb, get_str(wk, s), true);
 	sbuf_push(wk, ctx->sb, '\'');
 }
 
@@ -1866,16 +1921,24 @@ obj_to_s_opts(struct workspace *wk, obj o, struct sbuf *sb, struct obj_to_s_opts
 	}
 	case obj_dependency: {
 		struct obj_dependency *dep = get_obj_dependency(wk, o);
-		sbuf_pushs(wk, sb, "<dependency ");
+		sbuf_pushs(wk, sb, "<dep ");
 		if (dep->name) {
 			obj_to_s_str(wk, &ctx, dep->name);
 		}
 
-		sbuf_pushf(wk,
-			sb,
-			" | found: %s, pkgconf: %s>",
-			dep->flags & dep_flag_found ? "yes" : "no",
-			dep->type == dependency_type_pkgconf ? "yes" : "no");
+		const char *type = 0;
+		switch (dep->type) {
+		case dependency_type_declared: type = "declared"; break;
+		case dependency_type_pkgconf: type = "pkgconf"; break;
+		case dependency_type_threads: type = "threads"; break;
+		case dependency_type_external_library: type = "external_library"; break;
+		case dependency_type_appleframeworks: type = "appleframeworks"; break;
+		case dependency_type_not_found: type = "not_found"; break;
+		}
+
+		const char *found = dep->flags & dep_flag_found ? " found" : "";
+
+		sbuf_pushf(wk, sb, " %s machine:%s%s>", type, machine_kind_to_s(dep->machine), found);
 		break;
 	}
 	case obj_alias_target:
@@ -2021,7 +2084,7 @@ obj_to_s_opts(struct workspace *wk, obj o, struct sbuf *sb, struct obj_to_s_opts
 	}
 	case obj_typeinfo: {
 		struct obj_typeinfo *ti = get_obj_typeinfo(wk, o);
-		sbuf_pushs(wk, sb, "<typeinfo: ");
+		sbuf_pushf(wk, sb, "<typeinfo 0x4%x: ", o);
 		sbuf_pushs(wk, sb, typechecking_type_to_s(wk, ti->type));
 		sbuf_pushs(wk, sb, ">");
 		break;
@@ -2051,7 +2114,7 @@ obj_to_s(struct workspace *wk, obj o, struct sbuf *sb)
 		sbuf_pushf(wk, sb, fmt_buf, arg);                              \
 	}
 
-bool
+static bool
 obj_vasprintf(struct workspace *wk, struct sbuf *sb, const char *fmt, va_list ap)
 {
 	const char *fmt_start;
@@ -2132,7 +2195,7 @@ obj_vasprintf(struct workspace *wk, struct sbuf *sb, const char *fmt, va_list ap
 			if (*fmt == 'o') {
 				obj o = va_arg(ap, unsigned int);
 				if (get_obj_type(wk, o) == obj_string && got_hash) {
-					str_unescape(wk, sb, get_str(wk, o), false);
+					str_escape(wk, sb, get_str(wk, o), false);
 				} else {
 					struct obj_to_s_opts opts = { 0 };
 
@@ -2221,16 +2284,6 @@ obj_vasprintf(struct workspace *wk, struct sbuf *sb, const char *fmt, va_list ap
 	return true;
 }
 
-bool
-obj_asprintf(struct workspace *wk, struct sbuf *sb, const char *fmt, ...)
-{
-	va_list ap;
-	va_start(ap, fmt);
-	bool ret = obj_vasprintf(wk, sb, fmt, ap);
-	va_end(ap);
-	return ret;
-}
-
 uint32_t
 obj_vsnprintf(struct workspace *wk, char *buf, uint32_t len, const char *fmt, va_list ap)
 {
@@ -2255,10 +2308,10 @@ obj_snprintf(struct workspace *wk, char *buf, uint32_t len, const char *fmt, ...
 	return ret;
 }
 
-bool
+static bool
 obj_vfprintf(struct workspace *wk, FILE *f, const char *fmt, va_list ap)
 {
-	struct sbuf sb = { .flags = sbuf_flag_write, .buf = (void *)f };
+	SBUF_FILE(sb, f);
 
 	return obj_vasprintf(wk, &sb, fmt, ap);
 }
@@ -2283,66 +2336,156 @@ obj_printf(struct workspace *wk, const char *fmt, ...)
 	return ret;
 }
 
+static bool
+obj_vlprintf(struct workspace *wk, const char *fmt, va_list ap)
+{
+	FILE *f = _log_file();
+	struct sbuf *buf = _log_sbuf();
+	SBUF_FILE(sb, f);
+
+	if (f) {
+		buf = &sb;
+	}
+
+	return obj_vasprintf(wk, buf, fmt, ap);
+}
+
+bool
+obj_lprintf(struct workspace *wk, const char *fmt, ...)
+{
+	va_list ap;
+	va_start(ap, fmt);
+	bool ret = obj_vlprintf(wk, fmt, ap);
+	va_end(ap);
+	return ret;
+}
+
 /*
  * inspect - obj_to_s + more detail for some objects
  */
 
 static void
-obj_inspect_dep(struct workspace *wk, FILE *out, const char *pre, struct build_dep *dep)
+obj_inspect_dep(struct workspace *wk, const char *pre, struct build_dep *dep)
 {
-	obj_fprintf(wk, out, "%slink_language: %s\n", pre, compiler_language_to_s(dep->link_language));
-	obj_fprintf(wk, out, "%slink_whole: %o\n", pre, dep->link_whole);
-	obj_fprintf(wk, out, "%slink_with: %o\n", pre, dep->link_with);
-	obj_fprintf(wk, out, "%slink_with_not_found: %o\n", pre, dep->link_with_not_found);
-	obj_fprintf(wk, out, "%slink_args: %o\n", pre, dep->link_args);
-	obj_fprintf(wk, out, "%scompile_args: %o\n", pre, dep->compile_args);
-	obj_fprintf(wk, out, "%sinclude_directories: %o\n", pre, dep->include_directories);
-	obj_fprintf(wk, out, "%ssources: %o\n", pre, dep->sources);
-	obj_fprintf(wk, out, "%sobjects: %o\n", pre, dep->objects);
-	obj_fprintf(wk, out, "%sorder_deps: %o\n", pre, dep->order_deps);
-	obj_fprintf(wk, out, "%srpath: %o\n", pre, dep->rpath);
+	obj_lprintf(wk, "%slink_language: %s\n", pre, compiler_language_to_s(dep->link_language));
+	obj_lprintf(wk, "%slink_whole: %o\n", pre, dep->link_whole);
+	obj_lprintf(wk, "%slink_with: %o\n", pre, dep->link_with);
+	obj_lprintf(wk, "%slink_with_not_found: %o\n", pre, dep->link_with_not_found);
+	obj_lprintf(wk, "%slink_args: %o\n", pre, dep->link_args);
+	obj_lprintf(wk, "%scompile_args: %o\n", pre, dep->compile_args);
+	obj_lprintf(wk, "%sinclude_directories: %o\n", pre, dep->include_directories);
+	obj_lprintf(wk, "%ssources: %o\n", pre, dep->sources);
+	obj_lprintf(wk, "%sobjects: %o\n", pre, dep->objects);
+	obj_lprintf(wk, "%sorder_deps: %o\n", pre, dep->order_deps);
+	obj_lprintf(wk, "%srpath: %o\n", pre, dep->rpath);
 }
 
 void
-obj_inspect(struct workspace *wk, FILE *out, obj val)
+obj_inspect(struct workspace *wk, obj val)
 {
 	switch (get_obj_type(wk, val)) {
 	case obj_build_target: {
 		struct obj_build_target *tgt = get_obj_build_target(wk, val);
 
-		fprintf(out, "build_target:\n");
+		log_plain("build_target:\n");
 		if (tgt->name) {
-			obj_fprintf(wk, out, "    name: %o,\n", tgt->name);
+			obj_lprintf(wk, "    name: %o,\n", tgt->name);
 		}
-		obj_fprintf(wk, out, "    dep:\n");
-		obj_inspect_dep(wk, out, "        ", &tgt->dep);
-		obj_fprintf(wk, out, "    dep_internal:\n");
-		obj_inspect_dep(wk, out, "        ", &tgt->dep_internal);
+		obj_lprintf(wk, "    dep:\n");
+		obj_inspect_dep(wk, "        ", &tgt->dep);
+		obj_lprintf(wk, "    dep_internal:\n");
+		obj_inspect_dep(wk, "        ", &tgt->dep_internal);
 		break;
 	}
 	case obj_dependency: {
 		struct obj_dependency *dep = get_obj_dependency(wk, val);
 
-		fprintf(out, "dependency:\n");
+		log_plain("dependency:\n");
 
-		obj_fprintf(wk, out, "    found: %s\n", (dep->flags & dep_flag_found) ? "yes" : "no");
+		obj_lprintf(wk, "    found: %s\n", (dep->flags & dep_flag_found) ? "yes" : "no");
 
 		if (dep->name) {
-			obj_fprintf(wk, out, "    name: %o\n", dep->name);
+			obj_lprintf(wk, "    name: %o\n", dep->name);
 		}
 		if (dep->version) {
-			obj_fprintf(wk, out, "    version: %o\n", dep->version);
+			obj_lprintf(wk, "    version: %o\n", dep->version);
 		}
 		if (dep->variables) {
-			obj_fprintf(wk, out, "    variables: '%o'\n", dep->variables);
+			obj_lprintf(wk, "    variables: '%o'\n", dep->variables);
 		}
 
-		obj_fprintf(wk, out, "    type: %d\n", dep->type);
-		obj_fprintf(wk, out, "    dep:\n");
+		obj_lprintf(wk, "    type: %d\n", dep->type);
+		obj_lprintf(wk, "    dep:\n");
 
-		obj_inspect_dep(wk, out, "        ", &dep->dep);
+		obj_inspect_dep(wk, "        ", &dep->dep);
 		break;
 	}
-	default: obj_fprintf(wk, out, "%o\n", val);
+	default: obj_lprintf(wk, "%o\n", val);
+	}
+}
+
+/*
+ * object to json
+ */
+
+void
+obj_to_json(struct workspace *wk, obj o, struct sbuf *sb)
+{
+	switch (get_obj_type(wk, o)) {
+	case obj_array: {
+		sbuf_push(wk, sb, '[');
+		obj v;
+		obj_array_for_(wk, o, v, iter)
+		{
+			obj_to_json(wk, v, sb);
+			if (iter.i < iter.len - 1) {
+				sbuf_pushs(wk, sb, ", ");
+			}
+		}
+		sbuf_push(wk, sb, ']');
+		break;
+	}
+	case obj_dict: {
+		sbuf_push(wk, sb, '{');
+		uint32_t i = 0, len = get_obj_dict(wk, o)->len;
+		obj k, v;
+		obj_dict_for(wk, o, k, v) {
+			obj_to_json(wk, k, sb);
+			sbuf_pushs(wk, sb, ": ");
+			obj_to_json(wk, v, sb);
+
+			if (i < len - 1) {
+				sbuf_pushs(wk, sb, ", ");
+			}
+			++i;
+		}
+		sbuf_push(wk, sb, '}');
+		break;
+	}
+	case obj_number: sbuf_pushf(wk, sb, "%" PRId64, get_obj_number(wk, o)); break;
+	case obj_bool: sbuf_pushs(wk, sb, get_obj_bool(wk, o) ? "true" : "false"); break;
+	case obj_file:
+		o = *get_obj_file(wk, o);
+		/* fallthrough */
+	case obj_string: {
+		sbuf_push(wk, sb, '\"');
+		str_escape_json(wk, sb, get_str(wk, o));
+		sbuf_push(wk, sb, '\"');
+		break;
+	}
+	case obj_feature_opt: {
+		const char *s = (char *[]){
+			[feature_opt_enabled] = "enabled",
+			[feature_opt_disabled] = "disabled",
+			[feature_opt_auto] = "auto",
+		}[get_obj_feature_opt(wk, o)];
+		sbuf_pushf(wk, sb, "\"%s\"", s);
+		break;
+	}
+	case obj_null: {
+		sbuf_pushs(wk, sb, "null");
+		break;
+	}
+	default: error_unrecoverable("unable to convert %s to json", obj_type_to_s(get_obj_type(wk, o))); break;
 	}
 }
