@@ -6,18 +6,41 @@
 #include "compat.h"
 
 #include <inttypes.h>
+#include <stdarg.h>
 #include <string.h>
 
 #include "buf_size.h"
+#include "error.h"
 #include "lang/lexer.h"
 #include "lang/parser.h"
 #include "lang/typecheck.h"
+#include "lang/vm.h"
+#include "lang/workspace.h"
+#include "platform/assert.h"
 #include "tracy.h"
 
 /******************************************************************************
  * parser
  ******************************************************************************/
 
+struct parser;
+
+enum parse_stmt_flags {
+	parse_stmt_flag_eol_optional = 1 << 0,
+};
+
+struct parse_behavior {
+	void (*advance)(struct parser *p);
+	struct node *(*parse_stmt)(struct parser *p, enum parse_stmt_flags flags);
+	struct node *(*parse_list)(struct parser *p, enum node_type t, enum token_type end);
+};
+
+enum cm_parse_mode {
+	cm_parse_mode_command,
+	cm_parse_mode_command_args,
+	cm_parse_mode_conditional,
+};
+
 struct parser {
 	struct token previous, current, next;
 	struct lexer lexer;
@@ -26,7 +49,9 @@ struct parser {
 	struct source *src;
 	struct bucket_arr *nodes;
 	enum vm_compile_mode mode;
+	enum cm_parse_mode cm_mode;
 	uint32_t inside_loop;
+	obj doc_comment;
 
 	struct {
 		char msg[2048];
@@ -38,6 +63,8 @@ struct parser {
 	struct {
 		struct node_fmt previous, current;
 	} fmt;
+
+	struct parse_behavior behavior;
 };
 
 enum parse_precedence {
@@ -54,8 +81,8 @@ enum parse_precedence {
 	parse_precedence_primary,
 };
 
-typedef struct node *((*parse_prefix_fn)(struct parser *p));
-typedef struct node *((*parse_infix_fn)(struct parser *p, struct node *l));
+typedef struct node *((*parse_prefix_fn)(struct parser *p, bool assignment_allowed));
+typedef struct node *((*parse_infix_fn)(struct parser *p, struct node *l, bool assignment_allowed));
 
 struct parse_rule {
 	parse_prefix_fn prefix;
@@ -65,7 +92,7 @@ struct parse_rule {
 
 static struct node *parse_prec(struct parser *p, enum parse_precedence prec);
 static struct node *parse_expr(struct parser *p);
-static struct node *parse_block(struct parser *p, enum token_type types[], uint32_t types_len);
+static struct node *parse_block(struct parser *p, enum token_type types[], uint32_t types_len, enum parse_stmt_flags);
 
 /*******************************************************************************
  * misc api functions
@@ -108,10 +135,9 @@ node_type_to_s(enum node_type t)
 		nt(mod);
 		nt(not );
 		nt(index);
-		nt(method);
+		nt(member);
 		nt(call);
 		nt(assign);
-		nt(plusassign);
 		nt(foreach_args);
 		nt(foreach);
 		nt(if);
@@ -122,6 +148,7 @@ node_type_to_s(enum node_type t)
 		nt(func_def);
 		nt(return);
 		nt(group);
+		nt(maybe_id);
 	}
 
 	UNREACHABLE_RETURN;
@@ -138,6 +165,7 @@ fmt_node_to_s(struct workspace *wk, const struct node *n)
 	i += snprintf(&buf[i], BUF_SIZE_S - i, "%s", node_type_to_s(n->type));
 
 	switch (n->type) {
+	case node_type_maybe_id:
 	case node_type_id:
 	case node_type_id_lit:
 	case node_type_number:
@@ -160,6 +188,7 @@ node_to_s(struct workspace *wk, const struct node *n)
 	i += snprintf(&buf[i], BUF_SIZE_S - i, "%s", node_type_to_s(n->type));
 
 	switch (n->type) {
+	case node_type_maybe_id:
 	case node_type_id:
 	case node_type_id_lit:
 	case node_type_string: i += obj_snprintf(wk, &buf[i], BUF_SIZE_S - i, ":%o", n->data.str); break;
@@ -299,6 +328,11 @@ parse_advance(struct parser *p)
 	p->fmt.previous = p->fmt.current;
 	lexer_next(&p->lexer, &p->current);
 
+	if (p->current.type == token_type_doc_comment) {
+		p->doc_comment = p->current.data.str;
+		lexer_next(&p->lexer, &p->current);
+	}
+
 	while (p->current.type == token_type_error) {
 		parse_error(p, &p->current.location, "%s", get_cstr(p->wk, p->current.data.str));
 		p->err.unwinding = false;
@@ -362,7 +396,7 @@ static bool
 parse_accept(struct parser *p, enum token_type type)
 {
 	if (p->current.type == type) {
-		parse_advance(p);
+		p->behavior.advance(p);
 		return true;
 	}
 
@@ -397,7 +431,7 @@ parse_match(struct parser *p, enum token_type types[], uint32_t types_len)
 }
 
 static bool
-parse_expect(struct parser *p, enum token_type type)
+parse_expect_noadvance(struct parser *p, enum token_type type)
 {
 	if (p->current.type != type) {
 		parse_error(p,
@@ -408,7 +442,16 @@ parse_expect(struct parser *p, enum token_type type)
 		return false;
 	}
 
-	parse_advance(p);
+	return true;
+}
+
+static bool
+parse_expect(struct parser *p, enum token_type type)
+{
+	if (!parse_expect_noadvance(p, type)) {
+		return false;
+	}
+	p->behavior.advance(p);
 	return true;
 }
 
@@ -430,30 +473,60 @@ make_node_t(struct parser *p, enum node_type t)
 	return make_node(p, &(struct node){ .type = t });
 }
 
+static struct node *
+make_node_assign(struct parser *p, enum op_store_flags flags)
+{
+	struct node *n = make_node_t(p, node_type_assign);
+	n->data.type = flags;
+
+	switch (p->previous.type) {
+	case '=': {
+		break;
+	}
+	case token_type_plus_assign: {
+		n->data.type |= op_store_flag_add_store;
+		break;
+	}
+	default: UNREACHABLE;
+	}
+
+	return n;
+}
+
 /******************************************************************************
  * parsing functions
  ******************************************************************************/
 
 static struct node *
-parse_number(struct parser *p)
+parse_number(struct parser *p, bool assignment_allowed)
 {
 	return make_node_t(p, node_type_number);
 }
 
 static struct node *
-parse_id(struct parser *p)
+parse_id(struct parser *p, bool assignment_allowed)
 {
-	return make_node_t(p, node_type_id);
+	struct node *id = make_node_t(p, node_type_id);
+
+	if (assignment_allowed && (parse_accept(p, '=') || parse_accept(p, token_type_plus_assign))) {
+		struct node *n = make_node_assign(p, 0);
+		id->type = node_type_id_lit;
+		n->l = id;
+		n->r = parse_expr(p);
+		return n;
+	} else {
+		return make_node_t(p, node_type_id);
+	}
 }
 
 static struct node *
-parse_string(struct parser *p)
+parse_string(struct parser *p, bool assignment_allowed)
 {
 	return make_node_t(p, node_type_string);
 }
 
 static struct node *
-parse_bool(struct parser *p)
+parse_bool(struct parser *p, bool assignment_allowed)
 {
 	struct node *n = make_node_t(p, node_type_bool);
 	n->data.num = p->previous.type == token_type_true;
@@ -461,7 +534,7 @@ parse_bool(struct parser *p)
 }
 
 static struct node *
-parse_fstring(struct parser *p)
+parse_fstring(struct parser *p, bool assignment_allowed)
 {
 	uint32_t i, j;
 	const struct str *fstr = get_str(p->wk, p->previous.data.str);
@@ -558,7 +631,7 @@ parse_fstring(struct parser *p)
 }
 
 static struct node *
-parse_binary(struct parser *p, struct node *l)
+parse_binary(struct parser *p, struct node *l, bool assignment_allowed)
 {
 	enum node_type t;
 	enum token_type prev = p->previous.type;
@@ -592,7 +665,7 @@ parse_binary(struct parser *p, struct node *l)
 }
 
 static struct node *
-parse_ternary(struct parser *p, struct node *l)
+parse_ternary(struct parser *p, struct node *l, bool assignment_allowed)
 {
 	struct node *n;
 
@@ -606,21 +679,31 @@ parse_ternary(struct parser *p, struct node *l)
 }
 
 static struct node *
-parse_index(struct parser *p, struct node *l)
+parse_index(struct parser *p, struct node *l, bool assignment_allowed)
 {
-	struct node *n;
-
-	n = make_node_t(p, node_type_index);
-	n->l = l;
-	n->r = parse_prec(p, parse_precedence_assignment);
+	struct node *n, *key;
 
+	key = parse_prec(p, parse_precedence_assignment);
 	parse_expect(p, ']');
 
+	if ((p->mode & vm_compile_mode_language_extended) && assignment_allowed
+		&& (parse_accept(p, '=') || parse_accept(p, token_type_plus_assign))) {
+		n = make_node_assign(p, op_store_flag_member);
+		n->l = key;
+		n->r = make_node_t(p, node_type_list);
+		n->r->l = l;
+		n->r->r = parse_expr(p);
+	} else {
+		n = make_node_t(p, node_type_index);
+		n->l = l;
+		n->r = key;
+	}
+
 	return n;
 }
 
 static struct node *
-parse_unary(struct parser *p)
+parse_unary(struct parser *p, bool assignment_allowed)
 {
 	struct node *n;
 	switch (p->previous.type) {
@@ -634,7 +717,7 @@ parse_unary(struct parser *p)
 }
 
 static struct node *
-parse_grouping(struct parser *p)
+parse_grouping(struct parser *p, bool assignment_allowed)
 {
 	struct node *n = parse_prec(p, parse_precedence_assignment);
 	parse_expect(p, ')');
@@ -642,7 +725,7 @@ parse_grouping(struct parser *p)
 }
 
 static struct node *
-parse_grouping_fmt(struct parser *p)
+parse_grouping_fmt(struct parser *p, bool assignment_allowed)
 {
 	struct node *n = make_node_t(p, node_type_group);
 	n->l = parse_prec(p, parse_precedence_assignment);
@@ -737,6 +820,18 @@ parse_type(struct parser *p, type_tag *type, bool top_level)
 	return true;
 }
 
+static struct node *
+parser_get_doc_comment(struct parser *p)
+{
+	struct node *n = 0;
+	if (p->doc_comment) {
+		n = make_node_t(p, node_type_string);
+		n->data.str = p->doc_comment;
+		p->doc_comment = 0;
+	}
+	return n;
+}
+
 static struct node *
 parse_list(struct parser *p, enum node_type t, enum token_type end)
 {
@@ -759,15 +854,16 @@ parse_list(struct parser *p, enum node_type t, enum token_type end)
 		case node_type_args:
 			if (parse_match(p, (enum token_type[]){ token_type_identifier, ':' }, 2)) {
 				parse_expect(p, token_type_identifier);
-				key = parse_id(p);
+				key = parse_id(p, false);
 				key->type = node_type_string;
 				parse_expect(p, ':');
 			}
 			val = parse_expr(p);
 			break;
-		case node_type_def_args:
+		case node_type_def_args: {
+			struct node *doc = parser_get_doc_comment(p);
 			parse_expect(p, token_type_identifier);
-			val = parse_id(p);
+			val = parse_id(p, false);
 			parse_type(p, &type, true);
 
 			if (!type) {
@@ -776,6 +872,7 @@ parse_list(struct parser *p, enum node_type t, enum token_type end)
 
 			val->l = make_node_t(p, node_type_list);
 			val->l->data.type = type;
+			val->r = doc;
 
 			if (parse_accept(p, ':')) {
 				key = val;
@@ -787,6 +884,7 @@ parse_list(struct parser *p, enum node_type t, enum token_type end)
 				}
 			}
 			break;
+		}
 		default: UNREACHABLE; break;
 		}
 
@@ -832,24 +930,24 @@ parse_list(struct parser *p, enum node_type t, enum token_type end)
 }
 
 static struct node *
-parse_array(struct parser *p)
+parse_array(struct parser *p, bool assignment_allowed)
 {
-	return parse_list(p, node_type_array, ']');
+	return p->behavior.parse_list(p, node_type_array, ']');
 }
 
 static struct node *
-parse_dict(struct parser *p)
+parse_dict(struct parser *p, bool assignment_allowed)
 {
-	return parse_list(p, node_type_dict, '}');
+	return p->behavior.parse_list(p, node_type_dict, '}');
 }
 
 static struct node *
-parse_call(struct parser *p, struct node *l)
+parse_call(struct parser *p, struct node *l, bool assignment_allowed)
 {
 	struct node *n;
 	n = make_node_t(p, node_type_call);
 	n->r = l;
-	n->l = parse_list(p, node_type_args, ')');
+	n->l = p->behavior.parse_list(p, node_type_args, ')');
 
 	n->location = source_location_merge(l->location, p->previous.location);
 
@@ -860,25 +958,33 @@ parse_call(struct parser *p, struct node *l)
 }
 
 static struct node *
-parse_method(struct parser *p, struct node *l)
+parse_member(struct parser *p, struct node *l, bool assignment_allowed)
 {
-	struct node *n, *id, *args;
-
-	struct source_location loc = p->previous.location;
+	struct node *n, *id;
 
 	parse_expect(p, token_type_identifier);
-	id = parse_id(p);
+	id = parse_id(p, false);
+	id->type = node_type_id_lit;
 
-	parse_expect(p, '(');
-	n = parse_call(p, id);
-	n->type = node_type_method;
-	n->location = source_location_merge(loc, p->previous.location);
+	if ((p->mode & vm_compile_mode_language_extended) && assignment_allowed
+		&& (parse_accept(p, '=') || parse_accept(p, token_type_plus_assign))) {
+		n = make_node_assign(p, op_store_flag_member);
+		if (!(p->mode & vm_compile_mode_fmt)) {
+			id->type = node_type_string;
+		}
+		n->l = id;
+		n->r = make_node_t(p, node_type_list);
+		n->r->l = l;
+		n->r->r = parse_expr(p);
+	} else {
+		n = make_node_t(p, node_type_member);
+		n->l = l;
+		n->r = id;
 
-	args = make_node_t(p, node_type_args);
-	args->data = n->l->data;
-	args->r = l;
-	args->l = n->l;
-	n->l = args;
+		if (!(p->mode & vm_compile_mode_language_extended)) {
+			parse_expect_noadvance(p, '(');
+		}
+	}
 
 	return n;
 }
@@ -889,9 +995,12 @@ parse_func_params_and_body(struct parser *p, struct node *id)
 	struct node *n = make_node_t(p, node_type_func_def);
 
 	n->l = make_node_t(p, node_type_list);
-	n->l->l = id;
+	n->l->l = make_node_t(p, node_type_list);
+	n->l->l->l = id;
+	n->l->l->r = parser_get_doc_comment(p);
+
 	parse_expect(p, '(');
-	n->l->r = parse_list(p, node_type_def_args, ')');
+	n->l->r = p->behavior.parse_list(p, node_type_def_args, ')');
 
 	if (parse_accept(p, token_type_returntype)) {
 		parse_type(p, &n->data.type, true);
@@ -902,16 +1011,16 @@ parse_func_params_and_body(struct parser *p, struct node *id)
 		n->data.type = 0;
 	}
 
-	parse_expect(p, token_type_eol);
+	parse_accept(p, token_type_eol);
 
-	n->r = parse_block(p, (enum token_type[]){ token_type_endfunc }, 1);
+	n->r = parse_block(p, (enum token_type[]){ token_type_endfunc }, 1, parse_stmt_flag_eol_optional);
 	parse_expect(p, token_type_endfunc);
 
 	return n;
 }
 
 static struct node *
-parse_func(struct parser *p)
+parse_func(struct parser *p, bool assignment_allowed)
 {
 	return parse_func_params_and_body(p, 0);
 }
@@ -925,33 +1034,29 @@ parse_expr(struct parser *p)
 static struct node *
 parse_prec(struct parser *p, enum parse_precedence prec)
 {
-	parse_advance(p);
+	p->behavior.advance(p);
 
 	if (!p->parse_rules[p->previous.type].prefix) {
 		parse_error(p, 0, "expected expression, got %s", token_type_to_s(p->previous.type));
 		return 0;
 	}
 
-	struct node *l = p->parse_rules[p->previous.type].prefix(p);
+	bool assignment_allowed = prec <= parse_precedence_assignment;
+	struct node *l = p->parse_rules[p->previous.type].prefix(p, assignment_allowed);
 
 	while (prec <= p->parse_rules[p->current.type].precedence) {
-		parse_advance(p);
-		l = p->parse_rules[p->previous.type].infix(p, l);
+		p->behavior.advance(p);
+		l = p->parse_rules[p->previous.type].infix(p, l, assignment_allowed);
 	}
 
 	return l;
 }
 
 static struct node *
-parse_stmt(struct parser *p)
+parse_stmt(struct parser *p, enum parse_stmt_flags flags)
 {
 	struct node *n;
 
-	enum token_type assign_sequences[][2] = {
-		{ token_type_identifier, '=' },
-		{ token_type_identifier, token_type_plus_assign },
-	};
-
 	if (parse_accept(p, token_type_if)) {
 		struct node *parent;
 		parent = n = make_node_t(p, node_type_if);
@@ -960,7 +1065,7 @@ parse_stmt(struct parser *p)
 			n->l->l = p->previous.type == token_type_else ? 0 : parse_expr(p);
 			parse_expect(p, token_type_eol);
 			n->l->r = parse_block(
-				p, (enum token_type[]){ token_type_elif, token_type_else, token_type_endif }, 3);
+				p, (enum token_type[]){ token_type_elif, token_type_else, token_type_endif }, 3, 0);
 
 			if (!(parse_accept(p, token_type_elif) || parse_accept(p, token_type_else))) {
 				break;
@@ -977,11 +1082,11 @@ parse_stmt(struct parser *p)
 
 		parse_expect(p, token_type_identifier);
 		n->l->l = make_node_t(p, node_type_list);
-		n->l->l->l = parse_id(p);
+		n->l->l->l = parse_id(p, false);
 
 		if (parse_accept(p, ',')) {
 			parse_expect(p, token_type_identifier);
-			n->l->l->r = parse_id(p);
+			n->l->l->r = parse_id(p, false);
 		}
 
 		parse_expect(p, ':');
@@ -990,7 +1095,7 @@ parse_stmt(struct parser *p)
 		parse_expect(p, token_type_eol);
 
 		++p->inside_loop;
-		n->r = parse_block(p, (enum token_type[]){ token_type_endforeach }, 1);
+		n->r = parse_block(p, (enum token_type[]){ token_type_endforeach }, 1, 0);
 		--p->inside_loop;
 
 		parse_expect(p, token_type_endforeach);
@@ -1000,23 +1105,13 @@ parse_stmt(struct parser *p)
 		n = make_node_t(p, node_type_break);
 	} else if (parse_accept(p, token_type_func)) {
 		parse_expect(p, token_type_identifier);
-		n = parse_func_params_and_body(p, parse_id(p));
+		n = parse_func_params_and_body(p, parse_id(p, false));
 	} else if (parse_accept(p, token_type_return)) {
 		n = make_node_t(p, node_type_return);
 
 		if (p->current.type != token_type_eol) {
 			n->l = parse_expr(p);
 		}
-	} else if (parse_match(p, assign_sequences[0], ARRAY_LEN(assign_sequences[0]))
-		   || parse_match(p, assign_sequences[1], ARRAY_LEN(assign_sequences[1]))) {
-		parse_advance(p);
-
-		n = make_node_t(p, p->current.type == '=' ? node_type_assign : node_type_plusassign);
-		n->l = parse_id(p);
-		n->l->type = node_type_id_lit;
-
-		parse_advance(p);
-		n->r = parse_expr(p);
 	} else {
 		n = parse_expr(p);
 	}
@@ -1024,23 +1119,20 @@ parse_stmt(struct parser *p)
 	if (p->err.unwinding) {
 		while (p->err.unwinding && p->current.type != token_type_eof) {
 			switch (p->current.type) {
-			/* case token_type_func: */
-			/* case token_type_foreach: */
-			/* case token_type_endforeach: */
-			/* case token_type_endif: */
-			/* case token_type_else: */
-			/* case token_type_elif: */
-			/* case token_type_if: */
 			case token_type_eol: p->err.unwinding = false; break;
 			default: break;
 			}
 
 			if (p->err.unwinding) {
-				parse_advance(p);
+				p->behavior.advance(p);
 			}
 		}
 	} else {
-		parse_expect(p, token_type_eol);
+		if (flags & parse_stmt_flag_eol_optional) {
+			parse_accept(p, token_type_eol);
+		} else {
+			parse_expect(p, token_type_eol);
+		}
 	}
 
 	return n;
@@ -1063,7 +1155,7 @@ parse_block_skip_eol_and_check_terminator(struct parser *p, enum token_type type
 }
 
 static struct node *
-parse_block(struct parser *p, enum token_type types[], uint32_t types_len)
+parse_block(struct parser *p, enum token_type types[], uint32_t types_len, enum parse_stmt_flags flags)
 {
 	struct node *res, *n;
 
@@ -1078,9 +1170,11 @@ parse_block(struct parser *p, enum token_type types[], uint32_t types_len)
 			res = n = make_node_t(p, node_type_stmt);
 		}
 
-		n->l = parse_stmt(p);
+		n->l = p->behavior.parse_stmt(p, flags);
 
-		if (parse_block_skip_eol_and_check_terminator(p, types, types_len)) {
+		if ((flags & parse_stmt_flag_eol_optional) && p->previous.type != token_type_eol) {
+			break;
+		} else if (parse_block_skip_eol_and_check_terminator(p, types, types_len)) {
 			break;
 		}
 
@@ -1099,9 +1193,24 @@ parse_block(struct parser *p, enum token_type types[], uint32_t types_len)
 }
 
 static struct node *
-parse_impl(struct parser *p)
+parse_impl(struct workspace *wk,
+	struct source *src,
+	enum vm_compile_mode mode,
+	const struct parse_behavior *behavior,
+	const struct parse_rule *rules,
+	struct parser *p)
 {
 	TracyCZoneAutoS;
+
+	*p = (struct parser){
+		.wk = wk,
+		.nodes = &wk->vm.compiler_state.nodes,
+		.src = src,
+		.mode = mode,
+		.parse_rules = rules,
+		.behavior = *behavior,
+	};
+
 	enum lexer_mode lexer_mode = 0;
 	if (p->mode & vm_compile_mode_language_extended) {
 		lexer_mode |= lexer_mode_functions;
@@ -1111,13 +1220,12 @@ parse_impl(struct parser *p)
 	}
 	lexer_init(&p->lexer, p->wk, p->src, lexer_mode);
 
-	parse_advance(p);
+	p->behavior.advance(p);
 
-	struct node *n = parse_block(p, (enum token_type[]){ token_type_eof }, 1);
+	struct node *n = parse_block(p, (enum token_type[]){ token_type_eof }, 1, 0);
 
 	if (!p->err.count && !n) {
-		// This is an empty meson.build file.  Add a single dummy
-		// statement.
+		// This is an empty file.  Add a single dummy statement.
 		n = make_node_t(p, node_type_stmt);
 	}
 
@@ -1129,47 +1237,48 @@ parse_impl(struct parser *p)
 
 // clang-format off
 static const struct parse_rule parse_rules_base[token_type_count] = {
-	[token_type_number]     = { parse_number,   0,             0                           },
-	[token_type_identifier] = { parse_id,       0,             0                           },
-	[token_type_string]     = { parse_string,   0,             0                           },
-	[token_type_fstring]    = { parse_fstring,  0,             0                           },
-	[token_type_true]       = { parse_bool,     0,             0                           },
-	[token_type_false]      = { parse_bool,     0,             0                           },
-	['(']                   = { parse_grouping, parse_call,    parse_precedence_call       },
-	['[']                   = { parse_array,    parse_index,   parse_precedence_call       },
-	['{']                   = { parse_dict,     0,             0                           },
-	['+']                   = { 0,              parse_binary,  parse_precedence_term       },
-	['-']                   = { parse_unary,    parse_binary,  parse_precedence_term       },
-	['*']                   = { 0,              parse_binary,  parse_precedence_factor     },
-	['/']                   = { 0,              parse_binary,  parse_precedence_factor     },
-	['%']                   = { 0,              parse_binary,  parse_precedence_factor     },
-	['<']                   = { 0,              parse_binary,  parse_precedence_comparison },
-	['>']                   = { 0,              parse_binary,  parse_precedence_comparison },
-	[token_type_leq]        = { 0,              parse_binary,  parse_precedence_comparison },
-	[token_type_geq]        = { 0,              parse_binary,  parse_precedence_comparison },
-	[token_type_or]         = { 0,              parse_binary,  parse_precedence_or         },
-	[token_type_and]        = { 0,              parse_binary,  parse_precedence_and        },
-	[token_type_eq]         = { 0,              parse_binary,  parse_precedence_equality   },
-	[token_type_neq]        = { 0,              parse_binary,  parse_precedence_equality   },
-	[token_type_in]         = { 0,              parse_binary,  parse_precedence_equality   },
-	[token_type_not_in]     = { 0,              parse_binary,  parse_precedence_equality   },
-	['.']                   = { 0,              parse_method,  parse_precedence_call       },
-	['?']                   = { 0,              parse_ternary, parse_precedence_assignment },
-	[token_type_not]        = { parse_unary,    0,             0                           },
-	[token_type_func]       = { parse_func,     0,             0                           },
+	[token_type_number]      = { parse_number,   0,             0                           },
+	[token_type_identifier]  = { parse_id,       0,             0                           },
+	[token_type_string]      = { parse_string,   0,             0                           },
+	[token_type_fstring]     = { parse_fstring,  0,             0                           },
+	[token_type_true]        = { parse_bool,     0,             0                           },
+	[token_type_false]       = { parse_bool,     0,             0                           },
+	['(']                    = { parse_grouping, parse_call,    parse_precedence_call       },
+	['[']                    = { parse_array,    parse_index,   parse_precedence_call       },
+	['{']                    = { parse_dict,     0,             0                           },
+	['+']                    = { 0,              parse_binary,  parse_precedence_term       },
+	['-']                    = { parse_unary,    parse_binary,  parse_precedence_term       },
+	['*']                    = { 0,              parse_binary,  parse_precedence_factor     },
+	['/']                    = { 0,              parse_binary,  parse_precedence_factor     },
+	['%']                    = { 0,              parse_binary,  parse_precedence_factor     },
+	['<']                    = { 0,              parse_binary,  parse_precedence_comparison },
+	['>']                    = { 0,              parse_binary,  parse_precedence_comparison },
+	[token_type_leq]         = { 0,              parse_binary,  parse_precedence_comparison },
+	[token_type_geq]         = { 0,              parse_binary,  parse_precedence_comparison },
+	[token_type_or]          = { 0,              parse_binary,  parse_precedence_or         },
+	[token_type_and]         = { 0,              parse_binary,  parse_precedence_and        },
+	[token_type_eq]          = { 0,              parse_binary,  parse_precedence_equality   },
+	[token_type_neq]         = { 0,              parse_binary,  parse_precedence_equality   },
+	[token_type_in]          = { 0,              parse_binary,  parse_precedence_equality   },
+	[token_type_not_in]      = { 0,              parse_binary,  parse_precedence_equality   },
+	['.']                    = { 0,              parse_member,  parse_precedence_call       },
+	['?']                    = { 0,              parse_ternary, parse_precedence_assignment },
+	[token_type_not]         = { parse_unary,    0,             0                           },
+	[token_type_func]        = { parse_func,     0,             parse_precedence_assignment },
+};
+
+static const struct parse_behavior parse_behavior_base = {
+	.advance = parse_advance,
+	.parse_stmt = parse_stmt,
+	.parse_list = parse_list,
 };
 // clang-format on
 
 struct node *
 parse(struct workspace *wk, struct source *src, enum vm_compile_mode mode)
 {
-	return parse_impl(&(struct parser){
-		.wk = wk,
-		.nodes = &wk->vm.compiler_state.nodes,
-		.src = src,
-		.mode = mode,
-		.parse_rules = parse_rules_base,
-	});
+	struct parser p;
+	return parse_impl(wk, src, mode, &parse_behavior_base, parse_rules_base, &p);
 }
 
 struct node *
@@ -1179,15 +1288,8 @@ parse_fmt(struct workspace *wk, struct source *src, enum vm_compile_mode mode, o
 	memcpy(parse_rules, parse_rules_base, sizeof(parse_rules_base));
 	parse_rules['('].prefix = parse_grouping_fmt;
 
-	struct parser p = {
-		.wk = wk,
-		.nodes = &wk->vm.compiler_state.nodes,
-		.src = src,
-		.mode = mode,
-		.parse_rules = parse_rules,
-	};
-
-	struct node *n = parse_impl(&p);
+	struct parser p;
+	struct node *n = parse_impl(wk, src, mode, &parse_behavior_base, parse_rules, &p);
 
 	*raw_blocks = p.lexer.fmt.raw_blocks;
 
@@ -1195,3 +1297,235 @@ parse_fmt(struct workspace *wk, struct source *src, enum vm_compile_mode mode, o
 
 	return n;
 }
+
+/******************************************************************************
+* cmake
+******************************************************************************/
+
+static void
+cm_parse_advance(struct parser *p)
+{
+	struct lexer prev_lexer;
+
+	p->previous = p->current;
+	p->fmt.previous = p->fmt.current;
+
+	bool relex = false;
+
+relex:
+	prev_lexer = p->lexer;
+
+	switch (p->cm_mode) {
+	case cm_parse_mode_conditional: p->lexer.cm_mode = cm_lexer_mode_conditional; break;
+	default: break;
+	}
+
+	cm_lexer_next(&p->lexer, &p->current);
+	p->lexer.cm_mode = cm_lexer_mode_default;
+
+	while (p->current.type == token_type_error) {
+		parse_error(p, &p->current.location, "%s", get_cstr(p->wk, p->current.data.str));
+		p->err.unwinding = false;
+		prev_lexer = p->lexer;
+		cm_lexer_next(&p->lexer, &p->current);
+	}
+
+	if (p->current.type == token_type_eof) {
+		if (p->previous.type != token_type_eol) {
+			// simulate a eol at the end of the file even if we
+			// didn't get one
+			p->current.type = token_type_eol;
+		}
+	}
+
+	if (!relex) {
+		if (p->cm_mode == cm_parse_mode_command && p->current.type == token_type_identifier) {
+			struct lexer lexer_peek = p->lexer;
+			struct token next;
+
+			cm_lexer_next(&lexer_peek, &next);
+
+			if (next.type == '(') {
+				p->lexer = prev_lexer;
+				p->lexer.cm_mode = cm_lexer_mode_command;
+				relex = true;
+				goto relex;
+			}
+		}
+	}
+
+	if (p->cm_mode == cm_parse_mode_conditional && p->current.type == token_type_string) {
+		p->current.type = token_type_identifier;
+	} else if (p->cm_mode == cm_parse_mode_command_args && p->current.type == token_type_identifier) {
+		p->current.type = token_type_string;
+	}
+
+	/* LL("%d, previous: %s, current: ", p->cm_mode, token_to_s(p->wk, &p->previous)); */
+	/* log_plain("%s\n", token_to_s(p->wk, &p->current)); */
+
+	/* obj_fprintf(p->wk, log_file(), "%o, %o\n", p->fmt.previous.ws, p->fmt.current.ws); */
+
+	/* list_line_range(p->src, p->current.location, 0); */
+}
+
+static struct node *
+cm_parse_id(struct parser *p, bool assignment_allowed)
+{
+	if (p->cm_mode == cm_parse_mode_conditional) {
+		return make_node_t(p, node_type_maybe_id);
+	} else {
+		return make_node_t(p, node_type_id);
+	}
+}
+
+static struct node *
+cm_parse_string(struct parser *p, bool assignment_allowed)
+{
+	return make_node_t(p, node_type_string);
+}
+
+static struct node *
+cm_parse_with_mode(struct parser *p, enum cm_parse_mode mode, struct node *(parse_fun)(struct parser *))
+{
+	stack_push(&p->wk->stack, p->cm_mode, mode);
+	struct node *n = parse_fun(p);
+	stack_pop(&p->wk->stack, p->cm_mode);
+	return n;
+}
+
+static struct node *
+cm_parse_ignored_list(struct parser *p)
+{
+	parse_expect(p, '(');
+	p->behavior.parse_list(p, node_type_def_args, ')');
+	return 0;
+}
+
+static struct node *
+cm_parse_stmt(struct parser *p, enum parse_stmt_flags flags)
+{
+	struct node *n;
+	if (parse_accept(p, token_type_if)) {
+		struct node *parent;
+		parent = n = make_node_t(p, node_type_if);
+		while (true) {
+			n->l = make_node_t(p, node_type_list);
+			n->l->l = p->previous.type == token_type_else ?
+					  cm_parse_ignored_list(p) :
+					  cm_parse_with_mode(p, cm_parse_mode_conditional, parse_expr);
+			parse_expect(p, token_type_eol);
+			n->l->r = parse_block(
+				p, (enum token_type[]){ token_type_elif, token_type_else, token_type_endif }, 3, 0);
+
+			if (!(parse_accept(p, token_type_elif) || parse_accept(p, token_type_else))) {
+				break;
+			}
+
+			n = n->r = make_node_t(p, node_type_if);
+		}
+
+		parse_expect(p, token_type_endif);
+		cm_parse_ignored_list(p);
+
+		n = parent;
+	} else {
+		n = parse_expr(p);
+	}
+
+	if (p->err.unwinding) {
+		while (p->err.unwinding && p->current.type != token_type_eof) {
+			switch (p->current.type) {
+			case token_type_eol: p->err.unwinding = false; break;
+			default: break;
+			}
+
+			if (p->err.unwinding) {
+				p->behavior.advance(p);
+			}
+		}
+	} else {
+		parse_expect(p, token_type_eol);
+	}
+
+	return n;
+}
+
+static struct node *
+cm_parse_list(struct parser *p, enum node_type t, enum token_type end)
+{
+	struct node *n = make_node_t(p, t), *res = n;
+	uint32_t len = 0;
+
+	while (p->current.type != end && p->current.type != token_type_eol) {
+		n->l = parse_expr(p);
+		++len;
+		if (p->current.type != end) {
+			n = n->r = make_node_t(p, node_type_list);
+		}
+	}
+
+	parse_expect(p, end);
+
+	res->data.len.args = len;
+	return res;
+}
+
+static struct node *
+cm_parse_call(struct parser *p, struct node *l, bool assignment_allowed)
+{
+	struct node *n;
+
+	if (p->current.type == token_type_identifier) {
+		p->current.type = token_type_string;
+	}
+
+	stack_push(&p->wk->stack, p->cm_mode, cm_parse_mode_command_args);
+
+	n = make_node_t(p, node_type_call);
+	n->r = l;
+	n->l = p->behavior.parse_list(p, node_type_args, ')');
+
+	/* n->r->data.str = make_strf(p->wk, "cm_%s", get_cstr(p->wk, n->r->data.str)); */
+
+	stack_pop(&p->wk->stack, p->cm_mode);
+
+	n->location = source_location_merge(l->location, p->previous.location);
+
+	if (n->r->type == node_type_id) {
+		n->r->type = node_type_id_lit;
+	}
+	return n;
+}
+
+struct node *
+cm_parse(struct workspace *wk, struct source *src)
+{
+	struct parse_behavior behavior = {
+		.advance = cm_parse_advance,
+		.parse_stmt = cm_parse_stmt,
+		.parse_list = cm_parse_list,
+	};
+
+	// clang-format off
+	struct parse_rule parse_rules[token_type_count] = {
+		[token_type_identifier] = { cm_parse_id, 0, 0 },
+		[token_type_string] = { cm_parse_string, 0, 0 },
+		['('] = { parse_grouping, cm_parse_call, parse_precedence_call },
+		['<']                   = { 0,              parse_binary,  parse_precedence_comparison },
+		['>']                   = { 0,              parse_binary,  parse_precedence_comparison },
+		[token_type_leq]        = { 0,              parse_binary,  parse_precedence_comparison },
+		[token_type_geq]        = { 0,              parse_binary,  parse_precedence_comparison },
+		[token_type_or]         = { 0,              parse_binary,  parse_precedence_or         },
+		[token_type_and]        = { 0,              parse_binary,  parse_precedence_and        },
+		[token_type_eq]         = { 0,              parse_binary,  parse_precedence_equality   },
+		[token_type_neq]        = { 0,              parse_binary,  parse_precedence_equality   },
+		[token_type_in]         = { 0,              parse_binary,  parse_precedence_equality   },
+		[token_type_not_in]     = { 0,              parse_binary,  parse_precedence_equality   },
+	};
+	// clang-format on
+
+	struct parser p;
+	struct node *n = parse_impl(wk, src, 0, &behavior, parse_rules, &p);
+	/* print_ast(wk, n); */
+	return n;
+}
