@@ -5,31 +5,36 @@
 
 #include "compat.h"
 
+#include <stdarg.h>
 #include <stdlib.h>
 #include <string.h>
 
+#include "buf_size.h"
 #include "error.h"
 #include "lang/object.h"
 #include "lang/string.h"
 #include "lang/workspace.h"
 #include "log.h"
 #include "memmem.h"
+#include "platform/assert.h"
 #include "platform/mem.h"
 
 void
-str_unescape(struct workspace *wk, struct sbuf *sb, const struct str *ss, bool escape_whitespace)
+str_escape(struct workspace *wk, struct sbuf *sb, const struct str *ss, bool escape_whitespace)
 {
 	bool esc;
 	uint32_t i;
 
 	for (i = 0; i < ss->len; ++i) {
-		esc = ss->s[i] < 32;
+		esc = ss->s[i] < 32 || ss->s[i] == '\'';
 		if (!escape_whitespace && strchr("\t\n\r", ss->s[i])) {
 			esc = false;
 		}
 
 		if (esc) {
-			if (7 <= ss->s[i] && ss->s[i] <= 13) {
+			if (ss->s[i] == '\'') {
+				sbuf_pushf(wk, sb, "\\'");
+			} else if (7 <= ss->s[i] && ss->s[i] <= 13) {
 				sbuf_pushf(wk, sb, "\\%c", "abtnvfr"[ss->s[i] - 7]);
 			} else {
 				sbuf_pushf(wk, sb, "\\%d", ss->s[i]);
@@ -40,6 +45,27 @@ str_unescape(struct workspace *wk, struct sbuf *sb, const struct str *ss, bool e
 	}
 }
 
+void
+str_escape_json(struct workspace *wk, struct sbuf *sb, const struct str *ss)
+{
+	uint32_t i;
+	const char *esc = "\"\\";
+
+	for (i = 0; i < ss->len; ++i) {
+		if (strchr(esc, ss->s[i])) {
+			sbuf_pushf(wk, sb, "\\\"");
+		} else if (ss->s[i] < 32 || ss->s[i] > 126) {
+			if (8 <= ss->s[i] && ss->s[i] <= 13 && ss->s[i] != 11) {
+				sbuf_pushf(wk, sb, "\\%c", "btn_fr"[ss->s[i] - 8]);
+			} else {
+				sbuf_pushf(wk, sb, "\\u%04x", ss->s[i]);
+			}
+		} else {
+			sbuf_push(wk, sb, ss->s[i]);
+		}
+	}
+}
+
 bool
 str_has_null(const struct str *ss)
 {
@@ -414,6 +440,26 @@ str_contains(const struct str *str, const struct str *substr)
 	return !!memmem(str->s, str->len, substr->s, substr->len);
 }
 
+bool
+str_containsi(const struct str *str, const struct str *substr)
+{
+	if (substr->len > str->len) {
+		return false;
+	} else if (substr->len == str->len) {
+		return str_eqli(str, substr);
+	}
+
+	uint32_t i;
+	for (i = 0; i < str->len - substr->len; ++i) {
+		struct str a = { .s = str->s + i, .len = substr->len };
+		if (str_eqli(&a, substr)) {
+			return true;
+		}
+	}
+
+	return false;
+}
+
 obj
 str_join(struct workspace *wk, obj s1, obj s2)
 {
@@ -434,6 +480,12 @@ is_whitespace(char c)
 	return c == ' ' || c == '\t' || c == '\n' || c == '\r';
 }
 
+bool
+is_whitespace_except_newline(char c)
+{
+	return c == ' ' || c == '\t' || c == '\r';
+}
+
 bool
 str_to_i(const struct str *ss, int64_t *res, bool strip)
 {
@@ -464,7 +516,6 @@ str_to_i(const struct str *ss, int64_t *res, bool strip)
 obj
 str_split(struct workspace *wk, const struct str *ss, const struct str *split)
 {
-	static const char *whitespace = "\n\r\t ";
 	obj res;
 	make_obj(wk, &res, obj_array);
 
@@ -485,12 +536,12 @@ str_split(struct workspace *wk, const struct str *ss, const struct str *split)
 			}
 		} else {
 			start = i;
-			while (start < ss->len && strchr(whitespace, ss->s[start])) {
+			while (start < ss->len && is_whitespace(ss->s[start])) {
 				++start;
 			}
 
 			uint32_t end = start;
-			while (end < ss->len && !strchr(whitespace, ss->s[end])) {
+			while (end < ss->len && !is_whitespace(ss->s[end])) {
 				++end;
 			}
 
@@ -552,6 +603,20 @@ str_splitlines(struct workspace *wk, const struct str *ss)
 	return res;
 }
 
+bool
+str_split_in_two(const struct str *s, struct str *l, struct str *r, char split)
+{
+	const char *p;
+	if (!(p = memchr(s->s, split, s->len))) {
+		return false;
+	}
+
+	*l = (struct str){ s->s, p - s->s };
+	*r = (struct str){ s->s + l->len + 1, s->len - (l->len + 1) };
+
+	return true;
+}
+
 static bool
 str_has_chr(char c, const struct str *ss)
 {
@@ -568,7 +633,7 @@ str_has_chr(char c, const struct str *ss)
 obj
 str_strip(struct workspace *wk, const struct str *ss, const struct str *strip, enum str_strip_flag flags)
 {
-	const struct str *defstrip = &WKSTR(" \n\t");
+	const struct str *defstrip = &WKSTR(" \r\n\t");
 
 	if (!strip) {
 		strip = defstrip;
@@ -739,7 +804,7 @@ sbuf_push(struct workspace *wk, struct sbuf *sb, char s)
 {
 	if (sb->flags & sbuf_flag_write) {
 		FILE *out = (FILE *)sb->buf;
-		if (out == log_file()) {
+		if (out == _log_file()) {
 			log_plain("%c", s);
 		} else if (fputc(s, out) == EOF) {
 			error_unrecoverable("failed to write output to file");
@@ -759,7 +824,7 @@ sbuf_pushn(struct workspace *wk, struct sbuf *sb, const char *s, uint32_t n)
 {
 	if (sb->flags & sbuf_flag_write) {
 		FILE *out = (FILE *)sb->buf;
-		if (out == log_file()) {
+		if (out == _log_file()) {
 			log_plain("%.*s", n, s);
 		} else if (!fs_fwrite(s, n, out)) {
 			error_unrecoverable("failed to write output to file");
@@ -783,7 +848,7 @@ sbuf_pushs(struct workspace *wk, struct sbuf *sb, const char *s)
 {
 	if (sb->flags & sbuf_flag_write) {
 		FILE *out = (FILE *)sb->buf;
-		if (out == log_file()) {
+		if (out == _log_file()) {
 			log_plain("%s", s);
 		} else if (fputs(s, out) == EOF) {
 			error_unrecoverable("failed to write output to file");
@@ -804,22 +869,21 @@ sbuf_pushs(struct workspace *wk, struct sbuf *sb, const char *s)
 }
 
 void
-sbuf_pushf(struct workspace *wk, struct sbuf *sb, const char *fmt, ...)
+sbuf_vpushf(struct workspace *wk, struct sbuf *sb, const char *fmt, va_list args)
 {
 	uint32_t len;
-	va_list args, args_copy;
-	va_start(args, fmt);
 
 	if (sb->flags & sbuf_flag_write) {
 		FILE *out = (FILE *)sb->buf;
-		if (out == log_file()) {
+		if (out == _log_file()) {
 			log_plainv(fmt, args);
 		} else if (vfprintf((FILE *)sb->buf, fmt, args) < 0) {
 			error_unrecoverable("failed to write output to file");
 		}
-		goto done;
+		return;
 	}
 
+	va_list args_copy;
 	va_copy(args_copy, args);
 
 	len = vsnprintf(NULL, 0, fmt, args_copy);
@@ -830,8 +894,16 @@ sbuf_pushf(struct workspace *wk, struct sbuf *sb, const char *fmt, ...)
 	sb->len += len;
 
 	va_end(args_copy);
+}
+
+void
+sbuf_pushf(struct workspace *wk, struct sbuf *sb, const char *fmt, ...)
+{
+	va_list args;
+	va_start(args, fmt);
+
+	sbuf_vpushf(wk, sb, fmt, args);
 
-done:
 	va_end(args);
 }
 
@@ -885,3 +957,11 @@ sbuf_into_str(struct workspace *wk, struct sbuf *sb)
 		return make_strn(wk, sb->buf, sb->len);
 	}
 }
+
+void
+cstr_copy(char *dest, const char *src, uint32_t dest_len)
+{
+	uint32_t src_len = strlen(src) + 1;
+	assert(src_len <= dest_len);
+	memcpy(dest, src, src_len);
+}
