@@ -11,31 +11,36 @@
 #include <string.h>
 
 #include "args.h"
-#include "backend/backend.h"
 #include "backend/output.h"
+#include "buf_size.h"
 #include "cmd_install.h"
+#include "cmd_subprojects.h"
 #include "cmd_test.h"
 #include "embedded.h"
+#include "error.h"
 #include "external/libarchive.h"
 #include "external/libcurl.h"
 #include "external/libpkgconf.h"
 #include "external/samurai.h"
 #include "lang/analyze.h"
-#include "lang/compiler.h"
 #include "lang/fmt.h"
 #include "lang/func_lookup.h"
+#include "lang/object_iterators.h"
+#include "lang/parser.h"
 #include "lang/serial.h"
-#include "machine_file.h"
 #include "meson_opts.h"
 #include "options.h"
 #include "opts.h"
+#include "platform/assert.h"
 #include "platform/init.h"
 #include "platform/mem.h"
+#include "platform/os.h"
 #include "platform/path.h"
 #include "platform/run_cmd.h"
 #include "tracy.h"
+#include "ui.h"
 #include "version.h"
-#include "wrap.h"
+#include "vsenv.h"
 
 static bool
 ensure_in_build_dir(void)
@@ -70,8 +75,27 @@ ret:
 	return ret;
 }
 
+static void
+setup_platform_env(const char *build_dir, bool force)
+{
+	if (host_machine.sys == machine_system_windows) {
+		SBUF_manual(cache);
+		const char *cache_path = 0;
+		if (build_dir) {
+			path_copy(0, &cache, build_dir);
+			path_push(0, &cache, output_path.private_dir);
+			if (fs_dir_exists(cache.buf)) {
+				path_push(0, &cache, "vsenv.txt");
+				cache_path = cache.buf;
+			}
+		}
+
+		vsenv_setup(cache_path, force);
+	}
+}
+
 static bool
-cmd_exe(uint32_t argc, uint32_t argi, char *const argv[])
+cmd_exe(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
 	struct {
 		const char *feed;
@@ -189,7 +213,7 @@ ret:
 }
 
 static bool
-cmd_check(uint32_t argc, uint32_t argi, char *const argv[])
+cmd_check(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
 	struct {
 		const char *filename;
@@ -269,24 +293,34 @@ ret:
 }
 
 static bool
-cmd_analyze(uint32_t argc, uint32_t argi, char *const argv[])
+cmd_analyze(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
 	struct az_opts opts = {
-		.replay_opts = error_diagnostic_store_replay_include_sources,
 		.enabled_diagnostics = az_diagnostic_unused_variable | az_diagnostic_dead_code
 				       | az_diagnostic_redirect_script_error,
 	};
 
-	OPTSTART("luqO:W:i:td:") {
+	enum {
+		action_trace,
+		action_define,
+		action_default,
+	} action
+		= action_default;
+
+	static const struct command commands[] = {
+		[action_trace] = { "trace", 0, "print a tree of all meson source files that are evaluated" },
+		[action_define] = { "define <var>", 0, "lookup the definition of a variable" },
+		0,
+	};
+
+	OPTSTART("luqO:W:i:") {
 	case 'i': opts.internal_file = optarg; break;
 	case 'l':
 		opts.subdir_error = true;
-		opts.replay_opts &= ~error_diagnostic_store_replay_include_sources;
+		opts.replay_opts |= error_diagnostic_store_replay_dont_include_sources;
 		break;
 	case 'O': opts.file_override = optarg; break;
 	case 'q': opts.replay_opts |= error_diagnostic_store_replay_errors_only; break;
-	case 't': opts.eval_trace = true; break;
-	case 'd': opts.get_definition_for = optarg; break;
 	case 'W': {
 		bool enable = true;
 		const char *name = optarg;
@@ -322,13 +356,42 @@ cmd_analyze(uint32_t argc, uint32_t argi, char *const argv[])
 		"  -q - only report errors\n"
 		"  -O <path> - read project file with matching path from stdin\n"
 		"  -i <path> - analyze the single file <path> in internal mode\n"
-		"  -t - print a tree of all meson source files that are evaluated\n"
-		"  -d <var> - print the location of the definition of <var>\n"
 		"  -W [no-]<diagnostic> - enable or disable diagnostics\n"
 		"  -W list - list available diagnostics\n"
 		"  -W error - turn all warnings into errors\n",
-		NULL,
-		0)
+		commands,
+		-1);
+
+	uint32_t cmd_i = action_default;
+	if (!find_cmd(commands, &cmd_i, argc, argi, argv, true)) {
+		return false;
+	}
+	if (cmd_i != action_default) {
+		++argi;
+	}
+	action = cmd_i;
+
+	if (action == action_define) {
+		if (!check_operands(argc, argi, 1)) {
+			return false;
+		}
+	} else {
+		if (!check_operands(argc, argi, 0)) {
+			return false;
+		}
+	}
+
+	switch (action) {
+	case action_default: break;
+	case action_trace: {
+		opts.eval_trace = true;
+		break;
+	}
+	case action_define: {
+		opts.get_definition_for = argv[argi];
+		break;
+	}
+	}
 
 	if (opts.internal_file && opts.file_override) {
 		LOG_E("-i and -O are mutually exclusive");
@@ -351,7 +414,7 @@ cmd_analyze(uint32_t argc, uint32_t argi, char *const argv[])
 }
 
 static bool
-cmd_options(uint32_t argc, uint32_t argi, char *const argv[])
+cmd_options(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
 	struct list_options_opts opts = { 0 };
 	OPTSTART("am") {
@@ -369,7 +432,7 @@ cmd_options(uint32_t argc, uint32_t argi, char *const argv[])
 }
 
 static bool
-cmd_summary(uint32_t argc, uint32_t argi, char *const argv[])
+cmd_summary(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
 	OPTSTART("") {
 	}
@@ -398,7 +461,7 @@ ret:
 }
 
 static bool
-cmd_info(uint32_t argc, uint32_t argi, char *const argv[])
+cmd_info(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
 	LOG_W("the info subcommand has been deprecated, please use options / summary directly");
 
@@ -412,145 +475,16 @@ cmd_info(uint32_t argc, uint32_t argi, char *const argv[])
 	}
 	OPTEND(argv[argi], "", "", commands, -1);
 
-	cmd_func cmd = NULL;
-	;
-	if (!find_cmd(commands, &cmd, argc, argi, argv, false)) {
+	uint32_t cmd_i;
+	if (!find_cmd(commands, &cmd_i, argc, argi, argv, false)) {
 		return false;
 	}
 
-	assert(cmd);
-	return cmd(argc, argi, argv);
-}
-
-static const char *cmd_subprojects_subprojects_dir = "subprojects";
-
-static bool
-cmd_subprojects_check_wrap(uint32_t argc, uint32_t argi, char *const argv[])
-{
-	struct {
-		const char *filename;
-	} opts = { 0 };
-
-	OPTSTART("") {
-	}
-	OPTEND(argv[argi], " <filename>", "", NULL, 1)
-
-	opts.filename = argv[argi];
-
-	bool ret = false;
-
-	struct wrap wrap = { 0 };
-	if (!wrap_parse(opts.filename, &wrap)) {
-		goto ret;
-	}
-
-	ret = true;
-ret:
-	wrap_destroy(&wrap);
-	return ret;
-}
-
-struct cmd_subprojects_download_ctx {
-	const char *subprojects;
-};
-
-static enum iteration_result
-cmd_subprojects_download_iter(void *_ctx, const char *name)
-{
-	struct cmd_subprojects_download_ctx *ctx = _ctx;
-	uint32_t len = strlen(name);
-	SBUF_manual(path);
-
-	if (len <= 5 || strcmp(&name[len - 5], ".wrap") != 0) {
-		goto cont;
-	}
-
-	path_join(NULL, &path, ctx->subprojects, name);
-
-	if (!fs_file_exists(path.buf)) {
-		goto cont;
-	}
-
-	LOG_I("fetching %s", name);
-	struct wrap wrap = { 0 };
-	if (!wrap_handle(path.buf, ctx->subprojects, &wrap, true)) {
-		goto cont;
-	}
-
-	wrap_destroy(&wrap);
-cont:
-	sbuf_destroy(&path);
-	return ir_cont;
+	return commands[cmd_i].cmd(0, argc, argi, argv);
 }
 
 static bool
-cmd_subprojects_download(uint32_t argc, uint32_t argi, char *const argv[])
-{
-	bool res = false;
-
-	OPTSTART("") {
-	}
-	OPTEND(argv[argi], " <list of subprojects>", "", NULL, -1)
-
-	SBUF_manual(path);
-	path_make_absolute(NULL, &path, cmd_subprojects_subprojects_dir);
-
-	struct cmd_subprojects_download_ctx ctx = {
-		.subprojects = path.buf,
-	};
-
-	if (argc > argi) {
-		SBUF_manual(wrap_file);
-
-		for (; argc > argi; ++argi) {
-			path_join(NULL, &wrap_file, ctx.subprojects, argv[argi]);
-
-			sbuf_pushs(NULL, &wrap_file, ".wrap");
-
-			if (!fs_file_exists(wrap_file.buf)) {
-				LOG_E("wrap file for '%s' not found", argv[argi]);
-				goto ret;
-			}
-
-			if (cmd_subprojects_download_iter(&ctx, wrap_file.buf) == ir_err) {
-				goto ret;
-			}
-		}
-
-		sbuf_destroy(&wrap_file);
-		res = true;
-	} else {
-		res = fs_dir_foreach(path.buf, &ctx, cmd_subprojects_download_iter);
-	}
-
-ret:
-	sbuf_destroy(&path);
-	return res;
-}
-
-static bool
-cmd_subprojects(uint32_t argc, uint32_t argi, char *const argv[])
-{
-	static const struct command commands[] = {
-		{ "check-wrap", cmd_subprojects_check_wrap, "check if a wrap file is valid" },
-		{ "download", cmd_subprojects_download, "download subprojects" },
-		{ 0 },
-	};
-
-	OPTSTART("d:") {
-	case 'd': cmd_subprojects_subprojects_dir = optarg; break;
-	}
-	OPTEND(argv[0], "", "  -d <directory> - use an alternative subprojects directory\n", commands, -1)
-
-	cmd_func cmd;
-	if (!find_cmd(commands, &cmd, argc, argi, argv, false)) {
-		return false;
-	}
-
-	return cmd(argc, argi, argv);
-}
-static bool
-cmd_eval(uint32_t argc, uint32_t argi, char *const argv[])
+cmd_eval(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
 	struct workspace wk;
 	workspace_init_bare(&wk);
@@ -565,7 +499,7 @@ cmd_eval(uint32_t argc, uint32_t argi, char *const argv[])
 		break;
 	}
 	case 'b': {
-		vm_dbg_push_breakpoint(&wk, optarg);
+		vm_dbg_push_breakpoint_str(&wk, optarg);
 		break;
 	}
 	}
@@ -587,18 +521,13 @@ cmd_eval(uint32_t argc, uint32_t argi, char *const argv[])
 
 	struct source src = { 0 };
 
-	machine_init();
-
 	wk.vm.lang_mode = language_internal;
 
 	if (embedded) {
-		if (!(src.src = embedded_get(filename))) {
+		if (!(embedded_get(filename, &src))) {
 			LOG_E("failed to find '%s' in embedded sources", filename);
 			goto ret;
 		}
-
-		src.len = strlen(src.src);
-		src.label = filename;
 	} else {
 		if (!fs_read_entire_file(filename, &src)) {
 			goto ret;
@@ -617,7 +546,7 @@ cmd_eval(uint32_t argc, uint32_t argi, char *const argv[])
 	}
 
 	obj res;
-	if (!eval(&wk, &src, eval_mode_default, &res)) {
+	if (!eval(&wk, &src, build_language_meson, 0, &res)) {
 		goto ret;
 	}
 
@@ -628,7 +557,7 @@ ret:
 }
 
 static bool
-cmd_repl(uint32_t argc, uint32_t argi, char *const argv[])
+cmd_repl(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
 	struct workspace wk;
 	workspace_init(&wk);
@@ -644,7 +573,7 @@ cmd_repl(uint32_t argc, uint32_t argi, char *const argv[])
 }
 
 static bool
-cmd_dump_signatures(uint32_t argc, uint32_t argi, char *const argv[])
+cmd_dump_signatures(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
 	OPTSTART("") {
 	}
@@ -666,13 +595,234 @@ cmd_dump_signatures(uint32_t argc, uint32_t argi, char *const argv[])
 }
 
 static bool
-cmd_internal(uint32_t argc, uint32_t argi, char *const argv[])
+cmd_dump_docs(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
+{
+	log_set_file(stderr);
+
+	OPTSTART("") {
+	}
+	OPTEND(argv[argi], "", "", NULL, 0)
+
+	struct workspace wk;
+	workspace_init(&wk);
+
+	obj id;
+	make_project(&wk, &id, "dummy", wk.source_root, wk.build_root);
+	if (!setup_project_options(&wk, NULL)) {
+		UNREACHABLE;
+	}
+
+	dump_function_docs(&wk);
+
+	workspace_destroy(&wk);
+	return true;
+}
+
+static bool
+cmd_dump_toolchains(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
+{
+	struct obj_compiler comp = { 0 };
+	bool set_linker = false, set_static_linker = false;
+
+	const char *n1_args[32] = { "<value1>", "<value2>" };
+	struct args n1 = { n1_args, 2 };
+	struct toolchain_dump_opts opts = {
+		.s1 = "<value1>",
+		.s2 = "<value2>",
+		.b1 = true,
+		.i1 = 0,
+		.n1 = &n1,
+	};
+
+	OPTSTART("t:s:") {
+	case 't': {
+		if (strcmp(optarg, "list") == 0) {
+			printf("registered toolchains:\n");
+			enum toolchain_component component;
+			for (component = 0; component < toolchain_component_count; ++component) {
+				const struct toolchain_id *list = 0;
+				uint32_t i, count = 0;
+				switch (component) {
+				case toolchain_component_compiler:
+					list = compiler_type_name;
+					count = compiler_type_count;
+					break;
+				case toolchain_component_linker:
+					list = linker_type_name;
+					count = linker_type_count;
+					break;
+				case toolchain_component_static_linker:
+					list = static_linker_type_name;
+					count = static_linker_type_count;
+					break;
+				}
+
+				printf("  %s\n", toolchain_component_to_s(component));
+				for (i = 0; i < count; ++i) {
+					printf("    %s\n", list[i].id);
+				}
+			}
+			return true;
+		}
+
+		char *sep;
+		if (!(sep = strchr(optarg, '='))) {
+			LOG_E("invalid type: %s", optarg);
+			return false;
+		}
+
+		*sep = 0;
+
+		uint32_t component;
+		const char *type = sep + 1;
+
+		if (!toolchain_component_from_s(optarg, &component)) {
+			LOG_E("unknown toolchain component: %s", optarg);
+			return false;
+		}
+
+		switch ((enum toolchain_component)component) {
+		case toolchain_component_compiler:
+			if (!compiler_type_from_s(type, &comp.type[component])) {
+				LOG_E("unknown compiler type: %s", type);
+				return false;
+			}
+
+			if (!set_linker) {
+				comp.type[toolchain_component_linker] = compilers[comp.type[component]].default_linker;
+			}
+			if (!set_static_linker) {
+				comp.type[toolchain_component_static_linker]
+					= compilers[comp.type[component]].default_static_linker;
+			}
+			break;
+		case toolchain_component_linker:
+			set_linker = true;
+			if (!linker_type_from_s(type, &comp.type[component])) {
+				LOG_E("unknown linker type: %s", type);
+				return false;
+			}
+			break;
+		case toolchain_component_static_linker:
+			set_static_linker = true;
+			if (!static_linker_type_from_s(type, &comp.type[component])) {
+				LOG_E("unknown static_linker type: %s", type);
+				return false;
+			}
+			break;
+		}
+
+		break;
+	}
+	case 's': {
+		char *sep;
+		if (!(sep = strchr(optarg, '='))) {
+			LOG_E("invalid argument setting: %s", optarg);
+			return false;
+		}
+
+		*sep = 0;
+		++sep;
+
+		if (strcmp(optarg, "s1") == 0) {
+			opts.s1 = sep;
+		} else if (strcmp(optarg, "s2") == 0) {
+			opts.s2 = sep;
+		} else if (strcmp(optarg, "b1") == 0) {
+			if (strcmp(sep, "true") == 0) {
+				opts.b1 = true;
+			} else if (strcmp(sep, "false") == 0) {
+				opts.b1 = false;
+			} else {
+				LOG_E("invalid value for bool: %s", sep);
+				return false;
+			}
+		} else if (strcmp(optarg, "i1") == 0) {
+			int64_t res;
+			if (!str_to_i(&WKSTR(sep), &res, false)) {
+				LOG_E("invalid value for integer: %s", sep);
+				return false;
+			}
+
+			opts.i1 = res;
+		} else if (strcmp(optarg, "n1") == 0) {
+			n1.len = 0;
+
+			while (*sep) {
+				if (n1.len >= ARRAY_LEN(n1_args)) {
+					LOG_E("too many arguments for n1 value");
+					return false;
+				}
+
+				n1.args[n1.len] = sep;
+				++n1.len;
+
+				sep = strchr(sep, ',');
+				if (!sep) {
+					break;
+				}
+				*sep = 0;
+				++sep;
+			}
+		} else {
+			LOG_E("invalid setting name: %s", optarg);
+			return false;
+		}
+
+		break;
+	}
+	}
+	OPTEND(argv[argi],
+		"",
+		"  -t <component>=<type>|list - set the type for a component or list all component types\n"
+		"  -s <key>=<val> - set the value for a template argument\n",
+		NULL,
+		0)
+
+	struct workspace wk;
+	workspace_init(&wk);
+
+	obj id;
+	make_project(&wk, &id, "dummy", wk.source_root, wk.build_root);
+	if (!setup_project_options(&wk, NULL)) {
+		UNREACHABLE;
+	}
+
+	printf("compiler: %s, linker: %s, static_linker: %s\n",
+		compiler_type_name[comp.type[toolchain_component_compiler]].id,
+		linker_type_name[comp.type[toolchain_component_linker]].id,
+		static_linker_type_name[comp.type[toolchain_component_static_linker]].id);
+	printf("template arguments: s1: \"%s\", s2: \"%s\", b1: %s, i1: %d, n1: {",
+		opts.s1,
+		opts.s2,
+		opts.b1 ? "true" : "false",
+		opts.i1);
+	for (uint32_t i = 0; i < opts.n1->len; ++i) {
+		printf("\"%s\"", opts.n1->args[i]);
+		if (i + 1 < opts.n1->len) {
+			printf(", ");
+		}
+	}
+	printf("}\n");
+
+	toolchain_dump(&wk, &comp, &opts);
+
+	workspace_destroy(&wk);
+	return true;
+}
+
+static bool
+cmd_internal(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
 	static const struct command commands[] = {
+		{ "check", cmd_check, "check if a meson file parses" },
+		{ "dump_funcs", cmd_dump_signatures, "output all supported functions and arguments" },
+		{ "dump_toolchains", cmd_dump_toolchains, "output toolchain arguments" },
+		{ "dump_docs", cmd_dump_docs, "output docs" },
 		{ "eval", cmd_eval, "evaluate a file" },
 		{ "exe", cmd_exe, "run an external command" },
 		{ "repl", cmd_repl, "start a meson language repl" },
-		{ "dump_funcs", cmd_dump_signatures, "output all supported functions and arguments" },
+		{ "summary", cmd_summary, "print a configured project's summary" },
 		0,
 	};
 
@@ -680,33 +830,33 @@ cmd_internal(uint32_t argc, uint32_t argi, char *const argv[])
 	}
 	OPTEND(argv[argi], "", "", commands, -1);
 
-	cmd_func cmd = NULL;
-	;
-	if (!find_cmd(commands, &cmd, argc, argi, argv, false)) {
+	uint32_t cmd_i;
+	if (!find_cmd(commands, &cmd_i, argc, argi, argv, false)) {
 		return false;
 	}
 
-	assert(cmd);
-	return cmd(argc, argi, argv);
+	return commands[cmd_i].cmd(0, argc, argi, argv);
 }
 
 static bool
-cmd_samu(uint32_t argc, uint32_t argi, char *const argv[])
+cmd_samu(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
+	setup_platform_env(".", false);
+
 	return samu_main(argc - argi, (char **)&argv[argi], 0);
 }
 
 static bool
-cmd_test(uint32_t argc, uint32_t argi, char *const argv[])
+cmd_test(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
 	struct test_options test_opts = { 0 };
 
-	if (strcmp(argv[argi], "benchmark") == 0) {
+	OPTSTART("bs:d:Sfj:lvRe:o:") {
+	case 'b': {
 		test_opts.cat = test_category_benchmark;
 		test_opts.print_summary = true;
+		break;
 	}
-
-	OPTSTART("s:d:Sfj:lvRe:o:") {
 	case 'l': test_opts.list = true; break;
 	case 'e': test_opts.setup = optarg; break;
 	case 's':
@@ -760,6 +910,7 @@ cmd_test(uint32_t argc, uint32_t argi, char *const argv[])
 	}
 	OPTEND(argv[argi],
 		" [test [test [...]]]",
+		"  -b - run benchmarks instead of tests\n"
 		"  -d <mode> - change progress display mode (auto|dots|bar)\n"
 		"  -o <mode> - set output mode (term|html|json)\n"
 		"  -e <setup> - use test setup <setup>\n"
@@ -777,6 +928,8 @@ cmd_test(uint32_t argc, uint32_t argi, char *const argv[])
 		return false;
 	}
 
+	setup_platform_env(".", false);
+
 	test_opts.tests = &argv[argi];
 	test_opts.tests_len = argc - argi;
 
@@ -784,10 +937,10 @@ cmd_test(uint32_t argc, uint32_t argi, char *const argv[])
 }
 
 static bool
-cmd_install(uint32_t argc, uint32_t argi, char *const argv[])
+cmd_install(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
 	struct install_options opts = {
-		.destdir = getenv("DESTDIR"),
+		.destdir = os_get_env("DESTDIR"),
 	};
 
 	OPTSTART("nd:") {
@@ -808,8 +961,19 @@ cmd_install(uint32_t argc, uint32_t argi, char *const argv[])
 	return install_run(&opts);
 }
 
+static void
+cmd_setup_help(void)
+{
+	log_plain("\n");
+
+	struct list_options_opts list_opts = { 0 };
+	list_options(&list_opts);
+
+	log_plain("To see all options, including builtin options, use `muon options -a`.\n");
+}
+
 static bool
-cmd_setup(uint32_t argc, uint32_t argi, char *const argv[])
+cmd_setup(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
 	TracyCZoneAutoS;
 	bool res = false;
@@ -819,61 +983,48 @@ cmd_setup(uint32_t argc, uint32_t argi, char *const argv[])
 
 	uint32_t original_argi = argi + 1;
 
-	OPTSTART("D:c:b:") {
+	OPTSTART("D:b:") {
 	case 'D':
 		if (!parse_and_set_cmdline_option(&wk, optarg)) {
 			goto ret;
 		}
 		break;
-	case 'c': {
-		FILE *f;
-		if (!(f = fs_fopen(optarg, "rb:"))) {
-			goto ret;
-		} else if (!serial_load(&wk, &wk.compiler_check_cache, f)) {
-			LOG_E("failed to load compiler check cache");
-			goto ret;
-		} else if (!fs_fclose(f)) {
-			goto ret;
-		}
-		break;
-	}
 	case 'b': {
-		vm_dbg_push_breakpoint(&wk, optarg);
+		vm_dbg_push_breakpoint_str(&wk, optarg);
 		break;
 	}
 	}
-	OPTEND(argv[argi],
-		" <build dir>",
-		"  -D <option>=<value> - set project options\n"
-		"  -c <compiler_check_cache.dat> - path to compiler check cache dump\n"
-		"  -b <breakpoint> - set breakpoint\n",
-		NULL,
-		1)
+	OPTEND_CUSTOM(argv[argi], " <build dir>", "  -D <option>=<value> - set options\n", NULL, 1, cmd_setup_help())
 
 	const char *build = argv[argi];
 	++argi;
 
-	if (!workspace_setup_paths(&wk, build, argv[0], argc - original_argi, &argv[original_argi])) {
-		goto ret;
+	// Extract any relevant -D options that need to be handled very early.
+	// Currently this is only vsenv.  These haven't been added to any options
+	// dict yet so we need to manually scan the option_overrides array.
+	struct {
+		bool vsenv_force;
+	} opts = { 0 };
+	{
+		uint32_t i;
+		for (i = 0; i < wk.option_overrides.len; ++i) {
+			struct option_override *oo = arr_get(&wk.option_overrides, i);
+			if (oo->proj) {
+				continue;
+			}
+
+			if (str_eql(&WKSTR("vsenv"), get_str(&wk, oo->name))) {
+				opts.vsenv_force = str_eql(&WKSTR("true"), get_str(&wk, oo->val));
+			}
+		}
 	}
 
-	workspace_init_startup_files(&wk);
+	setup_platform_env(build, opts.vsenv_force);
 
-	uint32_t project_id;
-	if (!eval_project(&wk, NULL, wk.source_root, wk.build_root, &project_id)) {
+	if (!workspace_do_setup(&wk, build, argv[0], argc - original_argi, &argv[original_argi])) {
 		goto ret;
 	}
 
-	log_plain("\n");
-
-	if (!backend_output(&wk)) {
-		goto ret;
-	}
-
-	workspace_print_summaries(&wk, log_file());
-
-	LOG_I("setup complete");
-
 	res = true;
 ret:
 	workspace_destroy(&wk);
@@ -882,27 +1033,28 @@ ret:
 }
 
 static bool
-cmd_format(uint32_t argc, uint32_t argi, char *const argv[])
+cmd_format(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
-	if (strcmp(argv[argi], "fmt_unstable") == 0) {
-		LOG_W("the subcommand name fmt_unstable is deprecated, please use fmt instead");
-	}
-
 	struct {
 		char *const *filenames;
 		const char *cfg_path;
-		bool in_place, check_only, editorconfig;
+		bool in_place, check_only, editorconfig, print_failures;
 	} opts = { 0 };
 
-	OPTSTART("ic:qe") {
+	OPTSTART("ic:qel") {
 	case 'i': opts.in_place = true; break;
 	case 'c': opts.cfg_path = optarg; break;
 	case 'q': opts.check_only = true; break;
 	case 'e': opts.editorconfig = true; break;
+	case 'l':
+		opts.check_only = true;
+		opts.print_failures = true;
+		break;
 	}
 	OPTEND(argv[argi],
 		" <file>[ <file>[...]]",
 		"  -q - exit with 1 if files would be modified by muon fmt\n"
+		"  -l - like -q but also print failing filenames\n"
 		"  -i - format files in-place\n"
 		"  -c <muon_fmt.ini> - read configuration from muon_fmt.ini\n"
 		"  -e - try to read configuration from .editorconfig\n",
@@ -914,6 +1066,8 @@ cmd_format(uint32_t argc, uint32_t argi, char *const argv[])
 		return false;
 	}
 
+	log_set_file(stderr);
+
 	opts.filenames = &argv[argi];
 	const uint32_t num_files = argc - argi;
 
@@ -944,6 +1098,9 @@ cmd_format(uint32_t argc, uint32_t argi, char *const argv[])
 		}
 
 		fmt_ret = fmt(&src, out, opts.cfg_path, opts.check_only, opts.editorconfig);
+		if (!fmt_ret && opts.print_failures) {
+			printf("%s\n", opts.filenames[i]);
+		}
 cont:
 		if (opened_out) {
 			fs_fclose(out);
@@ -960,7 +1117,7 @@ cont:
 }
 
 static bool
-cmd_version(uint32_t argc, uint32_t argi, char *const argv[])
+cmd_version(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
 	printf("muon %s%s%s\nmeson compatibility version %s\nenabled features:\n",
 		muon_version.version,
@@ -991,10 +1148,10 @@ cmd_version(uint32_t argc, uint32_t argi, char *const argv[])
 	return true;
 }
 
-static bool cmd_main(uint32_t argc, uint32_t argi, char *argv[]);
+static bool cmd_main(void *_ctx, uint32_t argc, uint32_t argi, char *argv[]);
 
 static bool
-cmd_meson(uint32_t argc, uint32_t argi, char *const argv[])
+cmd_meson(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
 	++argi;
 
@@ -1010,7 +1167,7 @@ cmd_meson(uint32_t argc, uint32_t argi, char *const argv[])
 	argc = new_argc;
 	argv = new_argv;
 
-	bool res = cmd_main(argc, argi, (char **)argv);
+	bool res = cmd_main(0, argc, argi, (char **)argv);
 
 	workspace_destroy(&wk);
 	z_free(new_argv);
@@ -1019,24 +1176,31 @@ cmd_meson(uint32_t argc, uint32_t argi, char *const argv[])
 }
 
 static bool
-cmd_main(uint32_t argc, uint32_t argi, char *argv[])
+cmd_ui(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
+{
+	OPTSTART("") {
+	}
+	OPTEND(argv[argi], "", "", 0, 0);
+
+	return ui_main();
+}
+
+static bool
+cmd_main(void *_ctx, uint32_t argc, uint32_t argi, char *argv[])
 {
 	const struct command commands[] = {
-		{ "analyze", cmd_analyze, "run a static analyzer on the current project." },
-		{ "benchmark", cmd_test, "run benchmarks" },
-		{ "check", cmd_check, "check if a meson file parses" },
+		{ "analyze", cmd_analyze, "run a static analyzer" },
 		{ "fmt", cmd_format, "format meson source file" },
-		{ "fmt_unstable", cmd_format, NULL },
 		{ "info", cmd_info, NULL },
-		{ "install", cmd_install, "install project" },
+		{ "install", cmd_install, "install files" },
 		{ "internal", cmd_internal, "internal subcommands" },
-		{ "meson", cmd_meson, "meson compatible cli proxy" },
+		{ "meson", cmd_meson, NULL },
 		{ "options", cmd_options, "list project options" },
 		{ "samu", cmd_samu, have_samurai ? "run samurai" : NULL },
 		{ "setup", cmd_setup, "setup a build directory" },
 		{ "subprojects", cmd_subprojects, "manage subprojects" },
-		{ "summary", cmd_summary, "print a configured project's summary" },
 		{ "test", cmd_test, "run tests" },
+		{ "ui", cmd_ui, have_ui ? "run an interactive ui" : NULL },
 		{ "version", cmd_version, "print version information" },
 		{ 0 },
 	};
@@ -1067,12 +1231,12 @@ cmd_main(uint32_t argc, uint32_t argi, char *argv[])
 		commands,
 		-1)
 
-	cmd_func cmd;
-	if (!find_cmd(commands, &cmd, argc, argi, argv, false)) {
+	uint32_t cmd_i;
+	if (!find_cmd(commands, &cmd_i, argc, argi, argv, false)) {
 		goto ret;
 	}
 
-	res = cmd(argc, argi, argv);
+	res = commands[cmd_i].cmd(0, argc, argi, argv);
 
 ret:
 	sbuf_destroy(&argv0);
@@ -1090,6 +1254,7 @@ main(int argc, char *argv[])
 	path_init();
 
 	compilers_init();
+	machine_init();
 
 	bool res;
 	bool meson_compat = false;
@@ -1102,9 +1267,9 @@ main(int argc, char *argv[])
 	}
 
 	if (meson_compat) {
-		res = cmd_meson(argc, 0, argv);
+		res = cmd_meson(0, argc, 0, argv);
 	} else {
-		res = cmd_main(argc, 0, argv);
+		res = cmd_main(0, argc, 0, argv);
 	}
 
 	int ret = res ? 0 : 1;
