@@ -10,19 +10,25 @@
 #include <string.h>
 
 #include "backend/output.h"
+#include "buf_size.h"
 #include "embedded.h"
 #include "error.h"
+#include "functions/modules/subprojects.h"
+#include "lang/analyze.h"
 #include "lang/object_iterators.h"
 #include "lang/serial.h"
 #include "lang/typecheck.h"
 #include "log.h"
 #include "options.h"
+#include "platform/assert.h"
 #include "platform/filesystem.h"
+#include "platform/os.h"
 #include "platform/path.h"
+#include "wrap.h"
 
 bool initializing_builtin_options = false;
 
-static const char *build_option_type_to_s[build_option_type_count] = {
+const char *build_option_type_to_s[build_option_type_count] = {
 	[op_string] = "string",
 	[op_boolean] = "boolean",
 	[op_combo] = "combo",
@@ -31,9 +37,6 @@ static const char *build_option_type_to_s[build_option_type_count] = {
 	[op_feature] = "feature",
 };
 
-static bool
-set_option(struct workspace *wk, uint32_t node, obj opt, obj new_val, enum option_value_source source, bool coerce);
-
 static bool
 parse_config_string(struct workspace *wk, const struct str *ss, struct option_override *oo, bool key_only)
 {
@@ -124,7 +127,7 @@ log_option_override(struct workspace *wk, struct option_override *oo)
 		log_plain("%s:", get_cstr(wk, oo->proj));
 	}
 
-	obj_fprintf(wk, log_file(), "%s=%#o", get_cstr(wk, oo->name), oo->val);
+	obj_lprintf(wk, "%s=%#o", get_cstr(wk, oo->name), oo->val);
 	log_plain("'");
 }
 
@@ -168,7 +171,7 @@ check_invalid_subproject_option(struct workspace *wk)
 }
 
 static bool
-coerce_feature_opt(struct workspace *wk, const struct str *val, obj *res)
+coerce_feature_opt(struct workspace *wk, uint32_t ip, const struct str *val, obj *res)
 {
 	enum feature_opt_state f;
 
@@ -179,7 +182,7 @@ coerce_feature_opt(struct workspace *wk, const struct str *val, obj *res)
 	} else if (str_eql(val, &WKSTR("disabled"))) {
 		f = feature_opt_disabled;
 	} else {
-		vm_error(wk, "unable to coerce '%s' into a feature", val->s);
+		vm_error_at(wk, ip, "unable to coerce '%s' into a feature", val->s);
 		return false;
 	}
 
@@ -246,7 +249,7 @@ check_deprecated_option(struct workspace *wk, struct obj_option *opt, obj sval,
 
 		obj newopt;
 		if (get_option(wk, cur_proj, get_str(wk, opt->deprecated), &newopt)) {
-			set_option(wk, 0, newopt, sval, option_value_source_deprecated_rename, true);
+			set_option(wk, newopt, sval, option_value_source_deprecated_rename, true);
 		} else {
 			struct option_override oo = {
 				.proj = current_project(wk)->cfg.name,
@@ -312,12 +315,11 @@ coerce_option_override(struct workspace *wk, struct obj_option *opt, obj sval, o
 		} else if (str_eql(val, &WKSTR("false"))) {
 			b = false;
 		} else {
-			vm_error(wk, "unable to coerce '%s' into a boolean", val->s);
+			vm_error_at(wk, opt->ip, "unable to coerce '%s' into a boolean", val->s);
 			return false;
 		}
 
-		make_obj(wk, res, obj_bool);
-		set_obj_bool(wk, *res, b);
+		*res = make_obj_bool(wk, b);
 		break;
 	}
 	case op_integer: {
@@ -338,7 +340,7 @@ coerce_option_override(struct workspace *wk, struct obj_option *opt, obj sval, o
 		// do nothing, array values were already coerced above
 		break;
 	}
-	case op_feature: return coerce_feature_opt(wk, val, res);
+	case op_feature: return coerce_feature_opt(wk, opt->ip, val, res);
 	default: UNREACHABLE;
 	}
 
@@ -351,7 +353,7 @@ typecheck_opt(struct workspace *wk, uint32_t node, obj val, enum build_option_ty
 	enum obj_type expected_type;
 
 	if (type == op_feature && get_obj_type(wk, val) == obj_string) {
-		if (!coerce_feature_opt(wk, get_str(wk, val), res)) {
+		if (!coerce_feature_opt(wk, node, get_str(wk, val), res)) {
 			return false;
 		}
 
@@ -437,8 +439,8 @@ extend_array_option(struct workspace *wk, obj opt, obj new_val, enum option_valu
 	obj_array_extend_nodup(wk, o->val, new_val);
 }
 
-static bool
-set_option(struct workspace *wk, uint32_t node, obj opt, obj new_val, enum option_value_source source, bool coerce)
+bool
+set_option(struct workspace *wk, obj opt, obj new_val, enum option_value_source source, bool coerce)
 {
 	struct obj_option *o = get_obj_option(wk, opt);
 
@@ -477,7 +479,7 @@ set_option(struct workspace *wk, uint32_t node, obj opt, obj new_val, enum optio
 	o->source = source;
 
 	if (get_obj_type(wk, o->deprecated) == obj_bool && get_obj_bool(wk, o->deprecated)) {
-		vm_warning_at(wk, node, "option %o is deprecated", o->name);
+		vm_warning_at(wk, o->ip, "option %o is deprecated", o->name);
 	}
 
 	if (coerce) {
@@ -488,14 +490,14 @@ set_option(struct workspace *wk, uint32_t node, obj opt, obj new_val, enum optio
 		new_val = coerced;
 	}
 
-	if (!typecheck_opt(wk, node, new_val, o->type, o->name, &new_val)) {
+	if (!typecheck_opt(wk, o->ip, new_val, o->type, o->name, &new_val)) {
 		return false;
 	}
 
 	switch (o->type) {
 	case op_combo: {
 		if (!obj_array_in(wk, o->choices, new_val)) {
-			vm_error_at(wk, node, "'%o' is not one of %o", new_val, o->choices);
+			vm_error_at(wk, o->ip, "'%o' is not one of %o", new_val, o->choices);
 			return false;
 		}
 		break;
@@ -505,7 +507,7 @@ set_option(struct workspace *wk, uint32_t node, obj opt, obj new_val, enum optio
 
 		if ((o->max && num > get_obj_number(wk, o->max)) || (o->min && num < get_obj_number(wk, o->min))) {
 			vm_error_at(wk,
-				node,
+				o->ip,
 				"value %" PRId64 " is out of range (%" PRId64 "..%" PRId64 ")",
 				get_obj_number(wk, new_val),
 				(o->min ? get_obj_number(wk, o->min) : INT64_MIN),
@@ -519,7 +521,7 @@ set_option(struct workspace *wk, uint32_t node, obj opt, obj new_val, enum optio
 			obj val;
 			obj_array_for(wk, new_val, val) {
 				if (!obj_array_in(wk, o->choices, val)) {
-					vm_error_at(wk, node, "array element %o is not one of %o", val, o->choices);
+					vm_error_at(wk, o->ip, "array element %o is not one of %o", val, o->choices);
 					return false;
 				}
 			}
@@ -539,7 +541,7 @@ set_option(struct workspace *wk, uint32_t node, obj opt, obj new_val, enum optio
 bool
 create_option(struct workspace *wk, obj opts, obj opt, obj val)
 {
-	if (!set_option(wk, 0, opt, val, option_value_source_default, false)) {
+	if (!set_option(wk, opt, val, option_value_source_default, false)) {
 		return false;
 	}
 
@@ -616,13 +618,13 @@ set_binary_from_env(struct workspace *wk, const char *envvar, const char *dest)
 	}
 
 	const char *v;
-	if (!(v = getenv(envvar)) || !*v) {
+	if (!(v = os_get_env(envvar)) || !*v) {
 		return;
 	}
 
 	// TODO: implement something like shlex.split()
 	obj cmd = str_split(wk, &WKSTR(v), NULL);
-	set_option(wk, 0, opt, cmd, option_value_source_environment, false);
+	set_option(wk, opt, cmd, option_value_source_environment, false);
 }
 
 static void
@@ -633,11 +635,11 @@ set_compile_opt_from_env(struct workspace *wk, const char *name, const char *fla
 		UNREACHABLE;
 	}
 
-	if ((flags = getenv(flags)) && *flags) {
+	if ((flags = os_get_env(flags)) && *flags) {
 		extend_array_option(wk, opt, str_split(wk, &WKSTR(flags), NULL), option_value_source_environment);
 	}
 
-	if ((extra = getenv(extra)) && *extra) {
+	if ((extra = os_get_env(extra)) && *extra) {
 		extend_array_option(wk, opt, str_split(wk, &WKSTR(extra), NULL), option_value_source_environment);
 	}
 }
@@ -651,24 +653,24 @@ set_str_opt_from_env(struct workspace *wk, const char *env_name, const char *opt
 	}
 
 	const char *env_val;
-	if ((env_val = getenv(env_name)) && *env_val) {
-		set_option(wk, 0, opt, make_str(wk, env_val), option_value_source_environment, false);
+	if ((env_val = os_get_env(env_name)) && *env_val) {
+		set_option(wk, opt, make_str(wk, env_val), option_value_source_environment, false);
 	}
 }
 
 static bool
-init_builtin_options(struct workspace *wk, const char *script, const char *fallback)
+init_builtin_options(struct workspace *wk, const char *script)
 {
-	const char *opts;
-	if (!(opts = embedded_get(script))) {
-		opts = fallback;
+	struct source src;
+	if (!embedded_get(script, &src)) {
+		return false;
 	}
 
 	enum language_mode old_mode = wk->vm.lang_mode;
 	wk->vm.lang_mode = language_opts;
 	obj _;
 	initializing_builtin_options = true;
-	bool ret = eval_str_label(wk, script, opts, eval_mode_default, &_);
+	bool ret = eval(wk, &src, build_language_meson, 0, &_);
 	initializing_builtin_options = false;
 	wk->vm.lang_mode = old_mode;
 	return ret;
@@ -677,11 +679,7 @@ init_builtin_options(struct workspace *wk, const char *script, const char *fallb
 static bool
 init_per_project_options(struct workspace *wk)
 {
-	return init_builtin_options(wk,
-		"per_project_options.meson",
-		"option('default_library', type: 'string', value: 'static')\n"
-		"option('warning_level', type: 'string', value: '3')\n"
-		"option('c_std', type: 'string', value: 'c99')\n");
+	return init_builtin_options(wk, "options/per_project.meson");
 }
 
 static enum iteration_result
@@ -709,12 +707,34 @@ set_yielding_project_options_iter(struct workspace *wk, void *_ctx, obj _k, obj
 		return ir_cont;
 	}
 
-	if (!set_option(wk, 0, opt, po->val, option_value_source_yield, false)) {
+	if (!set_option(wk, opt, po->val, option_value_source_yield, false)) {
 		return ir_err;
 	}
 	return ir_cont;
 }
 
+static bool
+determine_option_file(struct workspace *wk, const char *cwd, struct sbuf *out)
+{
+	const char *option_file_names[] = {
+		"meson.options",
+		"meson_options.txt",
+	};
+
+	bool exists = false;
+	uint32_t i;
+	for (i = 0; i < ARRAY_LEN(option_file_names); ++i) {
+		path_join(wk, out, cwd, option_file_names[i]);
+
+		if (fs_file_exists(out->buf)) {
+			exists = true;
+			break;
+		}
+	}
+
+	return exists;
+}
+
 bool
 setup_project_options(struct workspace *wk, const char *cwd)
 {
@@ -726,28 +746,13 @@ setup_project_options(struct workspace *wk, const char *cwd)
 		return true;
 	}
 
-	const char *option_file_names[] = {
-		"meson.options",
-		"meson_options.txt",
-	};
-
 	SBUF(meson_opts);
-
-	bool exists = false;
-	uint32_t i;
-	for (i = 0; i < ARRAY_LEN(option_file_names); ++i) {
-		path_join(wk, &meson_opts, cwd, option_file_names[i]);
-
-		if (fs_file_exists(meson_opts.buf)) {
-			exists = true;
-			break;
-		}
-	}
+	bool exists = determine_option_file(wk, cwd, &meson_opts);
 
 	if (exists) {
 		enum language_mode old_mode = wk->vm.lang_mode;
 		wk->vm.lang_mode = language_opts;
-		if (!wk->vm.behavior.eval_project_file(wk, meson_opts.buf, false)) {
+		if (!wk->vm.behavior.eval_project_file(wk, meson_opts.buf, build_language_meson, 0)) {
 			wk->vm.lang_mode = old_mode;
 			return false;
 		}
@@ -768,6 +773,7 @@ setup_project_options(struct workspace *wk, const char *cwd)
 	bool ret = true;
 	struct option_override *oo;
 
+	uint32_t i;
 	for (i = 0; i < wk->option_overrides.len; ++i) {
 		oo = arr_get(&wk->option_overrides, i);
 
@@ -780,7 +786,7 @@ setup_project_options(struct workspace *wk, const char *cwd)
 		obj opt;
 		if (obj_dict_index_strn(wk, current_project(wk)->opts, name->s, name->len, &opt)
 			|| (is_master_project && obj_dict_index_strn(wk, wk->global_opts, name->s, name->len, &opt))) {
-			if (!set_option(wk, 0, opt, oo->val, oo->source, !oo->obj_value)) {
+			if (!set_option(wk, opt, oo->val, oo->source, !oo->obj_value)) {
 				ret = false;
 			}
 		} else {
@@ -797,46 +803,23 @@ setup_project_options(struct workspace *wk, const char *cwd)
 bool
 init_global_options(struct workspace *wk)
 {
-	if (!init_builtin_options(wk,
-		    "global_options.meson",
-		    "option('buildtype', type: 'string', value: 'debugoptimized')\n"
-		    "option('prefix', type: 'string', value: '/usr/local')\n"
-		    "option('bindir', type: 'string', value: 'bin')\n"
-		    "option('mandir', type: 'string', value: 'share/man')\n"
-		    "option('datadir', type: 'string', value: 'share')\n"
-		    "option('libdir', type: 'string', value: 'lib')\n"
-		    "option('includedir', type: 'string', value: 'include')\n"
-		    "option('wrap_mode', type: 'string', value: 'nopromote')\n"
-		    "option('force_fallback_for', type: 'array', value: [])\n"
-		    "option('pkg_config_path', type: 'string', value: '')\n"
-		    "option('c_args', type: 'array', value: [])\n"
-		    "option('c_link_args', type: 'array', value: [])\n"
-		    "option('werror', type: 'boolean', value: false)\n"
-		    "option('prefer_static', type: 'boolean', value: false)\n"
-
-		    "option('env.CC', type: 'array', value: ['cc'])\n"
-		    "option('env.NINJA', type: 'array', value: ['ninja'])\n"
-		    "option('env.AR', type: 'array', value: ['ar'])\n"
-		    "option('env.LD', type: 'array', value: ['ld'])\n")) {
+	if (!init_builtin_options(wk, "options/global.meson")) {
 		return false;
 	}
 
-	set_binary_from_env(wk, "CC", "env.CC");
-	set_binary_from_env(wk, "NINJA", "env.NINJA");
 	set_binary_from_env(wk, "AR", "env.AR");
-	set_binary_from_env(wk, "LD", "env.LD");
-	set_compile_opt_from_env(wk, "c_args", "CFLAGS", "CPPFLAGS");
-	set_compile_opt_from_env(wk, "c_link_args", "CFLAGS", "LDFLAGS");
-
-#ifdef MUON_BOOTSTRAPPED
+	set_binary_from_env(wk, "CC", "env.CC");
 	set_binary_from_env(wk, "CXX", "env.CXX");
+	set_binary_from_env(wk, "LD", "env.LD");
+	set_binary_from_env(wk, "NASM", "env.NASM");
+	set_binary_from_env(wk, "NINJA", "env.NINJA");
 	set_binary_from_env(wk, "OBJC", "env.OBJC");
 	set_binary_from_env(wk, "OBJCPP", "env.OBJCPP");
-	set_binary_from_env(wk, "NASM", "env.NASM");
+	set_compile_opt_from_env(wk, "c_args", "CFLAGS", "CPPFLAGS");
+	set_compile_opt_from_env(wk, "c_link_args", "CFLAGS", "LDFLAGS");
 	set_compile_opt_from_env(wk, "cpp_args", "CXXFLAGS", "CPPFLAGS");
 	set_compile_opt_from_env(wk, "cpp_link_args", "CXXFLAGS", "LDFLAGS");
 	set_str_opt_from_env(wk, "PKG_CONFIG_PATH", "pkg_config_path");
-#endif
 
 	return true;
 }
@@ -924,7 +907,7 @@ parse_and_set_option(struct workspace *wk, const struct parse_and_set_option_par
 		opt = newopt;
 	}
 
-	if (!set_option(wk, params->node, opt, oo.val, oo.source, !oo.obj_value)) {
+	if (!set_option(wk, opt, oo.val, oo.source, !oo.obj_value)) {
 		return false;
 	}
 
@@ -1066,6 +1049,21 @@ get_option_bool(struct workspace *wk, obj overrides, const char *name, bool fall
 	}
 }
 
+enum backend
+get_option_backend(struct workspace *wk)
+{
+	obj backend_opt;
+	get_option_value(wk, 0, "backend", &backend_opt);
+	const struct str *backend_str = get_str(wk, backend_opt);
+	if (str_eql(backend_str, &WKSTR("ninja"))) {
+		return backend_ninja;
+	} else if (str_eql(backend_str, &WKSTR("xcode"))) {
+		return backend_xcode;
+	} else {
+		UNREACHABLE;
+	}
+}
+
 /* options listing subcommand */
 
 struct make_option_choices_ctx {
@@ -1102,6 +1100,7 @@ make_option_choices_iter(struct workspace *wk, void *_ctx, obj val)
 struct list_options_ctx {
 	bool show_builtin;
 	const struct list_options_opts *list_opts;
+	const char *subproject_name;
 };
 
 static enum iteration_result
@@ -1134,7 +1133,10 @@ list_options_iter(struct workspace *wk, void *_ctx, obj key, obj val)
 		no_clr = "\033[0m";
 	}
 
-	obj_fprintf(wk, log_file(), "  %s%#o%s=", key_clr, key, no_clr);
+	{
+		const char *subp = ctx->subproject_name;
+		obj_lprintf(wk, "  -D %s%s%s%#o%s=", subp ? subp : "", subp ? ":" : "", key_clr, key, no_clr);
+	}
 
 	obj choices = 0;
 	obj selected = 0;
@@ -1194,16 +1196,11 @@ list_options_iter(struct workspace *wk, void *_ctx, obj key, obj val)
 	switch (opt->type) {
 	case op_boolean:
 	case op_combo:
-	case op_feature: obj_fprintf(wk, log_file(), "<%s>", get_cstr(wk, choices)); break;
+	case op_feature: obj_lprintf(wk, "<%s>", get_cstr(wk, choices)); break;
 	case op_string: {
 		const struct str *def = get_str(wk, opt->val);
-		obj_fprintf(wk,
-			log_file(),
-			"<%s>, default: %s%s%s",
-			get_cstr(wk, choices),
-			sel_clr,
-			def->len ? def->s : "''",
-			no_clr);
+		obj_lprintf(
+			wk, "<%s>, default: %s%s%s", get_cstr(wk, choices), sel_clr, def->len ? def->s : "''", no_clr);
 		break;
 	}
 	case op_integer:
@@ -1211,37 +1208,68 @@ list_options_iter(struct workspace *wk, void *_ctx, obj key, obj val)
 		if (opt->min || opt->max) {
 			log_plain(" where ");
 			if (opt->min) {
-				obj_fprintf(wk, log_file(), "%o <= ", opt->min);
+				obj_lprintf(wk, "%o <= ", opt->min);
 			}
 			log_plain("%sN%s", val_clr, no_clr);
 			if (opt->max) {
-				obj_fprintf(wk, log_file(), " <= %o", opt->max);
+				obj_lprintf(wk, " <= %o", opt->max);
 			}
 		}
-		obj_fprintf(wk, log_file(), ", default: %s%o%s", sel_clr, opt->val, no_clr);
+		obj_lprintf(wk, ", default: %s%o%s", sel_clr, opt->val, no_clr);
 
 		break;
 	case op_array:
 		log_plain("<%svalue%s[,%svalue%s[...]]>", val_clr, no_clr, val_clr, no_clr);
 		if (opt->choices) {
-			obj_fprintf(wk, log_file(), " where value in %s", get_cstr(wk, choices));
+			obj_lprintf(wk, " where value in %s", get_cstr(wk, choices));
 		}
 		break;
 	default: UNREACHABLE;
 	}
 
 	if (opt->source != option_value_source_default) {
-		obj_fprintf(wk, log_file(), "*");
+		obj_lprintf(wk, "*");
 	}
 
 	if (opt->description) {
-		obj_fprintf(wk, log_file(), "\n    %#o", opt->description);
+		obj_lprintf(wk, " - %#o", opt->description);
 	}
 
 	log_plain("\n");
 	return ir_cont;
 }
 
+static enum iteration_result
+list_options_for_subproject(struct workspace *wk, struct subprojects_common_ctx *ctx, const char *path)
+{
+	SBUF(cwd);
+
+	struct wrap wrap = { 0 };
+	if (!wrap_parse(path, &wrap)) {
+		return ir_err;
+	}
+
+	SBUF(meson_opts);
+	bool exists = determine_option_file(wk, wrap.dest_dir.buf, &meson_opts);
+
+	make_project(wk, &wk->cur_project, wrap.name.buf, wrap.dest_dir.buf, "");
+	current_project(wk)->cfg.name = make_str(wk, wrap.name.buf);
+
+	if (exists) {
+		if (!wk->vm.behavior.eval_project_file(wk, meson_opts.buf, build_language_meson, 0)) {
+			goto cont;
+		}
+	} else {
+		make_obj(wk, &current_project(wk)->opts, obj_dict);
+	}
+
+cont:
+	wrap_destroy(&wrap);
+	wk->cur_project = 0;
+
+	return ir_cont;
+}
+
 bool
 list_options(const struct list_options_opts *list_opts)
 {
@@ -1250,22 +1278,40 @@ list_options(const struct list_options_opts *list_opts)
 	workspace_init(&wk);
 	wk.vm.lang_mode = language_opts;
 
-	arr_push(&wk.projects, &(struct project){ 0 });
-	struct project *proj = arr_get(&wk.projects, 0);
-	make_obj(&wk, &proj->opts, obj_dict);
+	uint32_t idx;
+	make_project(&wk, &idx, "dummy", path_cwd(), "");
 
 	if (fs_file_exists("meson.build")) {
 		SBUF(meson_opts);
-		path_make_absolute(&wk, &meson_opts, "meson_options.txt");
+		bool exists = determine_option_file(&wk, ".", &meson_opts);
 
-		if (fs_file_exists(meson_opts.buf)) {
-			if (!wk.vm.behavior.eval_project_file(&wk, meson_opts.buf, false)) {
+		if (exists) {
+			if (!wk.vm.behavior.eval_project_file(&wk, meson_opts.buf, build_language_meson, 0)) {
 				goto ret;
 			}
+		} else {
+			make_obj(&wk, &current_project(&wk)->opts, obj_dict);
+		}
+
+		{
+			SBUF(subprojects_dir);
+			struct workspace az_wk = { 0 };
+			analyze_project_call(&az_wk);
+			path_make_absolute(&wk, &subprojects_dir, get_cstr(&az_wk, current_project(&az_wk)->subprojects_dir));
+
+			obj name = current_project(&az_wk)->cfg.name;
+			if (name) {
+				current_project(&wk)->cfg.name = make_str(&wk, get_cstr(&az_wk, name));
+			}
+
+			current_project(&wk)->subprojects_dir = sbuf_into_str(&wk, &subprojects_dir);
+
+			workspace_destroy(&az_wk);
+
+			subprojects_foreach(&wk, 0, 0, list_options_for_subproject);
 		}
 
 		if (list_opts->list_all) {
-			make_obj(&wk, &current_project(&wk)->opts, obj_dict);
 			if (!init_per_project_options(&wk)) {
 				goto ret;
 			}
@@ -1289,18 +1335,34 @@ list_options(const struct list_options_opts *list_opts)
 
 	struct list_options_ctx ctx = { .list_opts = list_opts };
 
-	if (get_obj_dict(&wk, current_project(&wk)->opts)->len) {
-		log_plain("project options:\n");
-		obj_dict_foreach(&wk, current_project(&wk)->opts, &ctx, list_options_iter);
-	} else if (!list_opts->list_all) {
+	bool had_project_options = false;
+
+	uint32_t i;
+	for (i = 0; i < wk.projects.len; ++i) {
+		struct project *proj = arr_get(&wk.projects, i);
+		if (get_obj_dict(&wk, proj->opts)->len) {
+			had_project_options = true;
+
+			const char *name = get_cstr(&wk, proj->cfg.name);
+			if (!name) {
+				name = "project";
+			}
+
+			log_plain("%s options:\n", name);
+
+			ctx.subproject_name = i == 0 ? 0 : name;
+			obj_dict_foreach(&wk, proj->opts, &ctx, list_options_iter);
+			ctx.subproject_name = 0;
+
+			log_plain("\n");
+		}
+	}
+
+	if (!had_project_options && !list_opts->list_all) {
 		log_plain("no project options defined\n");
 	}
 
 	if (list_opts->list_all) {
-		if (get_obj_dict(&wk, current_project(&wk)->opts)->len) {
-			log_plain("\n");
-		}
-
 		ctx.show_builtin = true;
 
 		log_plain("project builtin-options:\n");
