@@ -5,29 +5,34 @@
 
 #include "compat.h"
 
-#include <assert.h>
 #include <errno.h>
 #include <string.h>
-#include <unistd.h>
 
 #include "buf_size.h"
 #include "error.h"
 #include "lang/string.h"
 #include "log.h"
+#include "platform/assert.h"
+#include "platform/os.h"
 #include "platform/path.h"
 
 static struct {
-	char cwd_buf[BUF_SIZE_2k],
-	     tmp1_buf[BUF_SIZE_2k],
-	     tmp2_buf[BUF_SIZE_2k];
+	char cwd_buf[BUF_SIZE_2k], tmp1_buf[BUF_SIZE_2k], tmp2_buf[BUF_SIZE_2k];
 	struct sbuf cwd, tmp1, tmp2;
 } path_ctx;
 
+// These functions are defined in platform/<plat>/os.c
+//
+// They should only be called indirectly through path_chdir and path_cwd though
+// so the prototypes were removed from os.h.
+bool os_chdir(const char *path);
+char *os_getcwd(char *buf, size_t size);
+
 static void
 path_getcwd(void)
 {
 	sbuf_clear(&path_ctx.cwd);
-	while (!getcwd(path_ctx.cwd.buf, path_ctx.cwd.cap)) {
+	while (!os_getcwd(path_ctx.cwd.buf, path_ctx.cwd.cap)) {
 		if (errno == ERANGE) {
 			sbuf_grow(NULL, &path_ctx.cwd, path_ctx.cwd.cap);
 		} else {
@@ -39,12 +44,9 @@ path_getcwd(void)
 void
 path_init(void)
 {
-	sbuf_init(&path_ctx.cwd, path_ctx.cwd_buf, ARRAY_LEN(path_ctx.cwd_buf),
-		sbuf_flag_overflow_alloc);
-	sbuf_init(&path_ctx.tmp1, path_ctx.tmp1_buf, ARRAY_LEN(path_ctx.tmp1_buf),
-		sbuf_flag_overflow_alloc);
-	sbuf_init(&path_ctx.tmp2, path_ctx.tmp2_buf, ARRAY_LEN(path_ctx.tmp2_buf),
-		sbuf_flag_overflow_alloc);
+	sbuf_init(&path_ctx.cwd, path_ctx.cwd_buf, ARRAY_LEN(path_ctx.cwd_buf), sbuf_flag_overflow_alloc);
+	sbuf_init(&path_ctx.tmp1, path_ctx.tmp1_buf, ARRAY_LEN(path_ctx.tmp1_buf), sbuf_flag_overflow_alloc);
+	sbuf_init(&path_ctx.tmp2, path_ctx.tmp2_buf, ARRAY_LEN(path_ctx.tmp2_buf), sbuf_flag_overflow_alloc);
 	path_getcwd();
 }
 
@@ -64,7 +66,6 @@ _path_normalize(struct workspace *wk, struct sbuf *buf, bool optimize)
 	uint32_t part_len, i, slen = buf->len, blen = 0, sep_len;
 	bool loop = true, skip_part;
 
-	/* assert(slen && "path is empty"); */
 	if (!buf->len) {
 		return;
 	}
@@ -148,7 +149,7 @@ path_copy(struct workspace *wk, struct sbuf *sb, const char *path)
 bool
 path_chdir(const char *path)
 {
-	if (chdir(path) < 0) {
+	if (!os_chdir(path)) {
 		LOG_E("failed chdir(%s): %s", path, strerror(errno));
 		return false;
 	}
@@ -158,11 +159,17 @@ path_chdir(const char *path)
 }
 
 void
-path_cwd(struct workspace *wk, struct sbuf *sb)
+path_copy_cwd(struct workspace *wk, struct sbuf *sb)
 {
 	path_copy(wk, sb, path_ctx.cwd.buf);
 }
 
+const char *
+path_cwd(void)
+{
+	return path_ctx.cwd.buf;
+}
+
 void
 path_join_absolute(struct workspace *wk, struct sbuf *sb, const char *a, const char *b)
 {
@@ -226,8 +233,14 @@ path_relative_to(struct workspace *wk, struct sbuf *buf, const char *base_raw, c
 	 */
 
 	sbuf_clear(buf);
-	path_copy(wk, &path_ctx.tmp1, base_raw);
-	path_copy(wk, &path_ctx.tmp2, path_raw);
+
+	sbuf_clear(&path_ctx.tmp1);
+	sbuf_pushs(wk, &path_ctx.tmp1, base_raw);
+	_path_normalize(wk, &path_ctx.tmp1, true);
+
+	sbuf_clear(&path_ctx.tmp2);
+	sbuf_pushs(wk, &path_ctx.tmp2, path_raw);
+	_path_normalize(wk, &path_ctx.tmp2, true);
 
 	const char *base = path_ctx.tmp1.buf, *path = path_ctx.tmp2.buf;
 
@@ -260,9 +273,8 @@ path_relative_to(struct workspace *wk, struct sbuf *buf, const char *base_raw, c
 		common_end = i;
 	}
 
-	assert(i);
-	if (i == 1) {
-		/* -> base and path match only at root */
+	if (i <= 1) {
+		/* -> base and path match only at root, so take path */
 		path_copy(wk, buf, path);
 		return;
 	}
