@@ -16,8 +16,10 @@
 #include "buf_size.h"
 #include "lang/string.h"
 #include "log.h"
+#include "platform/assert.h"
 #include "platform/filesystem.h"
 #include "platform/mem.h"
+#include "platform/os.h"
 #include "platform/path.h"
 
 static bool
@@ -189,7 +191,7 @@ ret:
 }
 
 bool
-fs_copy_file(const char *src, const char *dest)
+fs_copy_file(const char *src, const char *dest, bool force)
 {
 	bool res = false;
 	FILE *f_src = NULL;
@@ -207,6 +209,10 @@ fs_copy_file(const char *src, const char *dest)
 		goto ret;
 	}
 
+	if (force) {
+		fs_make_writeable_if_exists(dest);
+	}
+
 	if (!(f_src = fs_fopen(src, "r"))) {
 		goto ret;
 	}
@@ -327,7 +333,7 @@ fs_make_symlink(const char *target, const char *path, bool force)
 const char *
 fs_user_home(void)
 {
-	return getenv("HOME");
+	return os_get_env("HOME");
 }
 
 bool
@@ -337,9 +343,6 @@ fs_is_a_tty_from_fd(int fd)
 	if (isatty(fd) == 1) {
 		return true;
 	} else {
-		if (errno != ENOTTY) {
-			LOG_W("isatty() failed: %s", strerror(errno));
-		}
 		return false;
 	}
 }
@@ -395,7 +398,7 @@ fs_find_cmd(struct workspace *wk, struct sbuf *buf, const char *cmd)
 		}
 	}
 
-	if (!(env_path = getenv("PATH"))) {
+	if (!(env_path = os_get_env("PATH"))) {
 		LOG_E("failed to get the value of PATH");
 		return false;
 	}
@@ -426,3 +429,22 @@ fs_find_cmd(struct workspace *wk, struct sbuf *buf, const char *cmd)
 
 	return false;
 }
+
+bool
+fs_has_extension(const char *path, const char *ext)
+{
+	char *s;
+
+	s = strrchr(path, '.');
+	if (!s) {
+		return false;
+	}
+
+	return strcmp(s, ext) == 0;
+}
+
+FILE *
+fs_make_tmp_file(const char *name, const char *suffix, char *buf, uint32_t len)
+{
+	return 0;
+}
