@@ -19,6 +19,7 @@
 #include "platform/assert.h"
 #include "platform/filesystem.h"
 #include "platform/mem.h"
+#include "platform/os.h"
 #include "platform/path.h"
 #include "platform/windows/log.h"
 #include "platform/windows/win32_error.h"
@@ -93,7 +94,7 @@ fs_exe_exists(const char *path)
 
 	size_low = GetFileSize(h, &size_high);
 	if ((size_low == INVALID_FILE_SIZE) && (GetLastError() != NO_ERROR)) {
-		LOG_I("can not get the size of file %s", path);
+		/* LOG_I("can not get the size of file %s", path); */
 		goto close_file;
 	}
 
@@ -108,44 +109,44 @@ fs_exe_exists(const char *path)
 	 * the rest is useless for us.
 	 */
 	if (size < 64) {
-		LOG_I("file %s is too small", path);
+		/* LOG_I("file %s is too small", path); */
 		goto close_file;
 	}
 
 	fm = CreateFileMapping(h, NULL, PAGE_READONLY, 0UL, 0UL, NULL);
 	if (!fm) {
-		LOG_I("Can not map file: %s", win32_error());
+		/* LOG_I("Can not map file: %s", win32_error()); */
 		goto close_file;
 	}
 
 	base = MapViewOfFile(fm, FILE_MAP_READ, 0, 0, 0);
 	if (!base) {
-		LOG_I("Can not view map: %s", win32_error());
+		/* LOG_I("Can not view map: %s", win32_error()); */
 		goto close_fm;
 	}
 
 	iter = base;
 
 	if (*((WORD *)iter) != 0x5a4d) {
-		LOG_I("file %s is not a MS-DOS file", path);
+		/* LOG_I("file %s is not a MS-DOS file", path); */
 		goto unmap;
 	}
 
 	offset = *((DWORD *)(iter + 0x3c));
 	if (size < offset + 24) {
-		LOG_I("file %s is too small", path);
+		/* LOG_I("file %s is too small", path); */
 		goto unmap;
 	}
 
 	iter += offset;
 	if ((iter[0] != 'P') && (iter[1] != 'E') && (iter[2] != '\0') && (iter[3] != '\0')) {
-		LOG_I("file %s is not a PE file", path);
+		/* LOG_I("file %s is not a PE file", path); */
 		goto unmap;
 	}
 
 	iter += 22;
 	if ((!((*((WORD *)iter)) & 0x0002)) || ((*((WORD *)iter)) & 0x2000)) {
-		LOG_I("file %s is not a binary file", path);
+		/* LOG_I("file %s is not a binary file", path); */
 		goto unmap;
 	}
 
@@ -188,7 +189,7 @@ fs_mkdir(const char *path, bool exist_ok)
 		if (exist_ok && GetLastError() == ERROR_ALREADY_EXISTS) {
 			return true;
 		}
-		LOG_E("failed to create directory %s: %s", path, win32_error());
+		LOG_E("failed to create directory \"%s\": %s", path, win32_error());
 		return false;
 	}
 
@@ -211,8 +212,12 @@ fs_rmdir(const char *path, bool force)
 }
 
 bool
-fs_copy_file(const char *src, const char *dest)
+fs_copy_file(const char *src, const char *dest, bool force)
 {
+	if (force) {
+		fs_make_writeable_if_exists(dest);
+	}
+
 	if (!CopyFile(src, dest, FALSE)) {
 		LOG_E("failed to copy file %s: %s", src, win32_error());
 		return false;
@@ -297,7 +302,7 @@ fs_make_symlink(const char *target, const char *path, bool force)
 const char *
 fs_user_home(void)
 {
-	return getenv("USERPROFILE");
+	return os_get_env("USERPROFILE");
 }
 
 static inline bool
@@ -469,18 +474,24 @@ fs_find_cmd(struct workspace *wk, struct sbuf *buf, const char *cmd)
 
 		if (fs_exe_exists(buf->buf)) {
 			return true;
-		} else {
-			if (!fs_has_extension(buf->buf, ".exe")) {
-				sbuf_pushs(wk, buf, ".exe");
-				if (fs_exe_exists(buf->buf)) {
-					return true;
-				}
-			}
-			return false;
 		}
+
+		if (!fs_has_extension(buf->buf, ".exe")) {
+			sbuf_pushs(wk, buf, ".exe");
+			if (fs_exe_exists(buf->buf)) {
+				return true;
+			}
+		}
+
+		return false;
 	}
 
-	if (!(env_path = getenv("PATH"))) {
+	if (strcmp(cmd, "cmd") == 0 || strcmp(cmd, "cmd.exe") == 0) {
+		sbuf_pushs(wk, buf, "cmd.exe");
+		return true;
+	}
+
+	if (!(env_path = os_get_env("PATH"))) {
 		LOG_E("failed to get the value of PATH");
 		return false;
 	}
@@ -542,3 +553,22 @@ fs_remove(const char *path)
 
 	return true;
 }
+
+FILE *
+fs_make_tmp_file(const char *name, const char *suffix, char *buf, uint32_t len)
+{
+	static uint32_t unique = 0;
+	++unique;
+
+	char tmp_dir[MAX_PATH + 1];
+	DWORD result = GetTempPath(sizeof(tmp_dir), tmp_dir);
+	if (result == 0) {
+		// Use '.' (current dir) as tmp dir if GetTempPath fails
+		tmp_dir[0] = '.';
+		tmp_dir[1] = 0;
+	}
+
+	snprintf(buf, len, "%s\\__muon_tmp_%d_%s.%s", tmp_dir, unique, name, suffix);
+
+	return fs_fopen(buf, "w+b");
+}
