@@ -10,6 +10,7 @@
 #include <string.h>
 
 #include "lang/string.h"
+#include "log.h"
 
 /*
  * reference:
@@ -29,6 +30,11 @@ path_is_absolute(const char *path)
 		return true;
 	}
 
+	// Handle unix paths as well
+	if (*path == '/') {
+		return true;
+	}
+
 	length = strlen(path);
 	if (length < 3) {
 		return false;
@@ -65,40 +71,3 @@ path_to_posix(char *path)
 		iter++;
 	}
 }
-
-void
-shell_escape(struct workspace *wk, struct sbuf *sb, const char *str)
-{
-	const char *need_escaping = "\"'$ \\><&#\n";
-	const char *s;
-	bool do_esc = false;
-
-	if (!*str) {
-		sbuf_pushs(wk, sb, "\"\"");
-		return;
-	}
-
-	for (s = str; *s; ++s) {
-		if (strchr(need_escaping, *s)) {
-			do_esc = true;
-			break;
-		}
-	}
-
-	if (!do_esc) {
-		sbuf_pushs(wk, sb, str);
-		return;
-	}
-
-	sbuf_push(wk, sb, '\"');
-
-	for (s = str; *s; ++s) {
-		if (*s == '\"') {
-			sbuf_pushs(wk, sb, "\\\"");
-		} else {
-			sbuf_push(wk, sb, *s);
-		}
-	}
-
-	sbuf_push(wk, sb, '\"');
-}
