@@ -16,6 +16,7 @@
 #include "buf_size.h"
 #include "error.h"
 #include "log.h"
+#include "platform/assert.h"
 #include "platform/filesystem.h"
 #include "platform/mem.h"
 #include "platform/path.h"
@@ -23,39 +24,37 @@
 #include "platform/timer.h"
 #include "platform/windows/win32_error.h"
 
-static uint32_t cnt_open = 0;
-
-#define record_handle(__h, v) _record_handle(__h, v, #__h)
+#define record_handle(__h, v) _record_handle(ctx, __h, v, #__h)
 
 static bool
-_record_handle(HANDLE *h, HANDLE v, const char *desc)
+_record_handle(struct run_cmd_ctx *ctx, HANDLE *h, HANDLE v, const char *desc)
 {
 	if (!v || v == INVALID_HANDLE_VALUE) {
 		return false;
 	}
 
-	++cnt_open;
+	++ctx->cnt_open;
 	*h = v;
 	return true;
 }
 
-#define close_handle(__h) _close_handle(__h, #__h)
+#define close_handle(__h) _close_handle(ctx, __h, #__h)
 
 static bool
-_close_handle(HANDLE *h, const char *desc)
+_close_handle(struct run_cmd_ctx *ctx, HANDLE *h, const char *desc)
 {
 	if (!*h || *h == INVALID_HANDLE_VALUE) {
 		return true;
 	}
 
-	assert(cnt_open);
+	assert(ctx->cnt_open);
 
 	if (!CloseHandle(*h)) {
 		LOG_E("failed to close handle %s:%p: %s", desc, *h, win32_error());
 		return false;
 	}
 
-	--cnt_open;
+	--ctx->cnt_open;
 	*h = INVALID_HANDLE_VALUE;
 	return true;
 }
@@ -67,7 +66,7 @@ enum copy_pipe_result {
 };
 
 static enum copy_pipe_result
-copy_pipe(struct win_pipe_inst *pipe, struct sbuf *sbuf)
+copy_pipe(struct run_cmd_ctx *ctx, struct win_pipe_inst *pipe, struct sbuf *sbuf)
 {
 	if (pipe->is_eof) {
 		return copy_pipe_result_finished;
@@ -75,45 +74,35 @@ copy_pipe(struct win_pipe_inst *pipe, struct sbuf *sbuf)
 
 	DWORD bytes_read;
 
-	if (pipe->is_pending) {
-		if (!GetOverlappedResult(pipe->handle, &pipe->overlapped, &bytes_read, TRUE)) {
-			switch (GetLastError()) {
-			case ERROR_BROKEN_PIPE:
-				pipe->is_eof = true;
-				if (!close_handle(&pipe->handle)) {
-					return copy_pipe_result_failed;
-				}
-				return copy_pipe_result_finished;
-			default: win32_fatal("GetOverlappedResult:"); return copy_pipe_result_failed;
-			}
-		} else {
-			if (bytes_read) {
-				sbuf_pushn(0, sbuf, pipe->overlapped_buf, bytes_read);
-				pipe->overlapped.Offset = 0;
-				pipe->overlapped.OffsetHigh = 0;
-				memset(pipe->overlapped_buf, 0, sizeof(pipe->overlapped_buf));
-			}
-
-			ResetEvent(pipe->overlapped.hEvent);
-		}
-	}
-
-	if (!ReadFile(
-		    pipe->handle, pipe->overlapped_buf, sizeof(pipe->overlapped_buf), &bytes_read, &pipe->overlapped)) {
-		switch (GetLastError()) {
-		case ERROR_BROKEN_PIPE:
+	if (!GetOverlappedResult(pipe->handle, &pipe->overlapped, &bytes_read, TRUE)) {
+		if (GetLastError() == ERROR_BROKEN_PIPE) {
 			pipe->is_eof = true;
 			if (!close_handle(&pipe->handle)) {
 				return copy_pipe_result_failed;
 			}
 			return copy_pipe_result_finished;
-		case ERROR_IO_PENDING: pipe->is_pending = true; break;
-		default: win32_fatal("ReadFile:"); return copy_pipe_result_failed;
 		}
-	} else {
-		pipe->is_pending = false;
-		if (bytes_read) {
-			sbuf_pushn(0, sbuf, pipe->overlapped_buf, bytes_read);
+
+		win32_fatal("GetOverlappedResult:");
+	}
+
+	if (pipe->is_reading && bytes_read) {
+		sbuf_pushn(0, sbuf, pipe->overlapped_buf, bytes_read);
+	}
+	memset(&pipe->overlapped, 0, sizeof(pipe->overlapped));
+	pipe->is_reading = true;
+
+	if (!ReadFile(
+		    pipe->handle, pipe->overlapped_buf, sizeof(pipe->overlapped_buf), &bytes_read, &pipe->overlapped)) {
+		if (GetLastError() == ERROR_BROKEN_PIPE) {
+			pipe->is_eof = true;
+			if (!close_handle(&pipe->handle)) {
+				return copy_pipe_result_failed;
+			}
+			return copy_pipe_result_finished;
+		}
+		if (GetLastError() != ERROR_IO_PENDING) {
+			win32_fatal("ReadFile:");
 		}
 	}
 
@@ -123,45 +112,29 @@ copy_pipe(struct win_pipe_inst *pipe, struct sbuf *sbuf)
 static enum copy_pipe_result
 copy_pipes(struct run_cmd_ctx *ctx)
 {
-	struct {
-		struct win_pipe_inst *pipe;
-		struct sbuf *sbuf;
-	} pipes[2];
-	HANDLE events[2];
-	uint32_t event_count = 0;
+	DWORD bytes_read;
+	struct win_pipe_inst *pipe;
+	OVERLAPPED *overlapped;
 
-#define PUSH_PIPE(__p, __sb)                        \
-	if (!(__p)->is_eof) {                       \
-		pipes[event_count].pipe = (__p);    \
-		pipes[event_count].sbuf = (__sb);   \
-		events[event_count] = (__p)->event; \
-		++event_count;                      \
+	if (!GetQueuedCompletionStatus(ctx->ioport, &bytes_read, (PULONG_PTR)&pipe, &overlapped, 100)) {
+		if (GetLastError() == WAIT_TIMEOUT) {
+			return copy_pipe_result_waiting;
+		} else if (GetLastError() != ERROR_BROKEN_PIPE) {
+			win32_fatal("GetQueuedCompletionStatus:");
+		}
 	}
 
-	PUSH_PIPE(&ctx->pipe_out, &ctx->out);
-	PUSH_PIPE(&ctx->pipe_err, &ctx->err);
-
-#undef PUSH_PIPE
-
-	if (!event_count) {
-		return copy_pipe_result_finished;
-	}
-
-	DWORD wait = WaitForMultipleObjects(event_count, events, FALSE, 100);
-
-	if (wait == WAIT_TIMEOUT) {
-		return copy_pipe_result_waiting;
-	} else if (wait == WAIT_FAILED) {
-		win32_fatal("WaitForMultipleObjects:");
-	} else if (WAIT_ABANDONED_0 <= wait && wait < WAIT_ABANDONED_0 + event_count) {
-		win32_fatal("WaitForMultipleObjects: abandoned");
-	} else if (wait >= WAIT_OBJECT_0 + event_count) {
-		win32_fatal("WaitForMultipleObjects: index out of range");
+	struct sbuf *sbuf = 0;
+	if (pipe == &ctx->pipe_out) {
+		sbuf = &ctx->out;
+	} else if (pipe == &ctx->pipe_err) {
+		sbuf = &ctx->err;
 	} else {
-		wait -= WAIT_OBJECT_0;
+		/* return copy_pipe_result_waiting; */
+		UNREACHABLE;
 	}
 
-	return copy_pipe(pipes[wait].pipe, pipes[wait].sbuf);
+	return copy_pipe(ctx, pipe, sbuf);
 }
 
 static void
@@ -172,9 +145,8 @@ run_cmd_ctx_close_pipes(struct run_cmd_ctx *ctx)
 	}
 
 	close_handle(&ctx->pipe_err.handle);
-	close_handle(&ctx->pipe_err.event);
 	close_handle(&ctx->pipe_out.handle);
-	close_handle(&ctx->pipe_out.event);
+	close_handle(&ctx->ioport);
 
 	// TODO stdin
 #if 0
@@ -195,7 +167,7 @@ run_cmd_collect(struct run_cmd_ctx *ctx)
 
 	bool loop = true;
 	while (loop) {
-		if (!(ctx->flags & run_cmd_ctx_flag_dont_capture)) {
+		if (!(ctx->flags & run_cmd_ctx_flag_dont_capture) && !(ctx->pipe_out.is_eof && ctx->pipe_err.is_eof)) {
 			if ((pipe_res = copy_pipes(ctx)) == copy_pipe_result_failed) {
 				return run_cmd_error;
 			}
@@ -248,17 +220,7 @@ open_pipes(struct run_cmd_ctx *ctx, struct win_pipe_inst *pipe, const char *name
 		name);
 	++uniq;
 
-	if (!record_handle(&pipe->event,
-		    CreateEvent(NULL, // default security attribute
-			    TRUE, // manual-reset event
-			    TRUE, // initial state = signaled
-			    NULL // unnamed event object
-			    ))) {
-		win32_fatal("CreateEvent:");
-	}
-
 	memset(&pipe->overlapped, 0, sizeof(pipe->overlapped));
-	pipe->overlapped.hEvent = pipe->event;
 
 	if (!record_handle(&pipe->handle,
 		    CreateNamedPipeA(pipe_name,
@@ -273,6 +235,10 @@ open_pipes(struct run_cmd_ctx *ctx, struct win_pipe_inst *pipe, const char *name
 		return false;
 	}
 
+	if (!CreateIoCompletionPort(pipe->handle, ctx->ioport, (ULONG_PTR)pipe, 0)) {
+		win32_fatal("CreateIoCompletionPort");
+	}
+
 	if (!ConnectNamedPipe(pipe->handle, &pipe->overlapped) && GetLastError() != ERROR_IO_PENDING) {
 		win32_fatal("ConnectNamedPipe:");
 		return false;
@@ -307,7 +273,6 @@ open_pipes(struct run_cmd_ctx *ctx, struct win_pipe_inst *pipe, const char *name
 	}
 
 	pipe->child_handle = output_write_child;
-	pipe->is_pending = true;
 	return true;
 }
 
@@ -318,6 +283,11 @@ open_run_cmd_pipe(struct run_cmd_ctx *ctx)
 		return true;
 	}
 
+	assert(ctx->ioport == 0);
+	if (!record_handle(&ctx->ioport, CreateIoCompletionPort(INVALID_HANDLE_VALUE, NULL, 0, 1))) {
+		win32_fatal("CreateIoCompletionPort:");
+	}
+
 	sbuf_init(&ctx->out, 0, 0, sbuf_flag_overflow_alloc);
 	sbuf_init(&ctx->err, 0, 0, sbuf_flag_overflow_alloc);
 
@@ -340,51 +310,81 @@ run_cmd_internal(struct run_cmd_ctx *ctx, char *command_line, const char *envstr
 
 	ctx->process = INVALID_HANDLE_VALUE;
 
+	LL("executing: ");
+	if (log_should_print(log_debug)) {
+		log_plain("%s\n", command_line);
+	}
+
 	if (envstr) {
+		LPSTR oldenv = GetEnvironmentStrings();
+
 		const char *k;
 		uint32_t i = 0;
-
+		LL("env:");
 		p = k = envstr;
-		for (;; ++p) {
+		for (; envc; ++p) {
 			if (!p[0]) {
 				if (!k) {
 					k = p + 1;
 				} else {
-					SetEnvironmentVariable(k, p + 1);
+					assert(*k);
+					if (log_should_print(log_debug)) {
+						log_plain(" %s='%s'", k, p + 1);
+					}
+
+					if (!SetEnvironmentVariable(k, p + 1)) {
+						LOG_E("failed to set environment var %s='%s': %s",
+							k,
+							p + 1,
+							win32_error());
+						FreeEnvironmentStrings(oldenv);
+						return false;
+					}
+
 					k = NULL;
+
 					if (++i >= envc) {
 						break;
 					}
 				}
 			}
 		}
-	}
-
-	if (log_should_print(log_debug)) {
-		LL("executing : '%s'\n", command_line);
-
-		if (envstr) {
-			const char *k;
-			uint32_t i = 0;
-			LL("env:");
-			p = k = envstr;
-			for (;; ++p) {
-				if (!p[0]) {
-					if (!k) {
-						k = p + 1;
-					} else {
-						log_plain(" %s='%s'", k, p + 1);
-						k = NULL;
-
-						if (++i >= envc) {
-							break;
-						}
-					}
-				}
-			}
 
+		if (log_should_print(log_debug)) {
 			log_plain("\n");
 		}
+
+		LPSTR newenv = GetEnvironmentStrings();
+
+		// Clear out current env based on newenv
+		char *var;
+		for (var = newenv; *var;) {
+			size_t len = strlen(var);
+			char *split = strchr(var, '=');
+			*split = 0;
+			SetEnvironmentVariable(var, 0);
+			*split = '=';
+			var += len + 1;
+		}
+
+		// Copy newenv into ctx->env
+		sbuf_init(&ctx->env, 0, 0, sbuf_flag_overflow_alloc);
+		sbuf_pushn(0, &ctx->env, newenv, var - newenv);
+		sbuf_push(0, &ctx->env, 0);
+
+		FreeEnvironmentStrings(newenv);
+
+		// Reset env based on oldenv
+		for (var = oldenv; *var;) {
+			size_t len = strlen(var);
+			char *split = strchr(var, '=');
+			*split = 0;
+			SetEnvironmentVariable(var, split + 1);
+			*split = '=';
+			var += len + 1;
+		}
+
+		FreeEnvironmentStrings(oldenv);
 	}
 
 	// TODO stdin
@@ -445,13 +445,17 @@ run_cmd_internal(struct run_cmd_ctx *ctx, char *command_line, const char *envstr
 
 	DWORD process_flags = 0;
 
-	res = CreateProcess(NULL,
+	if (strlen(command_line) >= 32767) {
+		LOG_E("command too long");
+	}
+
+	res = CreateProcessA(NULL,
 		command_line,
 		NULL,
 		NULL,
 		/* inherit handles */ TRUE,
 		process_flags,
-		NULL,
+		ctx->env.buf,
 		ctx->chdir,
 		&startup_info,
 		&process_info);
@@ -483,6 +487,35 @@ run_cmd_internal(struct run_cmd_ctx *ctx, char *command_line, const char *envstr
 	return run_cmd_collect(ctx) == run_cmd_finished;
 }
 
+static void
+run_cmd_push_argv(struct sbuf *cmd, struct sbuf *arg_buf, const char *arg, bool first)
+{
+	sbuf_clear(arg_buf);
+	shell_escape_no_dollar(0, arg_buf, arg);
+	sbuf_pushf(0, cmd, "%s%s", first ? "" : " ", arg_buf->buf);
+}
+
+static void
+run_cmd_push_arg(struct sbuf *cmd, struct sbuf *arg_buf, const char *arg)
+{
+	run_cmd_push_argv(cmd, arg_buf, arg, false);
+}
+
+static bool
+run_cmd_push_arg0(struct run_cmd_ctx *ctx, struct sbuf *cmd, struct sbuf *arg_buf, const char *arg)
+{
+	SBUF_manual(found_cmd);
+	if (!fs_find_cmd(0, &found_cmd, arg)) {
+		ctx->err_msg = "command not found";
+		sbuf_destroy(&found_cmd);
+		return false;
+	}
+
+	run_cmd_push_argv(cmd, arg_buf, found_cmd.buf, true);
+	sbuf_destroy(&found_cmd);
+	return true;
+}
+
 static bool
 argv_to_command_line(struct run_cmd_ctx *ctx,
 	struct source *src,
@@ -491,59 +524,53 @@ argv_to_command_line(struct run_cmd_ctx *ctx,
 	uint32_t argstr_argc,
 	struct sbuf *cmd)
 {
-	const char *argv0, *new_argv0 = NULL, *new_argv1 = NULL;
-
-	if (argstr) {
-		argv0 = argstr;
-	} else {
-		argv0 = argv[0];
-	}
-
-	if (!argv0 || !*argv0) {
-		LOG_E("missing argv0");
-		return false;
-	}
+	bool res = false;
+	SBUF_manual(arg_buf);
+	const char *argv0 = argstr ? argstr : argv[0];
 
 	sbuf_clear(cmd);
 
-	if (!path_is_basename(argv0)) {
-		if (fs_has_extension(argv0, ".bat")) {
-			sbuf_pushf(0, cmd, "cmd.exe /c \"%s\"", argv0);
-		} else if (fs_exists(argv0)) {
-			DWORD _binary_type;
-			if (!GetBinaryType(argv0, &_binary_type)) {
-				if (!run_cmd_determine_interpreter(src, argv0, &ctx->err_msg, &new_argv0, &new_argv1)) {
-					return false;
-				}
+	bool have_arg0 = false;
 
-				/* ignore /usr/bin/env on Windows */
-				if (strcmp(new_argv0, "/usr/bin/env") == 0 && new_argv1) {
-					argv0 = new_argv1;
-					new_argv1 = 0;
-				} else {
-					argv0 = new_argv0;
-				}
+	if (fs_has_extension(argv0, ".bat")) {
+		if (!run_cmd_push_arg0(ctx, cmd, &arg_buf, "cmd.exe")) {
+			goto ret;
+		}
+		run_cmd_push_arg(cmd, &arg_buf, "/c");
+		run_cmd_push_arg(cmd, &arg_buf, argv0);
+		have_arg0 = true;
+	} else if (fs_exists(argv0)) {
+		DWORD _binary_type;
+		if (!GetBinaryType(argv0, &_binary_type)) {
+			const char *new_argv0 = 0, *new_argv1 = 0;
+			if (!run_cmd_determine_interpreter(src, argv0, &ctx->err_msg, &new_argv0, &new_argv1)) {
+				return false;
 			}
 
-			sbuf_pushf(0, cmd, "\"%s\"", argv0);
-		} else {
-			sbuf_pushf(0, cmd, "\"%s\"", argv0);
-		}
-	} else {
-		SBUF_manual(found_cmd);
+			/* ignore /usr/bin/env on Windows */
+			if (strcmp(new_argv0, "/usr/bin/env") == 0 && new_argv1) {
+				new_argv0 = new_argv1;
+				new_argv1 = 0;
+			}
 
-		if (!fs_find_cmd(0, &found_cmd, argv0)) {
-			ctx->err_msg = "command not found";
-			sbuf_destroy(&found_cmd);
-			return false;
-		}
+			if (!run_cmd_push_arg0(ctx, cmd, &arg_buf, new_argv0)) {
+				goto ret;
+			}
 
-		sbuf_pushf(0, cmd, "\"%s\"", found_cmd.buf);
-		sbuf_destroy(&found_cmd);
+			if (new_argv1) {
+				run_cmd_push_arg(cmd, &arg_buf, new_argv1);
+			}
+
+			run_cmd_push_arg(cmd, &arg_buf, argv0);
+
+			have_arg0 = true;
+		}
 	}
 
-	if (new_argv1) {
-		sbuf_pushf(0, cmd, " \"%s\"", new_argv1);
+	if (!have_arg0) {
+		if (!run_cmd_push_arg0(ctx, cmd, &arg_buf, argv0)) {
+			goto ret;
+		}
 	}
 
 	if (argstr) {
@@ -554,7 +581,7 @@ argv_to_command_line(struct run_cmd_ctx *ctx,
 		for (;; ++p) {
 			if (!p[0]) {
 				if (i > 0) {
-					sbuf_pushf(0, cmd, " %s", arg);
+					run_cmd_push_arg(cmd, &arg_buf, arg);
 				}
 
 				if (++i >= argstr_argc) {
@@ -567,11 +594,14 @@ argv_to_command_line(struct run_cmd_ctx *ctx,
 	} else {
 		uint32_t i;
 		for (i = 1; argv[i]; ++i) {
-			sbuf_pushf(0, cmd, " %s", argv[i]);
+			run_cmd_push_arg(cmd, &arg_buf, argv[i]);
 		}
 	}
 
-	return true;
+	res = true;
+ret:
+	sbuf_destroy(&arg_buf);
+	return res;
 }
 
 bool
@@ -628,6 +658,9 @@ run_cmd_ctx_destroy(struct run_cmd_ctx *ctx)
 
 	sbuf_destroy(&ctx->out);
 	sbuf_destroy(&ctx->err);
+	sbuf_destroy(&ctx->env);
+
+	assert(ctx->cnt_open == 0);
 }
 
 bool
