@@ -14,6 +14,7 @@
 #include <strsafe.h>
 #include <stdlib.h>
 
+#include "lang/string.h"
 #include "log.h"
 #include "platform/windows/win32_error.h"
 
@@ -25,7 +26,6 @@ win32_error(void)
 	LPTSTR msg;
 	DWORD err;
 
-	*_msg = '\0';
 	err = GetLastError();
 
 	if (!FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
@@ -35,20 +35,21 @@ win32_error(void)
 		    (LPTSTR)&msg,
 		    0UL,
 		    NULL)) {
-		StringCchPrintf(_msg, sizeof(_msg), "FormatMessage() failed with error Id %ld", GetLastError());
+		snprintf(_msg, sizeof(_msg), "FormatMessage() failed with error Id %ld", GetLastError());
 		return _msg;
 	}
 
-	StringCchCat(_msg, sizeof(_msg), msg);
-	LocalFree(msg);
-
 	// strip trailing newlines from the error message
-	char *end = &_msg[strlen(_msg) - 1];
-	while (end > msg && strchr("\t\r\n ", *end)) {
+	char *end = &msg[strlen(msg) - 1];
+	while (end > msg && is_whitespace(*end)) {
 		*end = 0;
 		--end;
 	}
 
+	snprintf(_msg, sizeof(_msg), "%s (%lu)", msg, err);
+
+	LocalFree(msg);
+
 	return _msg;
 }
 
